//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_MANIPULATOR_H
#define GF_MANIPULATOR_H

#include <input/gupdate.h>
#include "gfdevice.h"
#include "gfpick.h"
#include "gflocat.h"
#include "gfmover.h"
#include "gfscene.h"
#include "gfbbox.h"
#include "gftop.h"

extern const GString gfManipulatorID;

//@ManMemo: A complex input class for object and camera manipulations.
class GFManipulator: public GFInputDevice {
  public:
    GFManipulator(const char*, GFLookatCamera&, GFScene&, 
		  GInputDevice* = (GInputDevice*)gNoInputDevice);
    virtual ~GFManipulator(void);

    void setObject(GFPrimitive*);
    // Set the primary primitive.

    void addObject(GFPrimitive*);
    // Add an object to the brothers list.
    // Must have the same parent as the primary
    // primitive.

    GFPrimitive* getObject(void) const {
	return(mPrimitive);
    }

    const GDLList<GFPrimitive>& getBrothers(void) {
        return(mBrothers);
    }

    virtual const char* getDescription(void) const {
	return("An integrated input device for both object and camera \
manipulations. The manipulator will be triggered after object selections, \
deselections and manipulations. And it may be linked to other manipulators \
for synchronization purposes. A manipulator also updates the associated \
display if necessary. All linked manipulators must invoke the same set of \
callback procedures to be consistent.\n\nA left mouse button click selects \
the hit object, a second click deselects. Left click and pressed Shift key \
selects the object's father. Left click and pressed Control key additionally \
selects a primitive, but only if the new primitive has the same father object \
as the primarily selected one. The selection of a primitive clears the list \
of brother objects. Mouse motions with pressed middle button invoke object/\
camera translations in y direction. Mouse motions with pressed right button \
invoke object translations in x-z plane or camera translations/rotations in \
x-z plane.\n\nA red bounding box marks the primarily selected object. Further \
blue boxes mark brother objects.");
    }
    void link(GFManipulator* pMani) {
	mLinks.append(pMani);
    }

    void unlink(GFManipulator* pMani) {
	mLinks.remove(pMani);
    }

    void translate(const GVector3&);

    void moveTo(const GPoint3&);

    void rotate(GFloat);

    void updateBoxes(void);

    void fixCamera(void) {
	mChangeCam = false;
	mLoc->off();
    }

    void drawBoxes(void);
    void drawCoordinates(const GVector3& MIN, const GVector3& MAX,
			 const GVector3& LITTLE_RASTER, const GVector3& BIG_RASTER);
  public:
    GFScene& getScene(void) {
	return(mScene);
    }

    GFLookatCamera& getCamera(void) {
	return(mRend);
    }

    virtual bool event(GEvent*);

    void linkUpdate(GFManipulator&);
    GAggregate<GFManipulator>& getLinkedManipulators(void) { return(mLinks); }
  public:
    static GFManipulator* getByName(const char*);
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    void trigger(void);
  private:
    // the primarily selected primitive:
    GFPrimitive* mPrimitive;
    // a list of additionally selected brothers:
    GDLList<GFPrimitive> mBrothers;

    GFLookatCamera& mRend;
    GFScene& mScene;
    bool mChangeCam;

    GFLocator* mLoc;
    GFMover* mMov;
    GUpdatePixmap* mUpd;
    GFBoundingBox* mBox;
    GFBoundingBox* mCenter;
    GFTop* mBrotherBoxes;
    GFTop* mCoord;
    
    GDLList<GFManipulator> mLinks;

    class PickCB: public GCallback {
      public:
	PickCB(GFManipulator& pManip): mManip(pManip) {}
	virtual void action(GInputDevice&);
      private:
	GFManipulator& mManip;
    };
    PickCB pickCB;

    class LocatorCB: public GCallback {
      public:
	LocatorCB(GFManipulator& pManip): mManip(pManip) {}
	virtual void action(GInputDevice&);
      private:
	GFManipulator& mManip;
    };
    LocatorCB locatorCB;

    class UpdateCB: public GCallback {
      public:
	UpdateCB(GFManipulator& pManip): mManip(pManip) {}
	virtual void action(GInputDevice&);
      private:
	GFManipulator& mManip;
    };
    UpdateCB updateCB;

    class LinkCB: public GCallback {
      public:
	LinkCB(GFManipulator& pManip): mManip(pManip) {}
	virtual void action(GInputDevice&);
      private:
	GFManipulator& mManip;
    };
    LinkCB linkCB;

    static GDLList<GFManipulator> sList;
};

#endif
