//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_PRIMITIVE_H
#define GF_PRIMITIVE_H

#include <ggeneric.h>
#include <gprimitv.h>
#include <model/gmodel.h>
#include <model/gresol.h>

#include "gfmaterl.h"
#include "gfobject.h"
#include "gfrender.h"

class GFScene;
class GFPrimitive;

extern int gfIsNotRoot;
extern bool& gfFillstyle;
extern const GString gfRoot;
extern GFScene* gfTopicalScene;
extern const GFPrimitive* gfNoFather;
extern bool gfTransparency;
extern bool gfBBoxUpdate;

typedef enum GFDirection {
    GF_NX = 1,
    GF_PX,
    GF_NY,
    GF_PY,
    GF_NZ,
    GF_PZ
};

typedef enum GFManipulationMode {
    GF_NO_MANIPULATION,
    GF_FREE_MANIPULATION,
    GF_SNAPPING,
    GF_RASTERIZATION,
    GF_SNAPPING_AND_RASTERIZATION
};

typedef enum GFAxis {
    GF_NO_AXIS,
    GF_X_AXIS,
    GF_Y_AXIS,
    GF_Z_AXIS,
    GF_ALL_AXES
};

class GFManipulationControl {
  public:
    GFManipulationControl(void): mMode(GF_FREE_MANIPULATION), mRadius(0.1), 
    mRaster(0.1) {} 
    void setMode(GFManipulationMode pMode) {
	mMode = pMode;
    }
    GFManipulationMode getMode(void) {
	return(mMode);
    }
    void setRadius(GFloat pRadius) {
	mRadius = pRadius;
    }
    GFloat getRadius(void) {
	return(mRadius);
    }
    void setRaster(GFloat pRaster) {
	mRaster = pRaster;
    }
    GFloat getRaster(void) {
	return(mRaster);
    }
  private:
    GFManipulationMode mMode;
    GFloat mRadius;
    GFloat mRaster;
};

extern GFManipulationControl gfXAxisControl, gfYAxisControl, gfZAxisControl;

typedef enum {
    GF_NOT_SELECTABLE,
    // The object cannot be selected.
    GF_SELECTABLE,
    // The object itself may be selected.
    GF_SELECT_FATHER
    // Try to select the father or nothing.	
} GFSelectionMode;

typedef enum {
    GF_PRIMITIVE,
    // The object is a primitive object
    // and is not built from other ones.
    GF_UNCHANGED,
    // The object hierarchy root by this object
    // was not changed since creation.
    GF_CHANGED
    // The object itself or a child was changed.
} GFWriteMode;

//@ManMemo: Abstract GF basic primitive.
class GFPrimitive: public GFObject, public GPrimitive, public GModelingAttribute, 
public GResolutionAttribute, public GFMaterialAttribute, public GShadable, public GRayTraceable {
  protected:
    //@ManMemo: Initialization of the primitive.
    GFPrimitive(const char*, GFPrimitive* = (GFPrimitive*)gfNoFather);
  public:
    virtual ~GFPrimitive(void);

    //@ManMemo: Check type compatibility.
    virtual bool isA(const GString& TYPE) const = 0;

    //@ManMemo: Inquire type identity.
    virtual const char* getClass(void) const = 0;
    
    // "positional modeling":
    // The position is the point with the lowest x,
    // lowest y and lowest z coordinate of the bounding
    // box in model coordinates.
    // set position relatively:
    void setPosition(const GPoint3& NEW_POSITION);

    //@ManMemo: Move to specified WC position.
    void moveTo(const GPoint3& NEW_POSITION);
    
    //@ManMemo: Translation in world coordinates.
    virtual void translate(const GVector3& VECTOR);

    //@ManMemo: Acumulating local rotation around all three axes.
    virtual void rotate(const GVector3& VECTOR);  
  public:
    //@ManMemo: Save ASCII description of object to output stream.
    void save(ostream& STREAM, bool SAVE_HEADER = false);
  private:
    //@ManMemo: Save ASCII representation of the constructor arguments.
    virtual void saveConstr(ostream& STREAM) = 0;

    //@ManMemo: Save ASCII representation of physically existent attributes.
    void saveModeling(ostream& STREAM, bool IS_ROOT);
 
    //@ManMemo: The object is an element of another object.
    void beAnElement(void);

    //@ManMemo: Insert passed primitive into element list and increase element pointer.
    void setElement(GFPrimitive* CHILD);
  public:
    const GString& saveName(bool FATHERS_NAME = false);
  protected:
    //@ManMemo: Save children, too.
    virtual void saveChildren(ostream& STREAM);
  protected:
    //@ManMemo: Mark recursively the object changed.
    void setChanged(void);

    //@ManMemo: Mark recursively the father changed.
    void setFatherChanged(void);

    //@ManMemo: Mark the object unchanged.
    void setUnchanged(void) {
	mWriteMode = GF_UNCHANGED;
    }
    
    bool wasChanged(void) const {
	return(mWriteMode == GF_CHANGED);
    }

    void XmodelingChanged(void);
    // Specific changed-function only to be called for the
    // object being changed.
  public:
    const GString& add(const GPoint3& POSITION, const char* CLASS, 
		       const char* ARGUMENTS);
    // Add an element object of specified CLASS at 
    // specified POSITION by evaluating
    // the passed constructor ARGUMENTS.
    // Returns name of created widget relatively.
    
    GFPrimitive* getFather(void) {
	return(mFather);
    }

    bool isElement(void) const {
	return(mIsElement);
    }

    GDLList<GFPrimitive>& getElements(void) {
	return(mElements);
    }

    GDLList<GFPrimitive>& getChildren(void) {
	return(mChildren);
    }

    GPoint3 getPosition(void) const;
    // METHOD: Inquire the local position.

    // set/get pickability:
    void setSelectionMode(GFSelectionMode pMode) {
	mMode = pMode;
    }
    GFSelectionMode getSelectionMode(void) const {
	return(mMode);
    }

    void setTranslationAxis(GFAxis pTrAxis) {
	mTrAxis = pTrAxis;
    }

    GFAxis getTranslationAxis(void) {
	return(mTrAxis);
    }

    void setRotationAxis(GFAxis pRtAxis) {
	mRtAxis = pRtAxis;
    }

    GFAxis getRotationAxis(void) {
	return(mRtAxis);
    }

    GFloat getDistance(GFDirection DIRECTION) const;

    void setCommand(const GString& COMMAND);
    const GString& getCommand(void) const;

    void setElementType(const GString& TYPE);
    const GString& getElementType(void) const;

    const GBounds3& getLocalBounds(void) const;
  public:
    // Shading interface:
    virtual void shade(void);
    // Ray-tracing interface:
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
    virtual const GSurface& getSurface(const GPoint3&);
    virtual GFloat getTransmission(void) const {
	return(getMaterial().getSurface().getTransparency());
    }

    virtual GFloat getNearest(const GRay3& pRay, GShadable*& rShad) {
	GFPrimitive* tPrim = (GFPrimitive*)rShad;
	GFloat tDist = getNearest(pRay, tPrim);
	rShad = (GShadable*)tPrim;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3& pRay, GRayTraceable*& rRay) {
	GFPrimitive* tPrim = (GFPrimitive*)rRay;
	GFloat tDist = getNearest(pRay, tPrim);
	rRay = (GRayTraceable*)tPrim;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3&, GFPrimitive*&);

    // compute and return bounding box in WC:
    const GBounds3& computeBounds(bool);
    // This may be done downwards the hierarchy (true),
    // or upwards (false).

    void logicCheck(bool);
    // evaluate logic check. If parameter is true call the
    // children recursively.

    const GSurface& getMappedSurface(const GPoint3&, const GSurface&);

  protected:
    virtual void materialChanged(void);
    virtual void resolutionChanged(void);
    virtual void modelingChanged(void);
    // The three methods are transferring changes down.
    // The have to be called in overloaded methods of subclasses.

    //@ManMemo: Set local modeling matrix.
    void setMatrix(const GMatrix4&);
  protected:
    virtual GFRenderable* getRenderable(void) {
	return(gfNoRenderable);
    } 
  private:
    //@ManMemo: Add a primitive as element.
    void add(GFPrimitive*);

    //@ManMemo: Remove a primitive from both children and element list.
    void remove(GFPrimitive*);

    virtual void addChild(GFPrimitive*);
    
    //@ManMemo: Get pointer to primitive with specified name.
    void boundingBox(GVector3&, GFloat, GFloat, GFloat);
  public:
    static GFPrimitive* getByName(const char*);
    static void hDelete(GFPrimitive*);
  protected:
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    GFPrimitive* mFather;
    // all components of the composite:
    GDLList<GFPrimitive> mChildren;
    // list of elements:
    // Elements is a sub-set of the children.
    GDLList<GFPrimitive> mElements;
    // Elements is a sub-set of the children.
    int mElCnt;
    GFSelectionMode mMode;
    GRayTraceableBounds mWCBounds; 
    GFWriteMode mWriteMode;
    bool mIsElement;
    GString mCommand;
    bool mCmdExists;
    bool mCmdCalled;
    GFAxis mTrAxis, mRtAxis;
    GString mElType;
    static GDLList<GFPrimitive> sNamedPrimitives;
};

template <class T>
//@ManMemo: Get names of primitives into a string list.
class GFGetNameIterator: public GIterator<T> {
  public:
    GFGetNameIterator(void): mList(20) {}

    void start(void) { mList = "{"; }
    void stop(void) { mList += '}'; }

    virtual bool exec(T& t) {
	mList += t.getName();
	mList += ' ';
	return(false);
    }
    const char* get(void) { return(mList); }
  private:
    GString mList;
};

inline void GFPrimitive::setFatherChanged(void) {
    if (getFather() != gfNoFather) getFather()->setChanged();
}

template <class T>
//@ManMemo: Save primitives to specified stream.
class GFSavePrimitiveIterator: public GIterator<T> {
  public:
    GFSavePrimitiveIterator(ostream& pStream): mStream(pStream) {}
    virtual bool exec(T& t) {
	t.save(mStream);
	return(false);
    }
  private:
    ostream& mStream;
};

// set/get a string specifying the virtual environment:
extern void gfSetEnvironment(const char*);
extern void gfSetEnvironment(const GString&);
extern const GString& gfGetEnvironment(void);

// compute distance to next primitive:
extern GFloat gfGetDistance(GPoint3 POINT, GVector3 DIRECTION);

#endif

