//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// $Id: 3ds.h,v 1.1 1996/02/21 02:20:15 juhu Exp $


#ifndef __3DS_H__
#define __3DS_H__

#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "util/gvector.h"
#include "render/graytrac.h"
#include "util/gcolor.h"
#include "util/gmatrix.h"
#include "util/gdyna.h"

#include "smoother.h"

//////////////////////////////////////////////////////////////////////////////
// @Utils@
//////////////////////////////////////////////////////////////////////////////
void error(const char *err);
void cleanupName (char *name);



//////////////////////////////////////////////////////////////////////////////
// @Types@
//////////////////////////////////////////////////////////////////////////////

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;


//////////////////////////////////////////////////////////////////////////////
// A class for DOS-binaries streams (Yuk!)
//////////////////////////////////////////////////////////////////////////////
class DOS_ifstream: public ifstream {
  public:
    DOS_ifstream() : ifstream() { }
    DOS_ifstream(const char *name, int mode=ios::in, int prot=0664)
    : ifstream(name, mode, prot) { }

  
    DOS_ifstream& operator >>(byte& ret) {
	read(&ret, 1);
      
	return (*this);
    }
    
    
    DOS_ifstream& operator >>(word& ret) {
	byte b1=0, b2=0;
	(*this) >> b1 >> b2;

	ret = (b2 << 8) + b1;
    
	return (*this);
    }
    
    DOS_ifstream& operator >>(dword& ret) {
	word ret1=0, ret2=0;
	(*this) >> ret1 ;
	(*this) >> ret2 ;
	ret = (ret2 << 16) + ret1;
	
	return (*this);
    }

    
    DOS_ifstream& operator>>(float& ret) {
	word ret1=0, ret2=0;
	dword *t;
	
	(*this) >> ret1 ;
	(*this) >> ret2 ;
	
	t = (dword*)&ret;
	*t = (ret2 << 16) + ret1;
    
	return (*this);
    }
   


    DOS_ifstream& operator >> (GVector3& point) {
	float x, y, z;
	(*this) >> x >> y >> z;
	point.setX(x); point.setY(y); point.setZ(z);
	return (*this);
    }

    DOS_ifstream& operator >> (GMatrix4& matrix) {
	for (register int i = 0; i < 4; i++) {
	    for (register int j = 0; j < 3; j++) {
		float tElem;
		(*this) >> tElem;
		matrix.set(i, j, tElem);
	    }
	}
    }
 
    DOS_ifstream& operator >> (char *name) {
	for (register int i = 0; i < 80; i++) {
	    get(name[i]);
	      
	    if (name[i] == '\0')
	      break;
	}

	return (*this);
    }
    
};




class Color {
  friend DOS_ifstream& operator >> (DOS_ifstream& rIN, Color& rColor);
  friend ostream& operator << (ostream& rOUT, const Color& rColor) {
      return (rOUT << rColor.mR << " " << rColor.mG << " " << rColor.mB);
  }
    
  public:
    Color(float pR=0.0, float pG=0.0, float pB=0.0) {
	mR = pR;
	mG = pG;
	mB = pB;
    }

    Color(const Color& pColor) {
	mR = pColor.mR;
	mG = pColor.mG;
	mB = pColor.mB;
    }
    
    void setRGB(float pR=0.0, float pG=0.0, float pB=0.0) {
	mR = pR;
	mG = pG;
	mB = pB;
    }
    
    void getRGB(float& rR, float& rG, float& rB) const {
	rR = mR ;
	rG = mG;
	rB = mB;
    }
    
  private:
    float mR, mG, mB;
    
};

class Percentage {
  friend DOS_ifstream& operator >> (DOS_ifstream& rIN, Percentage& rP);
  
  public:    
    float operator * (const float& pF) {
	return (mP * pF);
    }
    float operator () (void) const {
	return (mP);
    }
    
  private:
    float mP;
    
};

class Face {
  friend DOS_ifstream& operator >> (DOS_ifstream& rIN, Face& rF) {
      return ( rIN >> rF.a >> rF.b >> rF.c >> rF.flags);
  }
    
  public:
    word a, b, c; 
    word flags;
};


//////////////////////////////////////////////////////////////////////////////
// @NObject@
// Basis class for named objects
//////////////////////////////////////////////////////////////////////////////
class NObject {
  public:
    NObject(const char *pName) {
	strncpy(mName, pName, 80);
    }
    const char *getName(void) const {
	return(mName);
    }
    void setName(const char *pName) {
	strncpy(mName, pName, 80);
    }
    
    
  protected:
    char mName[80];
};




// Omni light command 
class OmniLight : public NObject {
  public:
    OmniLight(const char *pName, const GPoint3& pPos, const Color& pCol)
    :NObject(pName), mPos(pPos), mCol(pCol) {
    }
    const GPoint3& getPos(void) const {
	return(mPos);
    }
    const Color& getColor(void) const {
	return(mCol);
    }
    
  private:
    GPoint3 mPos; // Light position 
    Color mCol;   // Light colour 
};


// Spotlight command 
class  SpotLight: public OmniLight {
  public:
    SpotLight(const char *pName, 
	      const GPoint3& pPos, 
	      const Color& pCol, 
	      const GPoint3& pTarget,
	      float pHotspot, 
	      float pFalloff, 
	      bool pShadowFlag = true)
    :OmniLight(pName, pPos, pCol), mTarget(pTarget),
     mHotspot(pHotspot), mFalloff(pFalloff), mShadowFlag(pShadowFlag) {
     }
    

  private:
    GPoint3   mTarget;         // Spotlight target location 
    float     mHotspot;        // Hotspot angle (degrees) 
    float     mFalloff;        // Falloff angle (degrees) 
    int       mShadowFlag;    // Shadow flag (not used) 
};


// Camera command 
class Camera : public NObject {
  friend DOS_ifstream& operator >> (DOS_ifstream& rIN, Camera& rC) {
      return(rIN >> rC.mPos >> rC.mTarget >> rC.mBank >> rC.mLens);
  }
    
  public:
    Camera(void) 
    : NObject("") { }
    
    Camera(const char *pName, const GPoint3& pPos, const GPoint3& pTarget,
	   float pBank, float pLens)
    :NObject(pName), mPos(pPos), mTarget(pTarget),
     mBank(pBank), mLens(pLens) {
     }
    
    float getFOV(void);
    
    const GPoint3& getPos(void) const {
	return(mPos);
    }
    
    const GPoint3& getTarget(void) const {
	return(mTarget);
    }
    
  private: 
    GPoint3 mPos;           // Camera location 
    GPoint3 mTarget;        // Camera target 
    float  mBank;           // Banking angle (degrees) 
    float  mLens;           // Camera lens size (mm) 
    
};


// Material list 
class  Material : public NObject {
  public:
    Material(const char* pName, bool pExt) 
    : NObject(pName)
    {
	mExternal = pExt;
    }
    
  private:
    bool  mExternal;         // Externally defined material? 
};



//////////////////////////////////////////////////////////////////////////////
// Material property 
// @MatProp@
//////////////////////////////////////////////////////////////////////////////
class MatProp : public NObject {
  friend class Scene3DS;
    
  public:
    MatProp(const char *pName = "Default") 
    :NObject(pName)
    {
	mAmbient.setRGB();
	mDiffuse.setRGB(1, 1, 1);
	mSpecular.setRGB();
	mShininess = 4.824;
	mTransparency = 0.0;
	mReflection = 0.0;
	mSelfIllum = false;

	strcpy (mTexMap, "");
	mTexStrength = 0.0;

	strcpy (mBumpMap, "");
	mBumpStrength = 0.0;

    }
    
  public:
    Color  mAmbient;
    Color  mDiffuse;
    Color  mSpecular;
    float  mShininess;
    float  mTransparency;
    float  mReflection;
    bool   mSelfIllum;
    char   mTexMap[40];
    float  mTexStrength;
    char   mBumpMap[40];
    float  mBumpStrength;
};


//////////////////////////////////////////////////////////////////////////////
// A mesh object 
// @Mesh@
//////////////////////////////////////////////////////////////////////////////
class Mesh : public NObject {
    friend class Scene3DS;
  public:
    Mesh (const char *pName, int pNumVerts, int pNumFaces);
    ~Mesh(void);
    
    int getNumVerts(void) const {
	return(mNumVerts);
    }
    
    int getNumFaces(void) const {
	return(mNumFaces);
    }
    
    const GPoint3* getVertex(void) const {
	return (mVertex);
    }

    const GVector3* getNormals(void) const {
	return (mNormals);
    }

    const Face* getFaces(void) const {
	return (mFaces);
    }
    

    const Material** getMaterials(void) const {
	return (mMtl);
    }
    
    bool isHidden(void) const {
	return(mHidden);
    }

    bool isSmooth(void) const {
	return (mSmooth);
    }
    
    bool isShadow(void) const {
	return(mShadow);
    }

    bool computeNormals(void);
    
    void getSmoothData(PolySmoother *pSmoother) const;
    
  private:

    word  mNumVerts;        // Number of vertices 
    GPoint3 *mVertex;       // List of object vertices 
    word  mNumFaces;        // Number of faces 
    Face *mFaces;           // List of object faces 
    GVector3 *mNormals;     // list of normals for each face 
    Material **mMtl;        // Materials for each face 
    GMatrix4 mMatrix;       // Local mesh matrix 
    bool mHidden;           // Hidden flag 
    bool mShadow;           // Shadow flag 
    bool mSmooth;           // Smoothing flag
};




class Chunk {
  public:
    Chunk(DOS_ifstream& rIN) 
    :mIN(rIN) {
	start();
    }
        
    void start(void) {
	mStart = mIN.tellg();
	mIN >> mTag;
	mIN >> mLength;
	mEnd = mStart + mLength;
    }
        
    void end(void) { 
	mIN.seekg(mEnd);
    }

    void next(void) { 
	mIN.seekg(mEnd);
	start();
    }



    bool operator <= (const Chunk& pChunk) {
	return (mEnd <= pChunk.mEnd);
    }
    bool operator < (const Chunk& pChunk) {
	return (mEnd < pChunk.mEnd);
    }

    word tag(void) {
	return(mTag);
    }
    
  private:
    dword mStart;
    dword mEnd;
    dword mLength;
    word  mTag;
    DOS_ifstream& mIN;
    
};




//////////////////////////////////////////////////////////////////////////////
// A class for .3DS parsing
// @Scene3DS@
//
//////////////////////////////////////////////////////////////////////////////
class Scene3DS {
  public:
    Scene3DS(void);
    ~Scene3DS(void);
    
    // read the 3DS-file spezified by file\'s name 
    void read(const char *pInput);

    void setSmoothAngle(GFloat pAngle)  {
	mMinDot = cos(pAngle);
    }
    static GFloat mMinDot; // cosinus of the max. angle for smoothing
    
  protected:
    GBounds3 mWorldBound;

    Color mBGColor; // background color
    bool mUseBGColor;
    
    
    

    // input stream for DOS-binaries
    DOS_ifstream mIN;

    // lists of scene-elements 
    GDynaList<OmniLight *> mOmniList;
    GDynaList<SpotLight *> mSpotList;
    GDynaList<Camera    *> mCamList;
    GDynaList<Mesh      *> mMeshList;
    GDynaList<Material  *> mMtlList;
    GDynaList<MatProp   *> mMPropList ;
    
 
  private:
    
    void parseColor(GColorRGB &rColor);
    void parse3DS (const Chunk &pChunk);
    void parseMData(const Chunk &pChunk);
    void parseNamedObject (const Chunk &pChunk);
    void parseNTriObject (const Chunk &pChunk);
    void parsePointArray(void);
    void parseFaceArray (const Chunk &pChunk);
    Material* updateMaterials (char *pNewMatName, int pExt);
    void parseMshMatGroup(void);
    void parseSmoothGroup(void);
    void parseMatEntry (const Chunk &pChunk);
    void parseNCamera(void);
    void parseNDirectLight (const Chunk &pChunk);
    void parseDLSpotlight(void);



    static char mMsg[255];
    static char mObjName[80];
    static Mesh *mMesh;

};

#endif
