//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan 30 08:14:52 1996 - juhu@rz.tu-ilmenau.de
#ifndef __GT_OBJ_H__
#define __GT_OBJ_H__

#include <time.h>

#include "model/gmodel.h"
#include "gx/gx_obj.h"





class GTPrimitive: public GXPrimitive {
    // CONCEPT: Extended graphics kernel
    //
    // ABSTRACT PURPOSE: Base class for GT primitives 
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    
	

    GTPrimitive(int pF = GRayTraceable::PRIMITIVE, GSurface *pSurf = 0)
    : GXPrimitive(pF), mSurface(pSurf)     {}


    virtual ~GTPrimitive(void) {}

    // get surface at specified WC point:
    virtual const GSurface& getSurface(const GPoint3&) {
 	return((mSurface) ? *mSurface : gDummySurface);
    }
    virtual void setSurface(const GSurface *pSurface) {
 	mSurface = (GSurface *)pSurface;
    }
    virtual GFloat getTransmission(void) const {
	const GSurface  *tSurf = (mSurface) ? mSurface : &gDummySurface;
	return(tSurf->getTransparency());
    }
  
  protected:
    GSurface  *mSurface;
};



class GTSphere: public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE:   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

  public:
    GTSphere(const GPoint3& pCenter = GPoint3(0,0,0), GFloat pRad = 1.0)
    : GTPrimitive(), mCenter(pCenter), mRad(pRad)
    { 
	setFlags(getFlags() | GRayTraceable::CONVEX);
	computeBounds(); 
    }
    
    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3& rPoint) {
	return((rPoint - mCenter).unitize());
    }
    
    void setPos(const GPoint3& pPoint) { 
	mCenter = pPoint;
	computeBounds();
    }
    void setRad(GFloat pRad) { 
	mRad = pRad;
	computeBounds();
    }
    
  private:
    GPoint3 mCenter;
    GFloat mRad;
    virtual const class GBounds3 & computeBounds(void) {
	setWorldBounds() = GBounds3(mCenter - GVector3(mRad, mRad, mRad),
				    mCenter + GVector3(mRad, mRad, mRad));
	return(getWorldBounds());
	
    }
};


class GTDisc : public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: A class for planar disc   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    
    GTDisc(const GVector3& pC, const GVector3& pN, GFloat pR); 
    virtual ~GTDisc() {};

    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3&) {
	return (mNormal);
    }
    
  protected:
    GPoint3 mCenter;
    GVector3 mNormal;		
    GFloat mRad;
    GFloat mDist;			
				
    virtual  const GBounds3& computeBounds(void);
};



typedef GFloat GFloat2[2];

class GTPolygon : public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: A class for planar polygons (direct)   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    GTPolygon(void)
    : GTPrimitive(),
      mNumVerts(0), mVertices(0), mUV_Vertices(0), mValid(true) 
    { 
	setFlags(getFlags() | GRayTraceable::PLANAR);
    }
    
    GTPolygon(int pNumVerts, GPoint3 *pVertices);
    GTPolygon(const GVertexAggregate& pVertices) ;
    virtual ~GTPolygon();

    bool isValid(void) { return (mValid);}
    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3&) {
	return (mNormal);
    }
    
  protected:
    int mNumVerts;		
    GPoint3 *mVertices;	
    GFloat2 *mUV_Vertices;
    bool mValid;		

    GVector3 mNormal;		
    GFloat mDist;			
    	
				
    int mDominantAxis;			
    GFloat mUmin, mUmax;		
    GFloat mVmin, mVmax;		

    virtual void precompute(void);		

        
    virtual const class GBounds3 & computeBounds(void) {
	setWorldBounds() = GBounds3(GPoint3(MAXFLT, MAXFLT, MAXFLT), 
				    GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
	for (register int i = 0; i < mNumVerts; i++)
	    setWorldBounds() += mVertices[i];
	
	GVector3 tDiff(setWorldBounds().min() - setWorldBounds().max());
	for (i = 0; i < 3; i++)
	    if (fabs(tDiff[i]) < gEpsilon) {
		setWorldBounds().min()[i] -= gEpsilon;
		setWorldBounds().max()[i] += gEpsilon;
	    }
	return(getWorldBounds());
	
    }
};


// edge equation  A*X + B*Y + C = 0 
typedef struct {
    GFloat   A, B, C ;	
    int index;
} EdgeSet ;


class GTTriangle : public GTPolygon {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: Special class for triangles ()
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTTriangle(void) : GTPolygon(), mEdges(0) {}
    GTTriangle(GPoint3 *pVertices);
    ~GTTriangle(void);
    
    bool intersect(const GRay3&, GInterSectionList&);

  protected:
    
    EdgeSet *mEdges;
    
    virtual void precompute(void);
};


class GTSmoothTriangle : public GTTriangle {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: Special class for smooth-triangles ()
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTSmoothTriangle(void) : GTTriangle() {}
    GTSmoothTriangle(GPoint3 *pVertices, GVector3 *pNormals);
    virtual ~GTSmoothTriangle(void);

    GVector3 getNormal(const GPoint3&);

  protected:
    GVector3 *mNormals;
    
};




class GTSmoothPolygon : public GTPolygon {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: A class for Phong-Interp polygons
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    GTSmoothPolygon(void)
    : GTPolygon() { }
    
    GTSmoothPolygon(int pNumVerts, GPoint3 *pVertices, GVector3 *pNormals);
    GTSmoothPolygon(const GVertexAggregate& pVertices) ;
    virtual ~GTSmoothPolygon(void);

    GVector3 getNormal(const GPoint3&);
 
  protected:
    GVector3 *mNormals;
};



class GTCone: public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE:   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

  public:
    GTCone(const GPoint3& pBase, GFloat pBaseRad,
	   const GPoint3& pApex, GFloat pApexRad);
    virtual ~GTCone(void) {}
    //bool hitRay(const GRay3& pRay); 
    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3& rPoint);
    
  private:
    virtual const GBounds3& computeBounds(void);
    
    GModeling mModel;
    GFloat mStartDist;
    bool mCylFlag;
};


#endif
