//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan  9 21:22:57 1996 - juhu@rz.tu-ilmenau.de

//////////////////////////////////////////////////////////////////////////////
// GT NFF-parser, w/  extented commands for extended light sources
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
// + rectangular lightsource  
//   Syntax:
//    "lr" Xp Yp Zp Xl Yl Zl width height Wres Hres minS maxS candS [R G B]
//   Descr.: Extended lightsource with rectangular shape, defined by :
//    Xp Yp Zp -> Position (origin) of the shape,
//    Xl Yl Zl -> Lookat
//    width height -> dimension of the shape,
//    Wres Hres -> resolution for the sampling pattern (Jitter-grid over the shape)
//    minS maxS candS -> Parameters for adaptive supersampling ala Michell
//                       minS : min. number of samples to be taken (4)
//                       maxS : max. number of samples to be taken (16)
//                       candS: nummber of random candidates samples, wich are 
//                       required for the generation of  samples w/ approx. 
//                       Poisson-Disk-Distribution
//    [R G B] -> color (opt., default: RGB(1, 1, 1)) 
//

//////////////////////////////////////////////////////////////////////////////
// + spherical lightsource
//   Syntax:
//    "lr" Xp Yp Zp Rad [R G B]
//   Descr.:
//    Extended lightsource with spherical shape, defined by :
//    Xp Yp Zp -> Position (origin) of the sphere
//    Rad -> Radius of the sphere
//    [R G B] -> color (opt., default: RGB(1, 1, 1))
//

#include <fstream.h>
#include <time.h>

#include "render/graycam.h"
#include "render/gpixmap.h"

#include "gt_obj.h"
#include "gx/gx_defs.h"
#include "gx/gx_light.h"
#include "gx/gx_cam.h"

static void prerror(const char *err) {
   cerr << err << '\n';
   exit(3);
}

static long sPointL=0, sRectL=0, sSphL=0, sCone=0, sSph=0, sPoly=0, sBadPoly=0; 


void Read_NFF(const char *pFileName, 
	      GXRayTracer& rRT, 
	      GXPrimitiveList& rSC)
{
    static char nexttok[256];
    char cmd;
    ifstream inpf(pFileName);
    
    if (! inpf) {
	sprintf(nexttok, "Could not open %s", pFileName);
	prerror(nexttok);
    }
    nexttok[0] = '\0';
    
    GSurface      *tSurf = 0;

    while (! inpf.eof()) {
	if (nexttok[0]) {
	    cmd = nexttok[0];
	    nexttok[0] = '\0';
	}
	else {
	    inpf >> cmd;
	    if (! inpf)
	      break;
	}
	
	switch (cmd) {

	 case 'v':  
	    //// Camera
	      
	    GFloat fromx, fromy, fromz;
	    GFloat atx, aty, atz;
	    GFloat upx, upy, upz;
	    GFloat ang, hither;
	    int xres, yres;
	      
	    inpf >> nexttok;
	    if (strcmp(nexttok, "from"))
	      prerror("\"from\" expected");
	    inpf >> fromx >> fromy >> fromz >> nexttok;
	 
	    if (strcmp(nexttok, "at"))
	      prerror("\"at\" expected");
	    inpf >> atx >> aty >> atz >> nexttok;

	    if (strcmp(nexttok, "up"))
	      prerror("\"up\" expected");
	    inpf >> upx >> upy >> upz >> nexttok;

	    if (strcmp(nexttok, "angle"))
	      prerror("\"angle\" expected");
	    inpf >> ang >> nexttok;
	 
	    if (strcmp(nexttok, "hither"))
	      prerror("\"hither\" expected");
	    inpf >> hither >> nexttok;
	    
	    if (strcmp(nexttok, "resolution"))
	      prerror("\"resolution\" expected");
	    inpf >> xres >> yres;
	    
	    if( GXOptions::PixWidth == 0)
	      GXOptions::PixWidth = xres;
	    if(GXOptions::PixHeight == 0)
	      GXOptions::PixHeight = yres;
	    
	    nexttok[0] = '\0';
	    
	    GPoint3 tViewer(fromx, fromy, fromz); 
	    GPoint3 tLookAt(atx, aty, atz);
	    
	    rRT.setViewpoint(tViewer);
	    rRT.setRefpoint(tLookAt);
	    rRT.setZoom(ang);
	    
	    GVector3 tUp(upx, upy, upz);
	    tUp = tUp.unitize();
	    
	    GVector3 tN( tLookAt - tViewer );
	    tN = tN.unitize();

	    GVector3 ye(0, 1, 0);
	    // defaults
    
	    if(tN == GVector3(0, 1, 0))
	      ye = GVector3(0, 0, 1);
	    if(tN == GVector3(0, -1, 0))
	      ye = GVector3(0, 0, -1);
	    
	    GVector3 xe(ye.cross(tN)); 
	    xe = xe.unitize();
	    ye = tN.cross(xe); 
 
	    GVector3 tV = tUp - tN * (tUp.dot(tN));
	    
	    if(tV == GVector3(0, 0, 0)) {
		tV = ye;
		rRT.setTwist(0.0);
	    } else {
		tV = tV.unitize();
		GFloat tTheta = acos(tV.dot(ye));
		if((tV.cross(ye)).dot(tN) < 0.0)
		  tTheta = - tTheta;
		
		rRT.setTwist(M_R2D * tTheta);
	    }
	    
	    
	    
	    break;
	 
	 case 'b': {
	     
	     //// Background
	    
	     GFloat x, y, z;
	     inpf >> x >> y >> z;
	     rRT.setBackground(GColorRGB(x, y, z));
	     break;
	 }
	    //// Light
	 case 'l': {
		  
	     GFloat lx, ly, lz;
	     GFloat cx, cy, cz;
	     char *sc;
	     GXPointLight *tLight;
	     
	     // rectangular light source
	     if (inpf.peek() == 'r') {   
		 inpf >> cmd;
		 
		 GFloat ax, ay, az;
		 GFloat width, height;
		 int Wres, Hres;
		 int min, max, cand;
	    
		 inpf >> lx >> ly >> lz;
		 inpf >> ax >> ay >> az;
		 inpf >> width >> height;
		 inpf >> Wres >> Hres;
		 inpf >> min >> max >> cand;
		 tLight = new GXRectangularLight(GPoint3(lx, ly, lz),
						 GPoint3(ax, ay, az),
						 width, height,
						 Wres, Hres,
						 min, max, cand);
	
		 sRectL++;
	     }

	     // spherical light source
	     else if (inpf.peek() == 's') {   
		 inpf >> cmd;
	     
		 GFloat r;
		 int Wres, Hres;
		 inpf >> lx >> ly >> lz >> r >>Wres >> Hres ;
		 tLight = new GXSphericalLight(GPoint3(lx, ly, lz), r, Wres, Hres);
		 sSphL++;
	     }
	     
	     // point light source
	     else {
		 inpf >> lx >> ly >> lz;
		 tLight = new GXPointLight();
		 tLight->setPosition(GPoint3(lx, ly, lz));
		 sPointL++;
	     }
	     
	     // color of the light source
	     streampos tPos = inpf.tellg();
	     inpf >> nexttok;
	     cx = strtod(nexttok, &sc);
	     if (sc == &(nexttok[0])) {
		 tLight->setColor(GColorRGB(1,1,1));
		 inpf.seekg(tPos);
		 nexttok[0] = '\0';
	     }
	     else {
		 inpf >> cy >> cz;
		 tLight->setColor(GColorRGB(cx, cy, cz));
		 nexttok[0] = '\0';
	     }
	    
	     gGXLights->append(tLight);
	     break;
	 }
	    
	 case 'f':
	    //// Surface
	    
	      {  GFloat r, g, b, kd, ks, shine, t, ior;
		  inpf >> r >> g >> b >> kd >> ks >> shine >> t >> ior;
		  tSurf = new GSurface();
		  tSurf->setAmbient(GColorRGB(0, 0, 0));
		  tSurf->setDiffuse(GColorRGB(kd*r, kd*g, kd*b));
		  tSurf->setSpecular(GColorRGB(ks*r, ks*g, ks*b));
		  tSurf->setShininess(shine);
		  tSurf->setTransparency(t);
		  tSurf->setRefraction(ior);
	      }
	    break;
	    
	 
	 case 's':
	    //// sphere
	    
	      {  GFloat x, y, z, r;
		  inpf >> x >> y >> z >> r;
		  GTSphere *_pSph = new GTSphere(GPoint3(x, y, z), r);
		  _pSph->setSurface((tSurf) ? tSurf : &gDummySurface);
		  rSC.append(_pSph);

		  sSph++;
	      }
	    break;

	 case 'c':
	    //// cone
	    
	      {  GFloat bX, bY, bZ, bR, aX, aY, aZ, aR;
		  inpf >> bX >> bY >> bZ >> bR;
		  inpf >> aX >> aY >> aZ >> aR;
		  GTCone *pCone = new GTCone(GPoint3(bX, bY, bZ), bR,
					     GPoint3(aX, aY, aZ), aR);
		  pCone->setSurface((tSurf) ? tSurf : &gDummySurface);
		  rSC.append(pCone);

		  sCone++;
	      }
	    break;
	    
	 case 'p':
	    GTPolygon *tPoly;
	    
	    if (inpf.peek() == 'p') {   
		
		// polygon patch( $Phong!)
		inpf >> cmd;

		int tNum;
		inpf >> tNum;
		GPoint3 *tVerts = new GPoint3[tNum];
		GVector3 *tNorms = new GVector3[tNum];
		for (register int i = 0; i < tNum ; i++) {
		    GFloat x, y, z, nx, ny, nz;
		    inpf >> x >> y >> z >> nx >> ny >> nz;
		    tVerts[i] = GPoint3(x, y, z);
		    tNorms[i] = GVector3(nx, ny, nz);
		}
		
		if (tNum >  3)
		  tPoly = new GTSmoothPolygon(tNum, tVerts, tNorms);
		else
		  tPoly = new GTSmoothTriangle(tVerts, tNorms);
		
		
		delete [] tVerts;
		delete [] tNorms;
	    }
	    else {		
		int tNum;
		inpf >> tNum;
		GPoint3 *tVerts = new GPoint3[tNum];
		for (register int i = 0; i < tNum ; i++) {
		    GFloat x, y, z;
		    inpf >> x >> y >> z;
		    tVerts[i] = GVector3(x, y, z);
		}
		
		if(tNum > 3)
		  tPoly = new GTPolygon(tNum, tVerts);
		else
		  tPoly = new GTTriangle(tVerts);
		delete [] tVerts;
	    }

	    if(tPoly->isValid()) {
		tPoly->setSurface((tSurf) ? tSurf : &gDummySurface);
		rSC.append(tPoly);
		sPoly++;
	    }
	    else {
		gkillPointer(tPoly);
		cerr << "Warning: bad polygon removed!\n";
		sBadPoly++;
	    }
	    
	    break;
	 
	 case '#':
	    inpf.getline(nexttok, 256);
	    nexttok[0] = '\0';
	    break;
	}
    }

    cerr << "PLights\tRLights\tSLights\tCones\tSpheres\tPolygons\tBad\n";
    cerr << sPointL << "\t" << sRectL<< "\t" << sSphL << "\t" <<
      sCone << "\t" << sSph<< "\t" << sPoly<< "\t" << sBadPoly << "\n\n"; 



}



static void printHelp(void) {
    cerr << "NFF: The NFF-raytracer, based on the GX/GT kernel of GENERIC.\n\n"; 
    cerr << "Usage: nff [options] <filename>\n";
    cerr << "For any flag x, -x means ON, -x- means OFF\n";    
    cerr << "  -s<nnn>          speedup-methodes for primary/secondary/shadow rays:\n"; 
    cerr << "                   3->BSP, 5->Hyper-Octree, 2->Vista/Lightbuffer (default 232)\n";
    cerr << "  -bsp Level Num   setting up the BSP root-tree with :\n";
    cerr << "                   Level -> max. level of subdivision (default 8)\n";
    cerr << "                   Num   -> max. number of objects in a BSP-Node(default 8)\n";
    cerr << "  -hot Level Num   setting up the  root hyper-octree with :\n";
    cerr << "                   Level -> max. level of subdivision (default 4)\n";
    cerr << "                   Num   -> max. number of objects in a hypercube(default 8)\n";
    
    cerr << "  -a               antialiasing ON (default OFF)\n";
    cerr << "  -d               direct display ON (default ON)\n";
    cerr << "  -v               verbose message ON (default ON)\n";
    cerr << "  -p               pause before exit ON (default ON)\n";
    
    cerr << "  -b               activate ABVH for BSP-nodes (default ON)\n";
    cerr << "  -f<T|S>          set output file format: T->Targa, S->SGI-RGB\n";
    cerr << "  -oFile           set output-filename (default OFF)\n";
    cerr << "  -wWidth          set X-resolution\n";
    cerr << "  -hHeight         set Y-resolution\n\n";

}



int main(int argc, char *argv[]) {
    char tInFileName[256]="";
    char tBuf[256];	 
    bool tTargaFormat = true;
    
    if(argc<2) {
	printHelp();
	exit(3);
    }
    

    argc--;
    argv++;
    while (argc--) {
	if (argv[0][0] != '-') {
	    strcpy(tInFileName, argv[0]);
	    argv++;
	    continue;
	}
		
	// antialising ON/OFF
	else if (strncasecmp(argv[0]+1, "a", 1) == 0) {
	    GXOptions::Antialiasing = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::Antialiasing = false;
	}

	// direct display ON/OFF
	else if (strncasecmp(argv[0]+1, "d", 1) == 0) {
	    GXOptions::DisplayFlag = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::DisplayFlag = false;
	}
	// Verbose message ON/OFF
	else if (strncasecmp(argv[0]+1, "v", 1) == 0) {
	    GXOptions::VerboseFlag = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::VerboseFlag = false;
	}
	// pause before exit ON/OFF
	else if (strncasecmp(argv[0]+1, "p", 1) == 0) {
	    GXOptions::PauseFlag = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::PauseFlag = false;
	}
	
	// BSP setup
	else if (strncasecmp(argv[0]+1, "bsp", 3) == 0) {
	    // max. subdivision\'s level
	    if(argc--) GXOptions::BSP_MaxPartitionLevel = atoi(*(++argv));
	    
	    // max. number of objects in a subdivision cell
	    if(argc--) GXOptions::BSP_MaxObjInNode = atoi(*(++argv));
	    
	}
	
	// HyperOctree setup
	else if (strncasecmp(argv[0]+1, "hot", 3) == 0) {
	    // max. subdivision\'s level
	    if(argc--) GXOptions::HOT_MaxPartitionLevel = atoi(*(++argv));
	    
	    // max. number of objects in a subdivision cell
	    if(argc--) GXOptions::HOT_MaxObjInNode = atoi(*(++argv));
	    
	}

	// accelerartion methode
	else if (strncasecmp(argv[0]+1, "s", 1) == 0) {
	    
	    if(strncmp(argv[0]+2, "3", 1) == 0)
		GXOptions::PrimaryAcceleration = GX_BSP;
	    else if(strncmp(argv[0]+2, "5", 1) == 0)
		GXOptions::PrimaryAcceleration = GX_HOT;
	    else if(strncmp(argv[0]+2, "2", 1) == 0)
		GXOptions::PrimaryAcceleration = GX_BUF;

	    if(strncmp(argv[0]+3, "3", 1) == 0)
		GXOptions::SecondaryAcceleration = GX_BSP;
	    else if(strncmp(argv[0]+3, "5", 1) == 0)
		GXOptions::SecondaryAcceleration = GX_HOT;
	    
	    if(strncmp(argv[0]+4, "3", 1) == 0)
		GXOptions::ShadowAcceleration = GX_BSP;
	    else if(strncmp(argv[0]+4, "5", 1) == 0)
		GXOptions::ShadowAcceleration = GX_HOT;
	    else if(strncmp(argv[0]+4, "2", 1) == 0)
		GXOptions::ShadowAcceleration = GX_BUF;
	}
	
	// Output
	else if (strncasecmp(argv[0]+1, "o", 1) == 0) {
	    GXOptions::OutputFlag = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0) {
		GXOptions::OutputFlag = false;
		argv++;
		continue;
	    }
	    strcpy(tBuf, argv[0] + 2);
	}
	
	// ABVH ON/OFF ?
	else if (strncasecmp(argv[0]+1, "b", 1) == 0) {
	    GXOptions::ABVHFlag = true;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::ABVHFlag = false;
	}
	
	// output file format TGA or SGI/RGB ?
	else if (strncasecmp(argv[0]+1, "f", 1) == 0) {
	    
	    if(strncasecmp(argv[0]+2, "s", 1) == 0)
	      tTargaFormat = false;
	}
	
	// Height
	else if (strncasecmp(argv[0]+1, "h", 1) == 0) {
	    GXOptions::PixHeight = atoi(*argv + 2);
	}
	// Width
	else if (strncasecmp(argv[0]+1, "w", 1) == 0) {
	    GXOptions::PixWidth = atoi(*argv + 2);
	}
	else
	    cerr << "Ignoring unknown option " << argv[0] << '\n';
	argv++;
    }

    if(tInFileName[0]==0x0) {
	printHelp();
	exit(3);
    }
    
    
    
    
    //// intialize GENERIC-kernel..
    gInitialize();
   
    
    GXRayTracer  aRT;
    GXPrimitiveList aSc(true);
    gGXLights  = new GDLList<GXLight>(true);

    
    
    
    
    //// reading the NFF-file...
    Read_NFF(tInFileName, aRT, aSc);

    GPixmapDisplay  *aPix = 0;
    GImage *aOutFile = 0;
    if(GXOptions::DisplayFlag) {	      
	aPix = new GPixmapDisplay(tInFileName, 
				  GXOptions::PixWidth,
				  GXOptions::PixHeight);
	aRT.initRealization(*aPix);
    }
    
    if(GXOptions::OutputFlag) {
	if(tTargaFormat)
	  aOutFile = new GTargaImage("");
	else
	  aOutFile = new GSGIImage("");
	if(!strchr(tBuf, '.')) {
	    strcat(tBuf, ".");
	    strcat(tBuf, aOutFile->getExtension());
	}
	strcpy(GXOptions::OutFileName, tBuf);
	aOutFile->setFilename(GXOptions::OutFileName);
	aOutFile->setWidth(GXOptions::PixWidth);
	aOutFile->setHeight(GXOptions::PixHeight);
	aRT.setOutputImage(aOutFile);
    }
    
    //// and trace it now..
    aRT.rendering(aSc);
    
    if(GXOptions::PauseFlag) {
	cerr << "Press any key to exit...\n" ;
	getchar();
    }
    
    gkillPointer(aPix);
    gkillPointer(aOutFile);
    
    return (0);
}



