//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//////////////////////////////////////////////////////////////////////////////
// A test for different sampling patterns.
// The standard Poisson-Disk generation alg. is unstable, so i dont implement
// it in GX. I find the Michell's approximation is quite better.
// A note: For extended light sources, where the number of samples are fixed,
// N-rooks-sampling is the best pattern w/ lowest variance and discrepancy.
// For instance, i use this sampling pattern for spherical light source.  
//////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdlib.h>
#include "util/gvector.h"
#include "render/gpixmap.h"
#include "gx/gx_sampl.h"

int main(int argc, char *argv[]) {
    if(argc != 4) {
	cerr << "Usage: sampling <xres> <yres> <seed>\n";
	cerr << "Where: \n";
	cerr << " - xres, yres -> resolution in the canonical unit square\n";
	cerr << "                 (common values: [3 3], [5 5], [7 7]...\n";
	cerr << " - seed       -> start value for the random-sequence (1)\n\n"; 
	exit(3);
    }
    
    int sw = atoi(argv[1]);
    int sh = atoi(argv[2]);
    srandom(atoi(argv[3]));

    int ph = 400, pw = 400;
    int jh = 40, jw= 40;
    
    GPixmapDisplay  *aPix = new GPixmapDisplay("Sampling", pw, ph);
    GX_Sampling *tSampling;

    register int i;
    register int j;
    int sN = sw * sh;
    GPoint2 *tSamples = new GPoint2[sN];

    cerr << "Standard Poisson-Disk Sampling...\n";
    aPix->activate();
    aPix->clear(GColorRGB(0, 0, 0));
    aPix->swapBuffers();
    for (j = 0; j < ph; j+=jh) {
	for (i = 0; i < pw; i+=jw) {
	    register int k = 0;
	    while( k < sN) {
		tSamples[k].setX(Random_01());
		tSamples[k].setY(Random_01());
		bool tReject = false;
		for(register int l = 0; l < k; l++) {
		    GFloat tDist = (tSamples[k] - tSamples[l]).sqdist();
		    if(tDist < 0.1 * 0.1 ) {
			tReject = true;
			break;
		    }
		}
		
		if(!tReject) k++;
	    }

	    for(k = 0; k < sN; k++) 
	      aPix->putPixel((int)(i + floor((GFloat)jw*tSamples[k].getX())), 
			     (int)(j + floor((GFloat)jh*tSamples[k].getY())), 
			     GColorRGB(1, 1, 1));
	
	}
    } 

    ungetc(getchar(), stderr);

    tSampling = new GX_NRooksSampling(sw, sh);
    
    aPix->activate();
    aPix->clear(GColorRGB(0, 0, 0));
    aPix->swapBuffers();
    cerr << "Michell's Poisson-Disk Sampling...\n";
    
    for (j = 0; j < ph; j+=jh) {
	for (i = 0; i < pw; i+=jw) {
	    tSampling->generateSamples();
    
	    for(register int k = 0; k < tSampling->getNumber(); k++) {
		GPoint2 tSample = tSampling->getSample(k);
		aPix->putPixel((int)(i + floor((GFloat)jw*tSample.getX())), 
			       (int)(j + floor((GFloat)jh*tSample.getY())), 
			       GColorRGB(1, 1, 1));
	    }
	}
    } 
    
    ungetc(getchar(), stderr);
    delete tSampling;
    
    cerr << "Jittered Sampling...\n";
    aPix->activate();
    aPix->clear(GColorRGB(0, 0, 0));
    aPix->swapBuffers();

    tSampling = new GX_JitterSampling(sw, sh);
    for (j = 0; j < ph; j+=jh) {
	for (i = 0; i < pw; i+=jw) {
	    tSampling->generateSamples();
	    for(register int k = 0; k < tSampling->getNumber(); k++) {
		GPoint2 tSample = tSampling->getSample(k);
		aPix->putPixel((int)(i + floor((GFloat)jw*tSample.getX())), 
			       (int)(j + floor((GFloat)jh*tSample.getY())), 
			       GColorRGB(1, 1, 1));
	    }
	}
    } 
    
    ungetc(getchar(), stderr);
    delete tSampling;
        
    cerr << "Half-jittered Sampling...\n";
    aPix->activate();
    aPix->clear(GColorRGB(0, 0, 0));
    aPix->swapBuffers();
    
    tSampling = new GX_HalfJitterSampling(sw, sh);
    for (j = 0; j < ph; j+=jh) {
	for (i = 0; i < pw; i+=jw) {
	    tSampling->generateSamples();
	    for(register int k = 0; k < tSampling->getNumber(); k++) {
		GPoint2 tSample = tSampling->getSample(k);
		aPix->putPixel((int)(i + floor((GFloat)jw*tSample.getX())), 
			       (int)(j + floor((GFloat)jh*tSample.getY())), 
			       GColorRGB(1, 1, 1));
	    }
	}
    } 
    
    ungetc(getchar(), stderr);
    delete tSampling;
    
    
    cerr << "N-rooks Sampling...\n";
    aPix->activate();
    aPix->clear(GColorRGB(0, 0, 0));
    aPix->swapBuffers();
    tSampling = new GX_NRooksSampling(sw, sh);
    for (j = 0; j < ph; j+=jh) {
	for (i = 0; i < pw; i+=jw) {
	    tSampling->generateSamples();
	    for(register int k = 0; k < tSampling->getNumber(); k++) {
		GPoint2 tSample = tSampling->getSample(k);
		aPix->putPixel((int)(i + floor((GFloat)jw*tSample.getX())), 
			       (int)(j + floor((GFloat)jh*tSample.getY())), 
			       GColorRGB(1, 1, 1));
	    }
	}
    } 
    delete tSampling;

    ungetc(getchar(), stderr);
}
