//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////

#include "smoother.h"






// fuzzy comparison macro 
#define FUZZEQ(x,y)  (fabs((x)-(y))<(mFuzz))

int hashVertex(const GPoint3& pPt) {
    return ((int)(( (3*ABSQUANT(pPt.getX()) + 
		     5*ABSQUANT(pPt.getY()) + 
		     7*ABSQUANT(pPt.getZ())) * 
		    HASH_TABLE_SIZE)) % HASH_TABLE_SIZE);
}


// are these two vertices the same to with the tolerance? 
bool PolySmoother::compareVerts(GPoint3 *v0, GPoint3 *v1) {
    GFloat q0, q1;
    q0 = QUANT(v0->getX()); q1 = QUANT(v1->getX()); if (!FUZZEQ(q0, q1)) return(false);
    q0 = QUANT(v0->getY()); q1 = QUANT(v1->getY()); if (!FUZZEQ(q0, q1)) return(false);
    q0 = QUANT(v0->getZ()); q1 = QUANT(v1->getZ()); if (!FUZZEQ(q0, q1)) return(false);
    return(true);
}

PolySmoother::PolySmoother(int pHashTableSize)
{
    mHashTableSize = pHashTableSize;
    mHashTable = new GDynaList<HashNode*>[pHashTableSize];
    
    mEdgeTest = false;
    mMinDot = 0.2;
    mFuzz = mFuzzFraction = gEpsilon;
}

PolySmoother::~PolySmoother(void) {
    for(register int i=0; i < mHashTableSize; i++)
	for(GDynaListIterator<HashNode*> tI1(mHashTable[i]); tI1.more(); tI1.next())
	    delete (tI1.cur());
    gkillArray(mHashTable);
    

    for(GDynaListIterator<Polygon*> tI2(mPolygonTable); tI2.more(); tI2.next())
	delete (tI2.cur());
}

    
void PolySmoother::addPolygon(Polygon *pPoly) {
    
    if(!pPoly->mValid) {
	cerr << "Warning: Bad polygon removed!\n";
	return;
    }
    
    // add the polygon to polygons\'s table
    mPolygonTable.append(pPoly);
    
    // hash the polygon\'s vertices to the hashtable
    int tNumVerts = pPoly->mNumVerts;
    
    for (register int i=0; i < tNumVerts; i++) {
	// find the \'home adress\' of this key
	int tHashIndex = hashVertex(pPoly->mVertices[i]);
	// insert the key to the actual chain
	mHashTable[tHashIndex].prepend(new HashNode(pPoly, i));
    }
}


void PolySmoother::computeNormals(void) {
    
    computeFuzz();
    
    for (register int i=0; i < mHashTableSize; i++) {
	for(;;) {
	    // find the first WAITING node in this bucket
	    bool tFound = false;
	    
	    GDynaList<HashNode*> &tBucket = mHashTable[i];
	    int tBucketSize = tBucket.count();
	    for(register int j=0; j < tBucketSize; j++)
	    {
		HashNode *tNode = tBucket[j];
		if(tNode->mMarked != DONE) {
		    tFound = true;
		    break;
		}
	    }
	    
	    if (tFound) {
		processHashNode(i, j);
	    }
	    else
		break;
	} 
    }
}


void PolySmoother::computeFuzz(void) {
    GPoint3 tMin(MAXFLT, MAXFLT, MAXFLT), 
	tMax(-MAXFLT, -MAXFLT, -MAXFLT);

    for(GDynaListIterator<Polygon*> tItr(mPolygonTable); tItr.more(); tItr.next()) {
	Polygon *tPoly = tItr.cur();
	for (register int i=0; i < tPoly->mNumVerts; i++) {
	    tMin.setMin(tPoly->mVertices[i]);
	    tMax.setMax(tPoly->mVertices[i]);
	}
    } 
    
    GVector3 tDiff = tMax - tMin;
    GFloat tMaxDiff = gmax3(gabs(tDiff.getX()), 
			    gabs(tDiff.getY()), 
			    gabs(tDiff.getZ()));
    
    mFuzz = tMaxDiff * mFuzzFraction;
}

void PolySmoother::processHashNode(int pHomeI, int pBucketI) {
    GDynaList<HashNode*> &tBucket = mHashTable[pHomeI];
    HashNode *tFirstNode = tBucket[pBucketI];
    tFirstNode->mMarked = WORKING;
    
    GPoint3 *tFirstVert = &(tFirstNode->mPolygon->mVertices[tFirstNode->mVertexNum]);
    GVector3 *tHeadNormal = &(tFirstNode->mPolygon->mNormal);
    GVector3 tNormal = tFirstNode->mPolygon->mNormal;
    
    GVector3 *tTestNormal;
    GPoint3  *tTestVert;
    HashNode *tScanNode;
    


    int tBucketSize = tBucket.count();
    for(register int i = pBucketI + 1; i < tBucketSize ; i++) {
	tScanNode = tBucket[i];
	tTestVert = &(tScanNode->mPolygon->mVertices[tScanNode->mVertexNum]);
	
	if( compareVerts(tTestVert, tFirstVert) ) {
	    tTestNormal = &(tScanNode->mPolygon->mNormal);
	    if( !mEdgeTest || tTestNormal->dot(*tHeadNormal) > mMinDot) {
		tNormal += *tTestNormal;
		tScanNode->mMarked = WORKING;
	    } 
	}
    }
    
    tNormal.normalize();
    
    for(i = pBucketI; i < tBucketSize; i++) {
	tScanNode = tBucket[i];
    	if (tScanNode->mMarked == WORKING) {
	    tScanNode->mPolygon->mNormals[tScanNode->mVertexNum] = tNormal;
	    tScanNode->mMarked = DONE;
	}
    }
}

    



void PolySmoother::setFuzzFraction(GFloat pFuzz) {
    mFuzz = pFuzz;
}

void PolySmoother::enableEdgePreservation(GFloat pMinDot) {
    mEdgeTest = true;
    mMinDot = pMinDot;
}

void PolySmoother::disableEdgePreservation(void) {
    mEdgeTest = false;
}



////
////
////    
void Polygon::computeNormal(void) {
    GVector3 tN(0,0,0);
    
    for (register int i = 0; i < mNumVerts; i++) {
	int j = (i+1) % mNumVerts;
	GVector3 tT1 = mVertices[i] - mVertices[j];
	GVector3 tT2 = mVertices[i] + mVertices[j];
	tN += GVector3(tT1.getY() * tT2.getZ(),
			    tT1.getZ() * tT2.getX(),
			    tT1.getX() * tT2.getY());
    }
    if(tN.sqdist() < gEpsilon2) 
	mValid = false;
    else {
	mNormal = tN.unitize();
	mValid = true;
    }
}







