//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_BEAM_H__
#define __GX_BEAM_H__


/// A plane of a beam
class GXBeamPlane {
/** This class implements cutting-planes , which are used to contruct
  polyhedron-beams. The overlapping-test w/ a bbox uses the shaft-culling
  idea of E. Haines
  */
  public:

    /// construct the plane w/ the given normal and distance
    GXBeamPlane(const GVector3& pNormal, GFloat pD) 
    : mNormal(pNormal), mD(pD) {}
    
    /// construct the plane w/ the given plane and a given point on it
    GXBeamPlane(const GVector3& pNormal, const GPoint3 pPoint) {
	mNormal = pNormal;
	
	mNearI[0] = (mNormal.getX() > 0) ? G_LO_X : G_HI_X;
	mNearI[1] = (mNormal.getY() > 0) ? G_LO_Y : G_HI_Y;
	mNearI[2] = (mNormal.getZ() > 0) ? G_LO_Z : G_HI_Z;

	mD = mNormal.dot(pPoint);
    }
    
    void setNearIndex(short  pIx, short pIy, short pIz) {
	mNearI[0] = pIx; mNearI[1] = pIy; mNearI[2] = pIz; 
    }

    void getFarIndex(short& rIx, short& rIy, short& rIz) {
	rIx = (mNearI[0]+1) % 2; 
	rIy = 2 + (mNearI[1]+1) % 2; 
	rIz = 4 + (mNearI[2]+1) % 2; 
    }
    
    void getNearIndex(short& rIx, short& rIy, short& rIz) {
	rIx = mNearI[0]; 
	rIy = mNearI[1]; 
	rIz = mNearI[2]; 
    }

    
    bool isOverlap(GFloat *pExts) {
	return (mNormal.dot(GVector3(pExts[mNearI[0]], 
				     pExts[mNearI[1]],
				     pExts[mNearI[2]])) <= mD );
    }

    bool isOverlap(const GVector3& pPoint) {
	return (mNormal.dot(pPoint) <= mD );
    }
    
    /// overlapping-test w/ the given bbox.
    bool isOverlap(const GBounds3& pB) {
	GVector3 tPoint((mNearI[0] == G_LO_X) ? pB.getMin().getX() : pB.getMax().getX(),
			(mNearI[1] == G_LO_Y) ? pB.getMin().getY() : pB.getMax().getY(),
			(mNearI[2] == G_LO_Z) ? pB.getMin().getZ() : pB.getMax().getZ());
	return (mNormal.dot(tPoint) <= mD );
    }


  private:
    GVector3 mNormal;
    GFloat mD;
    short mNearI[3];

    ///
  friend class GXBeam;
    
};



/// A polyhedron-beam
class GXBeam : public GXAccelCell{
/** This class is very usefull for applications of ray\'s coherence.
  In GX, it is used for 5D, Light/Vista-buffer, shadow-culling etc..
  */

  public:
    ///
    GXBeam(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList);
    virtual ~GXBeam(void);
    
    
    ///
    bool   intersectShading(GRay3& pRay, GInterSectionList& rInter);
    /// 
    bool   intersectShadow(const GRay3& pRay, GInterSectionList& rInter);
    
    /// return the dominant axis for this beam
    int getDominantAxis(void) const { return (mFace);}

    /// return the number of candidates inside this beam
    int getNumOfChildren(void) const { 
	if(mCandidatesList)
	    return (mCandidatesList->count());
	else
	    return (0);
    }
    
    /// reclassify the candidates-list
    void classifyCandidatesList(void);

    ///
    void getClippedCandidatesList(const GRay3& pRay, 
				  GXObjectList &rList) const;
     
    /// overlapping-test w/ a bbox
    bool isOverlap(const GBounds3& box) const;

    /// overlapping-test w/ a ray in 5D
    bool isInside(GFloat *p5D);

  protected:
    /// direction face
    int mFace;
    
    /// min. & max. extents of the beam
    GFloat mExtents[10];
    
    /// the 4 shaft-planes of the beam
    GXBeamPlane *mPlanes[4];
    
    /// candidates list
    GXObjectList *mCandidatesList;
    

    
    void (*_constructBeamPlanes)(GFloat *pExts, GXBeamPlane *rPlanes[4] );
    GFloat (*_getMinDist)(const GRay3& pRay, const GBounds3 &pB);
    bool (*_isBehindVoxel)(const GBounds3 &pVoxel, const GBounds3 &pBound);
    

    
};



#endif
