//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_LIGHT_H__
#define __GX_LIGHT_H__


#include "render/glight.h"
#include "model/gmodel.h"
#include "gx/gx_defs.h"
#include "gx/gx_obj.h"
#include "gx/gx_shade.h"
#include "gx/gx_sampl.h"
#include "gx/gx_vbuf.h"


/// Basis class for GX-lightsources
class GXLight {
    
 public:
    /// CTor
    GXLight(GFloat pC1=1, GFloat pC2=0, GFloat pC3=0)
    : mC1(pC1), mC2(pC2), mC3(pC3) { }
    
    /// preprocess the internal datas
    virtual void preprocess(GXObjectList *pObjLists[6]) {}
    
    /// shade the I-point
    virtual GColorRGB shade           (GXShadingParams& rSP)=0;
    
    /// compute the DIFFUSE-component
    virtual GColorRGB shadeDiffuse    (const GXShadingParams &pSP) {
	return (GColorRGB(0, 0, 0));
    }
    
    /// compute the SPECULAR-component
    virtual GColorRGB shadeSpecular   (const GXShadingParams &pSP) {
	return (GColorRGB(0, 0, 0));
    }
 
    /// set the coefficients of the distance-attenuation-func.
    void setDistAtten(GFloat pC1, GFloat pC2, GFloat pC3) {
	mC1 = pC1;
	mC2 = pC2;
	mC3 = pC3;
    }
    
 protected:
    ///
    GFloat findDistAtten(GFloat pT) {
	GFloat tRet = 1 / (mC1 + pT * (mC2 + mC3 * pT));
	return ((tRet < 1) ? tRet : 1);
    }
 private:
    // lights coefficients for distance-attenuation 
    GFloat mC1, mC2, mC3;              

};


/// Ambient lightsource
class GXAmbientLight: public GXLight, virtual public GLight {
  public:
    ///
    GXAmbientLight(void) 
    : GXLight(), GLight() {}
    ///
    GColorRGB  shade(GXShadingParams &rSP);
};


/// A class for point lightsources
class GXPointLight: public GXLight, virtual public GPointLight {
    
 public:
    ///
    GXPointLight(void);
    ///
    virtual ~GXPointLight(void);
    ///
    virtual void preprocess(GXObjectList *pObjLists[6]);
    ///
    GColorRGB shade           (GXShadingParams& rSP);
    ///
    GColorRGB shadeDiffuse    (const GXShadingParams &pSP);
    ///
    GColorRGB shadeSpecular   (const GXShadingParams &pSP);
 
 protected:
    /// compute the light from the I-Point to the given target
    GColorRGB getLight(const GPoint3& pTarget, GXShadingParams &rSP);

    /// the modified light-buffer
    GXViewBuffer *mLightBuffer;

    /// reserved for shadow caching 
    GXCache *mShadowCache;    
    
};


/// A class for rectangular lightsources
class GXRectangularLight: public GXPointLight {
  public:
    ///
    GXRectangularLight(const GPoint3& pPosition, const GPoint3& pLookAt, 
		       GFloat pWidth, GFloat pHeight,
		       int pXRes = 2,
		       int pYRes = 2,
		       int pMinSamples = GXOptions::MinSamples,
		       int pMaxSamples = GXOptions::MaxSamples,
		       int pNumCandidates = GXOptions::NumCandidates);
    
    virtual ~GXRectangularLight(void);
    ///
    void setPosition(const GPoint3& pPosition);
    
    ///
    void preprocess(GXObjectList *pObjLists[6]);
    ///
    GColorRGB shade           (GXShadingParams& rSP);
 
 private:
    void computeModelling(void);
    
    GVector3 mLookAt;
    GFloat mWidth, mHeight;
    int mXRes, mYRes;
    GModeling mModel;
    GX_PoissonDiskSampling mSampling;
};


/// A class for spherical lightsources
class GXSphericalLight: public GXPointLight {
   
 public:
    ///
    GXSphericalLight(const GPoint3& pPosition, 
		     GFloat pRad,
		     int pXRes = 2,
		     int pYRes = 2);
    ///
    virtual ~GXSphericalLight(void);
    ///
    void setPosition(const GPoint3& pPosition);
    ///
    void preprocess(GXObjectList *pObjLists[6]);
    ///
    GColorRGB shade           (GXShadingParams& rSP);
 
 private:
    GFloat mRad;
    int mXRes, mYRes;
};


extern GDLList<GXLight>* gGXLights;

#endif
