//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_OBJ_H__
#define __GX_OBJ_H__

#include "render/graytrac.h"


///  Abstract class for GX objects
class GXObject: public GRayTraceable {
/** This class contains the extended interface
  of an GX-object to communicate w/ the GX-raytracer
  */
 
  public:

    /// CTor: default
    GXObject(int pF = GRayTraceable::PRIMITIVE): GRayTraceable(pF) {}
    /// DTor: virtual!
    virtual ~GXObject(void) {}
    
    /// return the number of children of this object
    virtual int getNumOfChildren(void) const { return(0);}

    /// return the heuristicaly intersection cost of this object
    virtual GFloat getCost(void) const = 0;

    /// Overlapping-test w/ the given point
    virtual bool isOverlap(const GPoint3& point) const = 0;

    /// Overlapping-test w/ the given box
    virtual bool isOverlap(const GBounds3& box) const = 0;
}; 


/// Base class for GX primitives
class GXPrimitive: public GXObject {
  public:
    
    GXPrimitive(int pF = GRayTraceable::PRIMITIVE)
    : GXObject(pF)    {}

    virtual ~GXPrimitive(void) {}
    
    ///
    virtual GFloat getCost(void) const{ return(0); }
    ///
    virtual bool isOverlap(const GPoint3& point) const {
	return (getWorldBounds().isOverlap(point));
    };
    ///
    virtual bool isOverlap(const GBounds3&    box) const {
	return (getWorldBounds().isOverlap(box));
    };


  
  protected:
    
    /// compute the bbox in WC
    virtual const GBounds3& computeBounds(void) {
	setWorldBounds().setMin(GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
	setWorldBounds().setMax(GPoint3(MAXFLT, MAXFLT, MAXFLT));
	return (getWorldBounds());
    }

};


/// Base class for acceleration cells
class GXAccelCell {
/** Interface for all types of subdivision cell for an acceleration methode.  
 */
  
  public:
    
    /// intersect the given SHADING-ray w/ all candidates of this cell
    virtual bool   intersectShading(GRay3& rRay, GInterSectionList& rInter) = 0;

    /// intersect the given SHADOW-ray w/ all candidates of this cell
    virtual bool   intersectShadow(const GRay3&  rRay, GInterSectionList&rInter) = 0;
    
};


/// Base class for acceleration methodes 
class GXAccelMethode {
/** Interface for all acceleration methodes 
 */

  public:
       
    /// intersect the given SHADING-ray w/ the ROOT-node
    virtual bool   intersectShading(GRay3& rRay, GInterSectionList& rInter) = 0;

    /// intersect the given SHADOW-ray w/ the ROOT-node
    virtual bool   intersectShadow(/// the ray
				   GRay3 &rRay, 
				   /// ref. to the I-result
				   GInterSectionList &rInter, 
				   /// the last opaque object intersected by the ray
				   GRayTraceable *&rLastOpaque, 
				   /// the last cell intersected by the ray
				   GXAccelCell *&rLastCell) = 0;
    
};



extern const int gMaxRecursion;


/// Intersection cache
class GXCache {
/** Define an efficient data struture for I-caching
  at each level of the ray\'s tree.
  */
 
  public:
    ///
    GXCache(int pLevel = 0,
	    GRayTraceable *pLastObject = 0,
	    GXAccelCell *pLastVoxel = 0,
	    GXCache *pRefraction = 0,
	    GXCache *pReflection = 0)
    : mLastObject(pLastObject),
      mLastVoxel(pLastVoxel),
      mRefraction(pRefraction),
      mReflection(pReflection) 
    {
	if(pLevel < gMaxRecursion) { 
	    mRefraction = new  GXCache(pLevel+1);
	    mReflection = new  GXCache(pLevel+1);
	}
    }
    
    ~GXCache(void) {
	gkillPointer(mRefraction);
	gkillPointer(mReflection);
    }
    
    /// examine the actual cache-data for a ray
    GXCache* getNode(const GRay3& pRay) {
	int tMask = 0x01;
	register GXCache *tCache = this;
    
	for (register int i = 0; i < pRay.getLevel(); ++i) {
	    if (tMask & pRay.getPath()) 
		tCache = tCache->mRefraction;
	    else             
		tCache = tCache->mReflection;
	    tMask = tMask << 1; 
	}
	return (tCache);
    }

    /// last recently hited object
    GRayTraceable   *mLastObject; 
    
    /// last recently hited voxel
    GXAccelCell     *mLastVoxel;  

  private:
    /// pointer to the refract. path in next level 
    GXCache         *mRefraction; 

    /// pointer to the reflect. path in next level
    GXCache         *mReflection; 
};



#endif















