//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_SAMPLE_H__
#define __GX_SAMPLE_H__

#include <stdlib.h>

//////////////////////////////////////////////////////////////////////////////
// UTILS
//////////////////////////////////////////////////////////////////////////////
// return a random value between [0, 1]
inline GFloat Random_01(void) {
    return ((GFloat) rand() / (GFloat)RAND_MAX);
}

// return a random value between [-1, 1]
inline GFloat Random_11(void) {
    return((GFloat) rand() * 2.0 / (GFloat) RAND_MAX - 1.0);
}
// return a random value between [a, b]
inline GFloat Random_ab(GFloat a, GFloat b) {
    if(a > b) gswap(a, b);
    return ((GFloat)rand() * (b -a) / (GFloat)RAND_MAX  + a);
}





/// Base class for different sampling methodes.
class GX_Sampling {
/** Each derived class muss overload the virtual
  function generateSamples(), wich generates an array of sample points
  in the unit square.
  */

  public:
    ///
    GX_Sampling(int pNumSamples) {
	mNumSamples = pNumSamples;
	mSamples = new GVector2[mNumSamples];
    }
    ///
    virtual ~GX_Sampling(void) {
	gkillArray(mSamples);
    }
    
    /// return number of samples
    int getNumber(void) {
	return (mNumSamples);
    }

    /// return an array of sample\'s positions
    const GVector2& getSample(int pIndex) const {
	return (mSamples[pIndex]);
    }
    
    /// generate the sampling pattern
    virtual void generateSamples(void) = 0;
    
  private:
    int mNumSamples;

  protected:
    /// array of samples
    GVector2 *mSamples;
    
};


/// Sampling with Poisson-Disk distribution
class GX_PoissonDiskSampling : public GX_Sampling {
/** 
  This sampling methode is statistical optimized
  for adaptive supersampling, based on Mitchells work. The number of
  samples varies from mMinS to mMaxS. The contrast-threshold is 
  RGB(0.4, 0.3, 0.6)
  */

  public:
    ///
    GX_PoissonDiskSampling(int pMinS = 4, 
			   int pMaxS = 16, 
			   int pCandidateS = 10) 
      : GX_Sampling(pMaxS), 
    mMinS(pMinS), mMaxS(pMaxS), mCandidateS(pCandidateS) {}
    ///
    virtual ~GX_PoissonDiskSampling(void) {}
    /// return the min. number of samples
    int getMinS(void) const { return (mMinS); }
    /// return the max. number of samples
    int getMaxS(void) const { return (mMaxS); }

    /// generate the sample  with the largest minimum-distance
    void generateSamples(void) {
	mSamples[0] = GVector2(Random_01(), Random_01());
    
	for(register int i=1; i < getNumber(); i++) {
	    GFloat tMinDist = -MAXFLT;

	    for(register int j=0; j < mCandidateS; j++) {
		GVector2 tCandidate(Random_01(), Random_01());
		
		// compute the min. distance
		GFloat tTestMinDist = MAXFLT;
		for(register int k = 0; k < i; k++) {
		    GFloat _t = (tCandidate - mSamples[k]).sqdist();
		    if(tTestMinDist > _t) 
		      tTestMinDist = _t;
		}

		// update the maximum from min. distances
		if(tMinDist < tTestMinDist) {
		    tMinDist = tTestMinDist;
		    mSamples[i] = tCandidate;
		}
	    }
	}
    }


  private:
    int mMinS, mMaxS, mCandidateS;
    GFloat getMinDist(int pNr, const GVector2&);
};



/// Jittered Sampling
class GX_JitterSampling  : public GX_Sampling {
    
  public:
    ///
    GX_JitterSampling(int pW, int pH) 
      : GX_Sampling(pW*pH), mW(pW), mH(pH) {}
    ///
    virtual ~GX_JitterSampling(void) {}
    ///
    void generateSamples(void) {
	for(register int i = 0; i < mH; i++) 
	  for(register int j = 0; j < mW; j++) 
	  mSamples[i*mH+j] = GPoint2(Random_ab((GFloat)j/mW, (GFloat)(j+1)/mW), 
				     Random_ab((GFloat)i/mH, (GFloat)(i+1)/mH)); 
    }
  
  private:
    int mW, mH;
};


/// Half-Jittered Sampling
class GX_HalfJitterSampling  : public GX_Sampling {
  public:
    ///
    GX_HalfJitterSampling(int pW, int pH) 
      : GX_Sampling(pW*pH), mW(pW), mH(pH) {}
    ///
    virtual ~GX_HalfJitterSampling(void) {}
    ///
    void generateSamples(void) {
	for(register int i = 0; i < mH; i++) 
	  for(register int j = 0; j < mW; j++) 
	  mSamples[i*mH+j] = GPoint2(Random_ab( ((GFloat)j+0.25)/mW, ((GFloat)j+0.75)/mW), 
				     Random_ab( ((GFloat)i+0.25)/mH, ((GFloat)i+0.75)/mH)); 
    }
    

  private:
    int mW, mH;
};


/// N-rooks Sampling 
class GX_NRooksSampling  : public GX_Sampling {
   
  public:
    ////
    GX_NRooksSampling(int pW, int pH) 
      : GX_Sampling(pW*pH), mW(pW), mH(pH) {}
    ///
    virtual ~GX_NRooksSampling(void) {}
    ///
    void generateSamples(void) {
	GFloat tWidth;
	register int i, j;

	tWidth = 1.0 / (GFloat)(mW * mH);

	for (i = 0; i < mH; i++) {
	    for (j = 0; j < mW; j++) {
		mSamples[i * mW + j].setX((GFloat)i * mW*tWidth + (GFloat)j*tWidth + Random_ab(0, tWidth));
		mSamples[i * mW + j].setY((GFloat)j * mH*tWidth + (GFloat)i*tWidth + Random_ab(0, tWidth));
	    }
	}

	for (i = 0; i < mH; i++) {
	    for (j = 0; j < mW; j++) {
		int k = (int) Random_ab(j, mW - 1);
		GFloat t = mSamples[i * mW + j].getX();
		mSamples[i * mW + j].setX(mSamples[i * mW + k].getX());
		mSamples[i * mW + k].setX(t);
	    }
	}

	for (i = 0; i < mW; i++) {
	    for (j = 0; j < mH; j++) {
		int k = (int) Random_ab(j, mH - 1);
		GFloat t = mSamples[j * mW + i].getY();
		mSamples[j * mW + i].setY(mSamples[k * mW + i].getY());
		mSamples[k * mW + i].setY(t);
	    }
	}
	
    }
    

  private:
    int mW, mH;

};



#endif
