//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_EVENT_H
#define G_EVENT_H

#include <string.h>
#include "input/gtime.h"

//@ManMemo: Basic event type.
class GEvent {
/*DOC:
Basic class of classes representing events.
An event has a unique type and a creation time.
*/
  public:
    //@ManMemo: Inquire class identity.
    virtual const char *getClass(void) const = 0;

    //@ManMemo: Check class identity. 
    virtual bool isA(const char* pClass) {
	if (pClass == getClass()) return(true);
	else return(!strcmp(pClass, getClass()));
    }

    //@ManMemo: Inquire event's creation time. 
    const GTime& getTime(void) const { return(mTime); }
  private:
    GTime mTime;
};

//@ManMemo: Base type for mouse events.
class GMouseEvent: public GEvent {
/*@DOC:
This event type represents mouse actions and the state of some special keys.
(0, 0) is the lower left.
*/
  public:
    //@ManMemo: Default initialization of members.
    GMouseEvent(void): left(false), middle(false), right(false),
    shift(false), ctrl(false), alt(false),
    x(0), y(0), w(0), h(0) {}
  public:
    //@ManMemo: State of the mouse buttons (true if pressed).
    bool left, middle, right;

    //@ManMemo: State of the special keys (true if pressed).
    bool shift, ctrl, alt;

    //@ManMemo: Mouse pointer coordinates and pixmap dimensions.
    int x, y, w, h;
};

extern const char* gMouseButtonEventID;

//@ManMemo: Mouse button event.
class GMouseButtonEvent: public GMouseEvent {
/*@DOC:
This event type represents mouse press/release actions and the state of some
special keys.
*/
  public:
    virtual const char* getClass(void) const { return(gMouseButtonEventID); }
};

extern const char* gMouseMotionEventID;

//@ManMemo: Mouse motion event.
class GMouseMotionEvent: public GMouseEvent {
/*@DOC:
This event type represents mouse motions and the state of some
special keys.
*/
  public:
    virtual const char* getClass(void) const { return(gMouseMotionEventID); }
};

extern const char* gSpaceballEventID;

//@ManMemo: Incremental space ball event.
class GSpaceballEvent: public GEvent {
  public:
    //@ManMemo: Default initialization of members.
    GSpaceballEvent(void): transX(0), transY(0), transZ(0),
    twistX(0), twistY(0), twistZ(0),
	b1(0), b2(0), b3(0), b4(0), b5(0), b6(0), b7(0), b8(0), 
	pick(0) {}
    const char* getClass(void) const { return(gSpaceballEventID); }
  public:
    //@ManMemo: Translation values of the spaceball.
    short transX, transY, transZ;

    //@ManMemo: Rotation values of the spaceball.
    short twistX, twistY, twistZ;

    //@ManMemo: Button states.
    bool b1, b2, b3, b4, b5, b6, b7, b8;

    //@ManMemo: Pickbutton state.
    bool pick;

    //@ManMemo: Number of periods of 0.25 ms since sending the last non-0 set of Spaceball data.
    short period;
};

class GPixmapDisplay;

//@ManMemo: Basic pixmap event.
class GPixmapEvent: public GEvent {
  protected:
    GPixmapEvent(GPixmapDisplay* tPixmap): mPixmap(tPixmap) {}
  public:
    GPixmapDisplay* getPixmap(void) { return(mPixmap); }
  private:
    GPixmapDisplay* mPixmap; 
};

extern const char* gUpdatePixmapEventID;

//@ManMemo: Update event for pixmaps.
class GUpdatePixmapEvent: public GPixmapEvent {
  public:
    //@ManMemo: Default initialization of members.
    GUpdatePixmapEvent(GPixmapDisplay* tPixmap): GPixmapEvent(tPixmap), mPartial(false) {}
    virtual const char* getClass(void) const { return(gUpdatePixmapEventID); }
  public:
    //@ManMemo: True if the coordinates are valid.  
    bool mPartial;

    //@ManMemo: (x0,y0) left upper corner of the area that must be updated, (x1,y1) right lower corner .
    int x0, y0, x1, y1;
};

extern const char* gResizePixmapEventID;

//@ManMemo: Resize event for pixmaps.
class GResizePixmapEvent: public GPixmapEvent {
  public:
    GResizePixmapEvent(GPixmapDisplay* tPixmap): GPixmapEvent(tPixmap) {}
    virtual const char* getClass(void) const { return(gResizePixmapEventID); }
};

#endif

