//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_MODELING_H
#define G_MODELING_H

#include "util/gmatrix.h"
#include "gattrib.h"

//@ManMemo: Modeling entity.
class GModeling {
/*@DOC:
Encapsulation of a spatial modeling property. Automatic individualization
is provided.
*/
  public:
    GModeling(void);
    virtual ~GModeling(void) {
	if (local) delete local;
	if (world) delete world;
    }
    
    GModeling& operator=(const GModeling& pModel) {
	individualize();
	*local = *pModel.local;
	*world = *pModel.world;
	setFatherModeling(pModel.getFatherModeling());
	return(*this);
    }
    
    //// modeling interface:
    // accumulationg rotation:
    void rotate(const GVector3&);
    // accumulating scaling:
    void scale(const GVector3&);
    // accumulating translation:
    void translate(const GVector3&);
    // replace translation terms:
    void setTranslation(const GPoint3&);

    void setMatrix(const GMatrix4&);

    const GMatrix4& getMatrix(void) const {
	if (local) return(local->mat);
	else return(gUnityMatrix4);
    }
    
    void reset(void);

    // conversion from/to modeling coordinates: 
    GVector3 mc2wc(const GVector3&) const;
    GVector3 wc2mc(const GVector3&) const;

    const GMatrix4& getWorldMatrix(void) const;
    const GMatrix4& getInvWorldMatrix(void) const;

    const GMatrix4& getGlobalMatrix(void) const;
    const GMatrix4& getInvGlobalMatrix(void) const;

    // access to inherited modeling:
    void setFatherModeling(const GModeling&); 
    const GModeling& getFatherModeling(void) const {
	return(*pFatherModeling);
    }
  private:
    void concat(void);
    void individualize(void);
    const GModeling* pFatherModeling;
    GMatrix4Pair* local;
    GMatrix4Pair* world;
};

extern const GModeling gNoModeling;

inline GModeling::GModeling(void): pFatherModeling(&gNoModeling), local(0), world(0) {}

//@ManMemo: A spatial modeling attribute.
class GModelingAttribute: public GAttribute {
  public:
    const GModeling& getModeling(void) const;

    // conversion from/to modeling coordinates: 
    GVector3 mc2wc(const GVector3& v) const {
	return(getModeling().mc2wc(v));
    }
    GVector3 wc2mc(const GVector3& v) const {
	return(getModeling().wc2mc(v));
    }

    const GMatrix4& getWorldMatrix(void) const {
	return(getModeling().getWorldMatrix());
    }
    const GMatrix4& getInvWorldMatrix(void) const {
	return(getModeling().getInvWorldMatrix());
    }

    void reference(const GModelingAttribute&);

    GModelingAttribute(void);
  protected:
    GModelingAttribute(const GModeling&);

    void setModeling(const GModeling&);
    virtual void modelingChanged(void) {}

    // Return private modeling.
    // If there isn't one, individualize it.
    GModeling& getPrivateModeling(void);
  private:
    GModeling* mModeling;
    static const GModelingAttribute attribute;
};

inline GModelingAttribute::GModelingAttribute(const GModeling& pModel): mModeling(0), GAttribute(attribute) {
    setModeling(pModel);
}

#endif
