//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_IMAGE_H
#define G_IMAGE_H

#include <stdio.h>

#include "render/grealiz.h"
#include "util/gcolor.h"
#include "util/gstring.h"
#include "goutput.h"

//@ManMemo: Abstract base class representing file based images.
class GImage: public GRealization {
  protected:
    //@ManMemo: Image initialization.
    GImage(const char* pFname): mBase(0), mXSize(0), mYSize(0), mFname(pFname) {} 
  public:
    virtual ~GImage(void) { 
	if (mBase) delete mBase;
    }

    //@ManMemo: Generic read function.
    virtual bool read(void) = 0;
    // Returns true if successful.

    //@ManMemo: Generic write function.
    virtual bool write(void) = 0;
    // Returns true if successful.

    //@ManMemo: Set filename.
    void setFilename(const char* f) {
	assert(f);
	mFname = f; 
    }

    //@ManMemo: Get file name.
    const char* getFilename(void) const { 
	return(mFname);
    }

    //@ManMemo: Get file name.
    const char* getName(void) const { 
	return(getFilename());
    }

    //@ManMemo: Set image width.
    void setWidth(int pX) {
	if (pX < 1) 
	    gOutput->error("Cannot except an image dimension less than one.");
	else {
	    mXSize = pX;
	    newGeometry(); 
	}
    }
    //@ManMemo: Get image width.
    int getWidth(void) const {
	return(mXSize); 
    }
    
    //@ManMemo: Set image height.
    void setHeight(int pY) {
	if (pY < 1) 
	  gOutput->error("Cannot except an image dimension less than one.");
	else {
	    mYSize = pY;
	    newGeometry(); 
	}
    }

    //@ManMemo: Get image height.
    int getHeight(void) const {
	return(mYSize);
    }

    //@ManMemo: Clear image with specified color.
    virtual void clear(const GColorRGB&) {
	gOutput->error("GImage::clear() called, which is an empty implementation.");
    }

    //@ManMemo: Write color into specified pixel.
    virtual void setColor(int, int, const GColorRGB&) = 0;

    //@ManMemo: Read color from specified pixel.
    virtual GColorRGB getColor(int, int) const = 0;

     //@ManMemo: Map coordinates in range 0 to 1 to pixel dimensions and return interpolated color.
    GColorRGB getIntColor(GFloat, GFloat) const;

    // @ManMemo: Return pre-defined image file extension.
    virtual const char* getExtension(void) const = 0;
    
    void putPixel(int pX, int pY, const GColorRGB& pColr) {
	setColor(pX, pY, pColr);
    }
    
    GColorRGB getPixel(int pX, int pY) {
	return(getColor(pX, pY));
    }
    
  protected:
    // returns true if indices are valid:
    bool checkIndices(int, int) const;
    
    virtual void newGeometry(void) = 0;
    // rearrange place for pixmap data
    
    unsigned long* mBase;
    // the pointer to a data array
  protected:
    int mXSize, mYSize;
    GString mFname;
};

#define GSGI_IMAGIC 	                0732
#define GSGI_TYPEMASK	         	0xff00
#define GSGI_BPPMASK			0x00ff
#define GSGI_ITYPE_VERBATIM		0x0000
#define GSGI_ITYPE_RLE		0x0100
#define GSGI_ISRLE(type)		(((type) & 0xff00) == GSGI_ITYPE_RLE)
#define GSGI_ISVERBATIM(type)	(((type) & 0xff00) == GSGI_ITYPE_VERBATIM)
#define GSGI_BPP(type)		((type) & GSGI_BPPMASK)
#define GSGI_RLE(bpp)		(GSGI_ITYPE_RLE | (bpp))
#define GSGI_VERBATIM(bpp)		(GSGI_ITYPE_VERBATIM | (bpp))

#define GSGI_RINTLUM (79)
#define GSGI_GINTLUM (156)
#define GSGI_BINTLUM (21)

#define GSGI_ILUM(r,g,b)     ((int)(GSGI_RINTLUM*(r)+GSGI_GINTLUM*(g)+GSGI_BINTLUM*(b))>>8)

#define GSGI_OFFSET_R	3	// this is byte order dependent 
#define GSGI_OFFSET_G	2
#define GSGI_OFFSET_B	1
#define GSGI_OFFSET_A	0
#define GSGI_CHANOFFSET(z)	(3-(z))	// this is byte order dependent 
#define GSGI_TAGLEN	(5)

typedef struct {
    unsigned short	imagic;		// stuff saved on disk . . 
    unsigned short 	type;
    unsigned short 	dim;
    unsigned short 	xsize;
    unsigned short 	ysize;
    unsigned short 	zsize;
    unsigned long 	min;
    unsigned long 	max;
    unsigned long	wastebytes;	
    char 		name[80];
    unsigned long	colormap;
    
    long 		file;		// stuff used in core only 
    unsigned short 	flags;
    short		dorev;
    short		x;
    short		y;
    short		z;
    short		cnt;
    unsigned short	*ptr;
    unsigned short	*base;
    unsigned short	*tmpbuf;
    unsigned long	offset;
    unsigned long	rleend;		// for rle images 
    unsigned long	*rowstart;	// for rle images 
    long		*rowsize;	// for rle images 
} GSGIImageStruct;

//@ManMemo: Encapsulation of SGI's RGB(A) image file format.
class GSGIImage: public GImage {
/*@Doc:
Encapsulation of SGI's RGB(A) image file format. Implementation
bases on code by Paul Haeberli, used with friendly permission.
*/
  public:
    GSGIImage(const char*);
    virtual ~GSGIImage(void) {};

    virtual bool read(void); 
    virtual bool write(void);

    virtual void setColor(int, int, const GColorRGB&);
    virtual GColorRGB getColor(int, int) const; 
    virtual const char* getExtension() const { 
	return("rgb"); 
    }
  private: 
    virtual void newGeometry(void) {
	if (mBase) delete mBase;
	mBase = new unsigned long [mXSize * mYSize + GSGI_TAGLEN];
    }
    int mZSize;
    //  zsize specifies what kind of image file to
    // write out.  if zsize is 1, the luminance of the pixels are
    // calculated, and a sinlge channel black and white image is saved.
    // If zsize is 3, an RGB image file is saved.  If zsize is 4, an
    // RGBA image file is saved.
    bool sizeOfImage(void);
    // get x,y of the image file, returns 1 if succesful
    // some utility statics:
    static void interleaveRow(unsigned char *lptr, unsigned char *cptr, int z, int n) {
	lptr += z;
	while(n--) {
	    *lptr = *cptr++;
	    lptr += 4;
	}
    }
    static void lumRow(unsigned char*, unsigned char*, int n);
    static int compressRow(unsigned char*, unsigned char*, int, int);
    static void readTab(FILE*, unsigned long*, int);
    static int writeTab(FILE*, unsigned long*, int);
    static void copyBw(long*, int);
    static void setAlpha(unsigned char*, int);
    static void expandRow(unsigned char*, unsigned char*, int);
    // byte order independent read/write of shorts and longs:
    static int putLong(FILE*, unsigned long);
    static void putShort(FILE*, unsigned short);
    static unsigned long getLong(FILE*);
    static unsigned short getShort(FILE*);
    // extract image data from core dumps:
    static void addLongImgTag(unsigned long *, int, int);

    void readHeader(FILE*);
    int writeHeader(FILE*);
  private:
    GSGIImageStruct image;
};

typedef struct {
    unsigned char ident_len, cmap_type, image_type;
    unsigned short int cmap_origin, cmap_len;
    unsigned char centry_size;
    unsigned short int image_xorg, image_yorg, image_width, image_height;
    unsigned char pixel_size, image_discr;
} GTargaImageHeader; 

typedef struct {
    unsigned char b, g, r;
    // Note b - g - r order.
} GTargaRGB;
  
//@ManMemo: Encapsulation of Targe file format.
class GTargaImage: public GImage {
  public:
    GTargaImage(const char*);
    virtual ~GTargaImage(void) {};
    virtual bool read(void); 
    virtual bool write(void);

    virtual void setColor(int, int, const GColorRGB&);
    virtual GColorRGB getColor(int, int) const; 
    virtual const char* getExtension(void) const { 
	return("tga"); 
    }

  private:
    int lh;             
    virtual void newGeometry(void) {
	if (mBase) delete mBase;
	mBase = new unsigned long [mXSize * mYSize];
    }
    int compressRow(GTargaRGB*, unsigned char*);
    bool sizeOfImage(void); 
    // get x,y of the image file, returns 1 if succesful
    // byte order independent read/write of shorts:
    void putShort(FILE*, unsigned short);
    unsigned short getShort(FILE*);

    GTargaImageHeader Header;
    void readHeader(FILE*);
    int writeHeader(FILE*);

};

typedef struct {
    unsigned int imagic, xsize, ysize;
} GRGBAImageHeader;

#define GRGBA_IMAGIC 0730

//@ManMemo: RGBA Image file encapsulation.
class GRGBAImage: public GImage {
  public:
    GRGBAImage(const char* pFname ): GImage(pFname) {}
    virtual ~GRGBAImage(void) {};
    virtual bool read(void);
    virtual bool write(void);

    virtual void setColor(int, int, const GColorRGB&);
    virtual GColorRGB getColor(int, int) const;
    virtual const char* getExtension(void) const {
	return("rgb"); 
    }
  private:
    GRGBAImageHeader image;
    static int putLong(int ,unsigned long);
    static int putShort(int, unsigned short);
    static unsigned long getLong(int);
    static unsigned short getShort(int);
    void readHeader(int);
    void writeHeader(int);

    virtual void newGeometry(void) {}
};

#endif
