//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_RAY_CAMERA_H
#define G_RAY_CAMERA_H

#ifdef GWINDOWS
#include <windows.h>
#else
#include <stdlib.h> 
// stdlib for RAND_MAX
#endif

#include "render/graytrac.h"
#include "render/grayligh.h"

#ifndef RAND_MAX
#define RAND_MAX 32767
#endif

#undef G_RAY_DEBUG
// set this for debugging the ray-tracing

class GRayInterSector {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An abstract intersector
    // class to be used for ray light sources.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual void intersect(const GRay3&, GInterSectionList&) = 0;
    // Intersect the related scene with a ray.
    // To be implemented in subclasses.
};

extern const int gMaxRecursion;
extern const GFloat gMinWeight;

//// some ray-tracing utility functions:
// (These could be (static) members of the ray-tracer class. However
// because of the "templatization" of the ray-tracer base class they are 
// not members.)

GVector3 gSpecularDirection(const GVector3& I, const GVector3& N);
// Compute specular direction from incident direction I and normal N.
// All vectors unitized.
    
int gTransmissionDirection(const GSurface* M1, const GSurface* M2, const GVector3& I, const GVector3& N, GVector3& T);
// Compute transmission direction T from incident direction I, normal N, 
//  going from medium M1 to M2, with refraction governed by the relative index
// of refraction according to Snell's law:
//     n1 * sin O1 = n2 * sin O2
// If there is total internal reflection, return 0, else set T and return 1.
// All vectors unitized.

void gGetSurfaces(const GSurface& SURFACE, int TRACE_CNT, bool ENTER, const GSurface*& ENTERING_SURFACE, 
const GSurface*& EXITING_SURFACE);
// Get surfaces in dependence of trace counter, object entering or leaving, etc.

template<class P, class R>
class GRayTracer: public GGRenderer<P, R>, public GRayInterSector {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Generic ray-tracer class.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  protected:
    GRayTracer(void): mBackgr(gBlack), mJitter(0.3), mSamples(3), mLineNumbers(false) {}

    virtual GRay3 createRay(const GPoint2&, GFloat pAspect = 1.0) = 0;
  public:
    virtual void intersect(const GRay3& pRay, GInterSectionList& pList) {
	GRayTracingIterator<P> tIter(pRay, pList);
	assert(mAggregate);
	mAggregate->doWithElements(tIter);
    }
    
    inline virtual void trace(int LEVEL, GFloat WEIGHT, const GRay3& WC_RAY, GColorRGB& COLOR);
    // trace ray into scene

    inline virtual void shade(int LEVEL, GFloat WEIGHT, const GVector3& PT, const GVector3& NO, const GVector3& I, const GInterSectionList& INTER, GColorRGB& COLOR);
    // Shade a surface point (recursively if necessary), stop recursing when level > gMaxRecursion
    // or when weight < gMinWeight, WEIGTH is the cumulative weight, PT is surface point, 
    // NO is normal vector, I is incident ray, INTER is the intersection list
  public:
    void setBackground(const GColorRGB& pBackgr) {
	mBackgr = pBackgr;
    }
    const GColorRGB& getBackground(void) const {
	return(mBackgr);
    }
    void setJitter(GFloat pJitter) {
	mJitter = pJitter;
    }
    void setSamples(int pSamples) {
	mSamples = pSamples;
    }

    virtual void initRealization(R& r) {
	GGRenderer<P, R>::initRealization(r);
	r.clear(gBlack);
    }
    inline virtual void rendering(const GAggregate<P>&);

    void writeLineNumbers(void) {
	mLineNumbers = true;
    }
  protected:
    int mTraceCnt;
  private:
    GColorRGB mBackgr;
    GFloat mJitter;
    int mSamples; 
    // Number of samples per pixel in each direction
    bool mLineNumbers;
};

template<class P, class R>
inline void GRayTracer<P, R>::rendering(const GAggregate<P>& a) { 
    GGRenderer<P, R>::rendering(a);
    int pw = mRealization->getWidth();
    int ph = mRealization->getHeight();
    GFloat wc = 1.0/(pw - 1);
    GFloat hc = 1.0/(ph - 1);
    GColorRGB tColr, tPixColr;
    GFloat tJi;
    const GFloat tCo = mJitter/RAND_MAX;
    const GFloat tS1 = 1.0 /(GFloat)mSamples;
    const GFloat tS2 = tS1 * tS1;

    for (int j = ph - 1; j >= 0; j--) {
	for (int i = 0; i < pw; i++) {
	    tPixColr = gBlack;
	    for (int sj = 0; sj < mSamples; sj++) {
		tJi = (sj + rand() * tCo) * tS1;
		GFloat yy = hc * ((GFloat)j + tJi);
		for (int si = 0; si < mSamples; si++) {
		    tJi = (si + rand() * tCo) * tS1;
		    GFloat xx = wc * ((GFloat)i + tJi);
		    trace(0, 1.0, createRay(GPoint2(xx, yy), (GFloat)pw/(GFloat)ph), tColr);
		    tPixColr += tColr;
		}
	    }
	    mRealization->putPixel(i, j, tPixColr * tS2);
	}
	if (mLineNumbers) cout << ph - j << " of " << ph << " lines traced." << endl;
    }
}

template<class P, class R>
inline void GRayTracer<P, R>::trace(int pLevel, GFloat pWeight, const GRay3& pRay, GColorRGB& pColor) {
    if (!pLevel) mTraceCnt = 0;

    GOneInterSectionList tList; 
    GRayTracingIterator<P> tIter(pRay, tList);
    mAggregate->doWithElements(tIter);
    if (tList.getNumber()) {
	const GInterSection& tInter = tList.getFirst();
	GRayTraceable& tPrim = tInter.getPrimitive();
	GPoint3 tPoint = pRay.compute(tInter.getT()); 
	GVector3 tNormal = tPrim.getNormal(tPoint);
	shade(pLevel, pWeight, tPoint, tNormal, pRay.getVector(), tList, pColor);
    } 
    else pColor = getBackground();
    
    // clip color to [0,1]:
    pColor.correct();
}

template<class P, class R>
inline void GRayTracer<P, R>::shade(int pLevel, GFloat pWeight, const GPoint3& pPoint, const GVector3& pNormal, 
				    const GVector3& pView, const GInterSectionList& pList, GColorRGB& pColor) {
    const GInterSection& tInter1 = pList.getFirst();
    GSurface tSurface1 = tInter1.getPrimitive().getSurface(pPoint);
    bool tEnter = tInter1.enter();

    pColor = gBlack;

    GRay3 tRay;
    tRay.setPoint(pPoint);
    tRay.setVector(gSpecularDirection(pView, pNormal));

    // if we are inside an object than invert the normal
    // to get the effect of a lamp:
    GVector3 n2 = pNormal;
    if (!tEnter && (pView.dot(n2) > 0)) n2 = n2.negate();

    // computing the light contributed
    // by the primary light sources:
    GRayLightShadeIterator<GRayLight> tIter(pColor, pPoint, n2, tRay.getVector(), tSurface1, self);
    gRayLights->doWithElements(tIter);

    GVector3 n = pNormal;
    if (pView.dot(n) > 0) n = n.negate();
    // flip normal if necessary
    
    // if we're not too deep then recurse:
    if (pLevel + 1 < gMaxRecursion) {
	// recurse on specular reflection ray if significant:  
	if (tEnter && tSurface1.getSpecular().abs() * pWeight > gMinWeight) {
	    GColorRGB tCol(gBlack);
#ifdef G_RAY_DEBUG
	    cout << "specular: " << tRay << endl;
#endif
	    trace(pLevel + 1, tSurface1.getSpecular().abs() * pWeight, tRay, tCol);
	    pColor = pColor + tCol * tSurface1.getSpecular();
	}

	// recurse on transmission ray if significant:
	if (tSurface1.getTransparency() * pWeight > gMinWeight ) {
	    const GSurface* sEnter;
	    const GSurface* sExit;
	    gGetSurfaces(tSurface1, mTraceCnt, tEnter, sEnter, sExit);
	    GVector3 tTransDir;
	    if (gTransmissionDirection(sExit, sEnter, pView, n, tTransDir)) {
		tRay.setVector(tTransDir);
#ifdef G_RAY_DEBUG
		cout << "transmission: " << tRay << endl;
#endif
		if (tEnter) mTraceCnt++;
		else mTraceCnt--;
		GColorRGB tCol(gBlack);  
		trace(pLevel + 1, tSurface1.getTransparency() * pWeight, tRay, tCol);
		pColor = pColor + tCol * tSurface1.getTransparency();
	    }
	}
    }
}

template<class P, class R>
class GLookatRayTracer: public GRayTracer<P, R>, public GLookatCamera {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A ray-tracer basing on an OpenGL
    // camera model.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual GRay3 inverseTransform(const GPoint2& pPoint, GFloat pAspect) const {
	return(GLookatCamera::inverseTransform(pPoint, pAspect));
    }
  protected:
    virtual GRay3 createRay(const GPoint2&, GFloat pAspect = 1.0);
};

template<class P, class R>
inline GRay3 GLookatRayTracer<P, R>::createRay(const GPoint2& pPoint, GFloat pAspect) {
    GRay3 tRay = inverseTransform(pPoint, pAspect);

    // TO_DO:
    //    tRay.setTmax();    
    //    tRay.setTmin();
    
    return(tRay);
}

#endif

