//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_RAY_TRACING
#define G_RAY_TRACING

#include "util/gmatrix.h"
#include "util/gsurface.h"
#include "util/gtopolgy.h"
#include "util/gheap.h"
#include "render/grender.h"

class GRayTraceable;
extern const GSurface gDummySurface;

//@ManMemo: A data type for holding material properties and other data of an intersection point.
class GInterSection {
  public:
    GInterSection(void): mPrim(0), mT(0), mEnter(false) {}
    GInterSection(GRayTraceable& pPrim, GFloat pT, bool pEnter): mPrim(&pPrim), mT(pT), mEnter(pEnter) {}

    GRayTraceable& getPrimitive(void) const { 
	return(*mPrim); 
    }
    GFloat getT(void) const { 
	return(mT); 
    }
    bool enter(void) const {
	return(mEnter);
    }

    GInterSection& operator=(const GInterSection& pInter) {
	mPrim = pInter.mPrim;
	mT = pInter.mT;
	mEnter = pInter.mEnter;
	return(*this);
    }

    // compare two intersections:
    bool operator < (const GInterSection&) const;
  
  private:
    // the hitted primitive:
    GRayTraceable* mPrim;

    // the distance from ray origin to intersection point:
    GFloat mT; 
    // true if entering the primitive, false else
    bool mEnter; 
};

//@ManMemo: An abstract aggregate storing  intersection points.
class GInterSectionList {
  protected:
    GInterSectionList(void): mTransmission(1.0) {}
  public:
    // return first intersection:
    virtual const GInterSection& getFirst(void) const = 0;

    // add intersection:
    virtual void add(const GInterSection&) = 0;

    virtual int getNumber(void) const = 0;
    // METHOD: Return number of intersections.

    GFloat getTransmission(void) const {
	return(mTransmission);
    }
    void updateTransmission(GFloat pTrans) {
	mTransmission = mTransmission * pTrans;
    }
    

  protected:
    GFloat mTransmission;
    // A scalar in the range 0 to 1. 1 means
    // total transmission, 0 stands for no transmission.
    // The variable contains the lowest transmission found.
    // NOTE: This not a correct but a somehow useful way
    // to let shine light thru unopaque primitives.
};

//@ManMemo: An intersection list that stores just the nearest point. 
class GOneInterSectionList: public GInterSectionList {
  public:
    GOneInterSectionList(void): mHit(false) {}

    // return first intersection:
    virtual const GInterSection& getFirst(void) const {
	return(mInter);
    }

    // add intersection:
    virtual void add(const GInterSection& s);

    virtual int getNumber(void) const {
	return(mHit ? 1 : 0);
    }
    // return one or zero
  private:
    bool mHit;
    GInterSection mInter;
};

//@ManMemo: An intersection list storing all valid intersections in descendant-order.
class GAllInterSectionList: public GInterSectionList, public GHeap<GInterSection> {
  public:
    GAllInterSectionList(void) 
      : GInterSectionList(), GHeap<GInterSection>() {}

    // return first intersection:
    const GInterSection& getFirst(void) const {
	return(GHeap<GInterSection>::head());
    }

    // add intersection:
    void add(const GInterSection& pInter);
 
    int getNumber(void) const {
	return(GHeap<GInterSection>::count());
    }

    void reset(void) {
	GHeap<GInterSection>::reset();
	mTransmission = 1.0;
    }  

};

template <class P>
//@ManMemo: Basic iterator for ray-tracing.
class GRayTracingIterator: public GIterator<P> {
  public:
    GRayTracingIterator(const GRay3& pRay, GInterSectionList& pList): mRay(pRay), mList(pList), mHit(false) {}
    virtual bool exec(P& p) { 
	if (p.intersect(mRay, mList)) mHit = true;
	return(false);
    }
    void reset(void) {
	mHit = false;
    }
    // METHOD: Reset hit flag.
    bool getHit(void) const { return(mHit); }
    // METHOD: Return true if there was a hit.
  private:
    const GRay3& mRay;
    GInterSectionList& mList;
    bool mHit;
};

//@ManMemo: An axis aligned box in world coordinates.
class GBounds3 {
    friend ostream& operator<<(ostream&, const GBounds3&);
  public:
    GBounds3(void) {}
    GBounds3(const GPoint3& pMin, const GPoint3& pMax)
      : mMin(pMin), mMax(pMax) {}
   
    GBounds3(const GBounds3 &pBox) {
	mMin = pBox.mMin;
	mMax = pBox.mMax;
    }
   
    // INTERSECTION
    friend GBounds3 operator &(const GBounds3& pA, const GBounds3& pB) {
	if (pA.isOverlap(pB)) 
	  return(GBounds3(Vec3Max(pA.mMin, pB.mMin), 
			  Vec3Min(pA.mMax, pB.mMax)));
	else	       
	  return(GBounds3(GPoint3(0, 0, 0), GPoint3(0, 0, 0)));
    }

    // UNION
    friend GBounds3 operator |(const GBounds3& pA, const GBounds3& pB) {
	return (GBounds3(Vec3Min(pA.mMin, pB.mMin), 
			 Vec3Max(pA.mMax, pB.mMax)));
    }

    // INTERFACE
    void min(const GVector3& pMin) { mMin = pMin; } 
    GVector3& min(void) { return(mMin); }
    const GVector3& min(void) const { return(mMin); }

    void max(const GVector3& pMax) { mMax = pMax; } 
    GVector3& max(void) { return(mMax); }
    const GVector3& max(void) const { return(mMax); }
 
    // set/get minimum:
    void setMin(const GVector3& pMin) { mMin = pMin; }
    const GVector3& getMin(void) const { return(mMin); }

    // set/get maximum:
    void setMax(const GVector3& pMax) { mMax = pMax; }
    const GVector3& getMax(void) const { return(mMax); }

    
    const GBounds3& operator+=(const GPoint3& pPt) {
	mMin.setMin(pPt);
	mMax.setMax(pPt);
	return(self);
    }

    const GBounds3& operator+=(const GBounds3& pBox) {
	mMin.setMin( pBox.mMin);
	mMax.setMax( pBox.mMax);
	return(self);
    }

    GFloat getSurfaceArea() const  {
	GPoint3 tExt = mMax - mMin;
	return (tExt.getX() * (tExt.getY() + tExt.getZ())+ 
		tExt.getY() * tExt.getZ());
    }

    bool isOverlap(const GBounds3& y) const {
	return !(mMax.getX() < y.mMin.getX() || mMin.getX() > y.mMax.getX() ||
		 mMax.getY() < y.mMin.getY() || mMin.getY() > y.mMax.getY() ||
		 mMax.getZ() < y.mMin.getZ() || mMin.getZ() > y.mMax.getZ()  );
    }

    bool isOverlap(const GPoint3& pt) const {
	return !(mMax.getX() < pt.getX() || mMin.getX() > pt.getX() ||
		mMax.getY() < pt.getY() || mMin.getY() > pt.getY() ||
		mMax.getZ() < pt.getZ() || mMin.getZ() > pt.getZ()  );
    }

    
    // intersect w/ a ray. If hit, tmin & tmax are 2 I-distances  
    bool findMinMax(const GRay3& ray, GFloat &tmin, GFloat &tmax) const;

    // TRANSFORM
    friend GBounds3 operator *(const GMatrix4& pM, const GBounds3& pB);

  private:
    GVector3 mMin, mMax;
};

//@ManMemo:  Basic renderable for ray-tracers.
class GRayTraceable {
  public:
    // object-type
    enum {
	PRIMITIVE    = 0x0,
	    BOUND        = 0x1,
	    CSG          = 0x2,
	    GROUP        = 0x4,
	    CONVEX       = 0x8,
	    ONESIDED     = 0x10,
	    PLANAR       = 0x20,
	    BOUNDED      = 0x40
    };
  public:
    GRayTraceable(int F = PRIMITIVE): mFlags(F), mLastRayID(-1) {}
  
    // return the type of primitives  
    int getFlags(void) const { return(mFlags); }
    void setFlags(int pFlags) { mFlags = pFlags; }

    bool isBound(void) const { return (mFlags & BOUND); }
    bool isCSG(void) const { return(mFlags & CSG); }
    bool isGroup(void) const { return(mFlags & GROUP); }
    bool isConvex(void) const { return(mFlags & CONVEX); }
    bool isOnesided(void) const { return(mFlags & ONESIDED); }
    bool isPlanar(void) const { return(mFlags & PLANAR); }
    bool isBounded(void) const { return(mFlags & BOUNDED); }

    // mailbox-stuff 
    bool checkID(const GRay3& pRay) const { 
	return (mLastRayID == pRay.getID()); 
    }
    void updateID(const GRay3& pRay) {
	mLastRayID = pRay.getID();
    }
    
    // If the rayID is in the mailbox, just adding all
    // valid intersections to rInter. 
 
    bool checkMailBox(bool& rHit, const GRay3& pRay, GInterSectionList& rInter) {
	if( checkID(pRay)) {
	    while(mInterHeap.count()) {
		rInter.add(mInterHeap.remove());
		rHit = true;
	    }
	    return (true);
	}
	else {
	    updateID(pRay);
	    mInterHeap.reset();
	    return (false);
	}
    }
    
    void sendMail(const GInterSection& pI) { 
	mInterHeap.add(pI);
    }
    
    // Get nearest sub renderable by tracing
    // the ray or 0.
    // Return the distance to the entry point:
    virtual GFloat getNearest(const GRay3&, GRayTraceable*& rRay) {
	rRay = 0;
	return(0.0);
    }

    virtual bool intersect(const GRay3&, GInterSectionList& list) = 0;
    // METHOD: Add intersections to $$list$$, return true if there was a hit.

    // One can overload this methode for faster shadow-computation!
    virtual bool hitRay(const GRay3& pRay) {
	GOneInterSectionList tList;
	return(intersect(pRay, tList));
    }
    
    
    // get normal vector at specified WC point:
    virtual GVector3 getNormal(const GPoint3& POINT) = 0;

    // get surface at specified WC point:
    virtual const GSurface& getSurface(const GPoint3&) {
	return(gDummySurface);
    }

    // Return 0 for no transmission or 1 for opaque
    // objects or an intermediate value.
    virtual GFloat getTransmission(void) const {
	return(0.0);
    }

    // bounding box in the local coordinates system:
    virtual const GBounds3& getBounds(void) const { return(mBounds); }
    // bounding box in the global coordinates system:
    virtual const GBounds3& getWorldBounds(void) const { return(mWorldBounds); }

  protected:
    GBounds3& setBounds(void) { return(mBounds); }
    GBounds3& setWorldBounds(void) { return(mWorldBounds); }
    
  private:
    // local bounding box
    GBounds3 mBounds;
    GBounds3 mWorldBounds;
    
    // type
    int mFlags;	 
    // ID-Nr of the last tested ray 
    long mLastRayID;

    // mailbox
    GAllInterSectionList mInterHeap;
};

//@ManMemo: A utility primitive for ray-tracing. To be parameterized in sub-classes.
class GRayTraceableQuadric: public GRayTraceable {
  protected: 
    // initialize quadric coefficients:
    // (quad, mixed, linear, const)
    GRayTraceableQuadric(const GVector3&, const GVector3&, const GVector3&, GFloat);
  public:
    // incremental transformations on the quadric:
    void translate(const GVector3&);
    void rotate(const GVector3&);
    void scale(const GVector3&);
  public:
    virtual GFloat getNearest(const GRay3&, GRayTraceable*&);
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
  private:
    void matrix2coefficients(void);
    void coefficients2matrix(void);
  private:
    GVector3 mQuadric, mMixed, mLinear;
    GFloat mConst;
    GMatrix4 mMat;
};

//@ManMemo: An ideal sphere basing on the quadric. The Radius is 1.
class GRayTraceableSphere: public GRayTraceableQuadric {
  public:
    GRayTraceableSphere(void): GRayTraceableQuadric(GVector3(1.0, 1.0, 1.0), GVector3(0.0, 0.0, 0.0), GVector3(0.0, 0.0, 0.0), -1.0) {
	setBounds().setMin(GVector3(-1.0, -1.0, -1.0));
	setBounds().setMax(GVector3(1.0, 1.0, 1.0));
    }
};

//@ManMemo: A ray-traceable rectangular bounding box. The dimensions have to be set on the fly.
class GRayTraceableBounds: public GRayTraceable {
    friend ostream& operator<<(ostream&, const GRayTraceableBounds&);
  public:
    // setup dimensions:
    // (min, max)
    void set(GFloat, GFloat, GFloat, GFloat, GFloat, GFloat); 
    virtual GFloat getNearest(const GRay3&, GRayTraceable*&);

    void setMin(const GVector3& pMin) { setBounds().setMin(pMin); } 
    const GVector3& getMin(void) const { return(getBounds().getMin()); }

    void setMax(const GVector3& pMax) { setBounds().setMax(pMax); } 
    const GVector3& getMax(void) const { return(getBounds().getMax()); }
  public:
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
};

typedef bool (*GIFuncPtr)(const GPoint3&, const GPoint3&, const GPoint3&);

//@ManMemo: A planar polygon without support of vertex colors.
class GRayTraceablePolygon: public GRayTraceable, virtual public GPolygon {
  public:
    // Create an empty polygon:
    GRayTraceablePolygon(void);
    // Create a polygon from a const list of vertices:
    GRayTraceablePolygon(const GVertexAggregate&);
    virtual ~GRayTraceablePolygon(void);
    
    virtual void geometryChanged(void);

    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&) {
	GVector3 tNormal(GPolygon::getNormal());
	return(tNormal);
    }  
  private:
    GIFuncPtr ifunc;

    GPoint3* mPolygon;
    int mPolySegs;
    bool mPolyValid;
    GFloat mDist;
    // distance of plane
};


#endif

