//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_RENDER_H
#define G_RENDER_H

#include "grealiz.h"
#include "util/gtempl.h"
#include "util/gvector.h"
#include "util/gmatrix.h"

//@ManMemo: Very abstract base class for renderers creating 2D pixmaps.
class GRenderer {
  public:
    // convert a normalized 2D point into a ray in the
    // virtual coordinates (world coordinates):
    virtual GRay3 inverseTransform(const GPoint2&, GFloat) const = 0;
    // The float value specifies the aspect ratio.
};

template<class P, class R>
// P stands for the primitive type to render
// R stands for the Realization to render in
//@ManMemo: Templatized abstract renderer rendering  a generic aggregate into a generic realization.
class GGRenderer: public GRenderer {
  protected:
    GGRenderer(void): mAggregate(0), mRealization(0) {}
  public:
    // the essential operation of a renderer:
    virtual void initRealization(R& pRealization) {
	mRealization = &pRealization;
	mRealization->activate();
    }
    virtual void rendering(const GAggregate<P>& pAggregate) {
	mAggregate = &pAggregate;
    }  
    // reset references:
    void reset(void) {
	mAggregate = 0;
	mRealization = 0;
    }
  protected:
    // The last used aggregate and realization:
    const GAggregate<P>* mAggregate;
    R* mRealization;
};

template<class R>
//@ManMemo: An iterator that gets the nearest generic renderable by tracing a ray.
class GGetNearestIterator: public GIterator<R> {
  public:
    GGetNearestIterator(const GRay3& pRay): mRay(pRay), mRender(0), mDist(0.0) {}
    virtual bool exec(R& r) {
	R* tRender;
	GFloat tDist;
	tDist = r.getNearest(mRay, tRender);
	if (tRender == 0) return(false);
	if (tDist <= 0.0) return(false);
	if (!mRender || (tDist < mDist)) {
	    mDist = tDist;
	    mRender = tRender;
	}
	return(false);
    }
    R* getRenderable(void) const { return(mRender); }
    GFloat getDistance(void) const { return(mDist); }
  private:
    const GRay3& mRay;
    R* mRender;
    GFloat mDist;
};

//@ManMemo: An abstract renderer basing on the lookat() camera model of OpenGL.
class GLookatCamera {
  public:
    GLookatCamera(void): mView(GPoint3(0.0, 0.0, 10.0)), 
    mRef(GPoint3(0.0, 0.0, 0.0)), mFar(100.0), mNear(0.01), 
    mTwist(0.0), mZoom(40.0), mPersp(true) {}

    void setViewpoint(const GPoint3& pView) { 
	mView = pView; 
	computeMatrix();
    }
    const GPoint3& getViewpoint(void) const { return(mView); }

    void setRefpoint(const GPoint3& pRef) { 
	mRef = pRef; 
	computeMatrix();
    }
    const GPoint3& getRefpoint(void) const { return(mRef); }

    void setFar(GFloat pFar) { mFar = pFar; }
    GFloat getFar(void) const { return(mFar); }

    void setNear(GFloat pNear) { mNear = pNear; }
    GFloat getNear(void) const { return(mNear); }

    void setTwist(GFloat pTwist) { 
	mTwist = pTwist; 
	computeMatrix();
    }
    GFloat getTwist(void) const { return(mTwist); }

    void setZoom(GFloat pZoom) { 
	mZoom = pZoom; 
	computeMatrix();
    }
    GFloat getZoom(void) const { return(mZoom); }
    
    void perspectivic(void) {
	mPersp = true;
    }
    void orthographic(void) {
	mPersp = false;
    }
    bool isPerspectivic(void) const {
	return(mPersp);
    }
  public:
    GRay3 inverseTransform(const GPoint2&, GFloat = 1.0) const;
  protected:
    void computeMatrix(void);
    const GMatrix4& getMatrix(void) const {
	return(mMat);
    }
  private:
    GPoint3 mView, mRef;
    GFloat mFar, mNear, mTwist, mZoom;
    GMatrix4 mMat;
    bool mPersp;
};

// a utility function to compute view up vector
GVector3 gCalcVUP(const GVector3&, const GVector3&, GFloat);

#endif
