//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_SOLID_TEXTURE_H
#define G_SOLID_TEXTURE_H

#include "gdata.h"
#include "util/gcolor.h"
#include "util/gmatrix.h"
#include "util/gvector.h"

extern const int gMaxNoise;

//@ManMemo: Spezification of how to combine texture with surface properties.
typedef enum GSolidTextureSampleMode {
    G_NO_SAMPLE = 0, 
    G_REPLACE_SAMPLE, // replace surface components {ambient/
    // diffuse or specular or etc.} of the primitive
    G_ADD_SAMPLE, // add texture to the surface of the primitive
    G_MUL_SAMPLE // multiply texture with the surface of the primitive
};

//@ManMemo: Spezification of how to combine texture with primitive's normal vector.
typedef enum GSolidTextureNormalMode {
    G_NO_NORMAL = 0, 
    G_REPLACE_NORMAL, // replace the normal of the primitive
    G_ADD_NORMAL // add textures normal to the normal of the primitive
};

//@ManMemo: Abstract base class of procedural 3D textures.
class GSolidTexture: public GGraphicalData {
  protected:
    GSolidTexture(void);
  public:
    virtual GColorRGB getSample(const GPoint3&) const {
	GColorRGB tColor(gBlack);
	return(tColor);
    }
    virtual GVector3 getNormal(const GVector3&) const {
	GVector3 tVector(gOrigin3);
	return(tVector);
    }

    GSolidTextureSampleMode getSampleMode(void) { 
	return(smode); 
    }
    GSolidTextureNormalMode getNormalMode(void) {
	return(nmode); 
    }

    void setMatrix(const GMatrix4&m) {
	mMatrix = m;
    }
    const GMatrix4& get_matrix(void) const {
	return(mMatrix);
    }
  protected:
    static int ***noiseTable;

    GMatrix4 mMatrix;

    GSolidTextureSampleMode smode;
    GSolidTextureNormalMode nmode;

    GFloat getNoise(const GPoint3&) const;
    GColorRGB getRGBNoise(const GPoint3&, GFloat) const;
    GFloat getTurbulence(const GPoint3&, int) const;
};

extern int base(GFloat); 
extern GFloat dsqr(GFloat);
extern GFloat dmod(GFloat, GFloat);
extern GFloat frac(GFloat d);

//@ManMemo: Simulation of poly blobs.
class GPolyBlobTexture: public GSolidTexture {
  public:
    GPolyBlobTexture(void) {}
    virtual GColorRGB getSample(const GPoint3 &) const;
};

//@ManMemo: Simulation of noise.
class GNoiseTexture: public GSolidTexture {
  public:
    GNoiseTexture(const GColorRGB& _col): col(_col) {}

    void setColor(const GColorRGB& _col) {
	col = _col;
    }
    const GColorRGB& getColor(void) const {
	return(col); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;
  public:
    GColorRGB col;
};

//@ManMemo: A unicolor texture. 
class GUniColorTexture: public GSolidTexture {
/*@Doc: 
The texture consists homogeneously of only one color. 
Objects of this class can be used in other textures.
*/
  public:
    GUniColorTexture(const GColorRGB& _col): col(_col) {}
    void setColor(const GColorRGB& _col) {
	col = _col; 
    }
    const GColorRGB& getColor(void) const {
	return(col); 
    }
    virtual GColorRGB getSample(const GPoint3 &) const { 
	GColorRGB c(col); 
	return(c); 
    }
  private:
    GColorRGB col;
};

//@ManMemo: Simulation of wood consisting of a bright and a dark surface.
class GWoodTexture: public GSolidTexture {
  public:
    GWoodTexture(GSolidTexture* t0, GSolidTexture* t1):
    xlight(0), tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GWoodTexture(void) {
	tex0->deref();
	tex1->deref(); 
    }

    void setBright(const int i) {
	xlight = i; 
    }
    int getBright(void) const {
	return(xlight); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;
  private:
    int xlight;
    GSolidTexture* tex0, * tex1;
};

//@ManMemo: A chessboard-like 3D texture consisting of two other textures.
class GCheckerTexture: public GSolidTexture {
  public:
    GCheckerTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GCheckerTexture(void) { 
	tex0->deref(); tex1->deref(); 
    }
    virtual GColorRGB getSample(const GPoint3 &) const;
  private:
    GSolidTexture* tex0, * tex1;
};

//@ManMemo: A grid consisting of two further textures.
class GGridTexture: public GSolidTexture {
  public:
    GGridTexture(GSolidTexture* t0, GSolidTexture* t1):
    xwidth(0.1), tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GGridTexture(void) {
	tex0->deref(); tex1->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;

    void setWidth(const GFloat d) { 
	xwidth = d; 
	if (xwidth < 0) xwidth = 0; 
	if (xwidth > 1) xwidth = 1; 
    }
    GFloat getWidth(void) const {
	return(xwidth); 
    }

  private:
    GFloat xwidth;
    GSolidTexture* tex0, * tex1;
};

//@ManMemo: A turbulence texture built from two further surfaces.
class GTurbulenceTexture: public GSolidTexture {
  public:
    GTurbulenceTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1), xiter(10), xlevel(0.5) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GTurbulenceTexture(void) {
	tex0->deref(); tex1->deref(); 
    }
    
    virtual GColorRGB getSample(const GPoint3 &) const;

    void setIter(const int i) {
	xiter = i;
	if (xiter < 0) xiter = 0;
    }
    int getIter(void) const {
	return(xiter); 
    }

    void setLevel(const GFloat d) {
	xlevel = d; 
    }
    GFloat getLevel(void) const {
	return(xlevel); 
    }
  private:
    GSolidTexture* tex0, * tex1;
    int xiter;
    GFloat xlevel;
};

//@ManMemo: Interpolation between two textures.
class GInterpolationTexture: public GSolidTexture {
  public:
    GInterpolationTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1), xsteps(2) {
	tex0->ref(); tex1->ref(); 
    }
    virtual ~GInterpolationTexture(void) {
	tex0->deref(); tex1->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3&) const;

    void setSteps(const int i) { 
	xsteps = i; if (xsteps < 0) xsteps = 0;
    }
    int getSteps(void) const { 
	return(xsteps); 
    }
  private:
    GSolidTexture* tex0, * tex1;
    int xsteps;
};

//@ManMemo: This texture truncs another continuous texture.
class GTruncTexture: public GSolidTexture {
  public:
    GTruncTexture(GSolidTexture* _tex): tex(_tex) {
	xsteps = 2; tex->ref();
    }
    virtual ~GTruncTexture(void) { 
	tex->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;

    void setSteps(const int i) { 
	xsteps = i; 
	if (xsteps < 0) xsteps = 0;
    }
    int getSteps(void) const {
	return(xsteps); 
    }
  private:
    GSolidTexture* tex;
    int xsteps;
};

//@ManMemo: Simulation of marble.
class GMarbleTexture: public GSolidTexture {
/*@Doc: 
This texture simulates marple by linear interpolating between two other textures. 
*/
  public:
    GMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1):
    tex0(_tex0),  tex1(_tex1) {
	tex0->ref(); tex1->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20;
    }
    virtual ~GMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); 
    }
    virtual GColorRGB getSample(const GPoint3&) const;

    void setWidths(const GFloat _w[4]) {
	w[0] = _w[0]; w[1] = _w[1]; w[2] = _w[2]; w[3] = _w[3];
    }
    GFloat* getWidths(void) const { 
	return((GFloat*)w); 
    };
  private:
    GSolidTexture* tex0, * tex1;
    GFloat w[4];
};

//@ManMemo: Marble simulation.
class GXLinearMarbleTexture: public GSolidTexture {
/*@Doc: 
This texture simulates marple by 
linear interpolating between four other textures.
*/
  public:
    GXLinearMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1, 
			  GSolidTexture* _tex2, GSolidTexture* _tex3):
    tex0(_tex0), tex1(_tex1), tex2(_tex2), tex3(_tex3) {
	tex0->ref(); tex1->ref(); tex2->ref(); tex3->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20; w[4] = 25; w[5] = 30; w[6] = 35; 
    }
    virtual ~GXLinearMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); tex2->deref(); tex3->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3&) const;

    void setWidths(const GFloat _w[7]) {
	for (int i = 0; i < 7; i++) 
	    w[i] = _w[i];
    }
    GFloat* getWidths(void) const {
	return((GFloat*)w); 
    }
  private:
    GSolidTexture* tex0, * tex1, * tex2, * tex3;
    GFloat w[7];
};

//@ManMemo: Marble simulation.
class GXCubicMarbleTexture: public GSolidTexture {
/*@Doc: 
This texture simulates marple by a cubic interpolation between four other textures.
*/
  public:
    GXCubicMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1, 
			 GSolidTexture* _tex2, GSolidTexture* _tex3) :
    tex0(_tex0), tex1(_tex1), tex2(_tex2), tex3(_tex3) {
	tex0->ref(); tex1->ref(); tex2->ref(); tex3->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20; w[4] = 25;
    }
    virtual ~GXCubicMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); tex2->deref(); tex3->deref(); 
    }

    virtual GColorRGB getSample( const GPoint3 &) const;

    void setWidths(const GFloat _w[5]) { 
	for (int i = 0; i < 5; i++) 
	    w[i] = _w[i]; 
    }
    GFloat* getWidths(void) const { 
	return((GFloat*)w); 
    }
  private:
    GSolidTexture* tex0, * tex1, * tex2, * tex3;
    GFloat w[5];
};

//@ManMemo: Lattice bump mapping.
class GLatticeTexture: public GSolidTexture {
/*@Doc: 
This demo texture creates a bump mapping in the x-y plane. The space between
the grid is 1, but can be changed via the texture matrix. This texture is very
useful for planar primitives, e.g. a quader or a plane. Surface components
will not be changed directly.
*/
  public:
    GLatticeTexture(void) { 
	smode = G_NO_SAMPLE;
	nmode = G_REPLACE_NORMAL; 
    }
    virtual GVector3 getNormal(const GPoint3&) const;
};

#endif

