//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GDYNA_LIST_H
#define GDYNA_LIST_H

#include "goutput.h"
#include "util/gtempl.h"

extern int gDynaListGetBestSize(int count, unsigned size);

template<class T> 
//@ManMemo: A list class, implemented with dynamic array
class GDynaList: public GAggregate<T> { 
  public:  
    GDynaList(int size = 0);  
    GDynaList(const GDynaList&);  
    
    virtual ~GDynaList(void);  
    
    int count(void) const;  
    T& operator [](int index);  
    const T& operator [](int index) const;  
    
    void prepend(const T&);  
    void append(const T&);  
    void insert(int index, const T&);  
    void remove(int index);  
    void reset(void);
    void free(void);
    
    void swap(int iA, int iB);  
    
    virtual void setElement(int pIndex, T& pElement);
    virtual T& getElement(int pIndex) const ;
    
    virtual void doWithElements(GIterator<T>& func) const {
        T* tItems = mItems;
        for (int i = 0; i < getNumber(); i++) {
            if (func.exec(*tItems++)) break;
        } 
    }
  private:  
    T* mItems;  
    int mSize;  
    int mFree;  
};  
 

template<class T>
inline GDynaList<T>::GDynaList(int size) {  
    if (size > 0) {  
        mSize = gDynaListGetBestSize(size, sizeof(T));  
        mItems = new T[mSize];  
    } else {  
        mSize = 0;  
        mItems = 0;  
    }  
    setNumber(0);  
    mFree = 0;  
}  

template<class T>
inline GDynaList<T>::GDynaList(const GDynaList<T>& pList) {  
    mSize = pList.mSize;
    mFree = pList.mFree;
    setNumber(pList.getNumber());
    mItems = new T[mSize];  
    for(register int i=0; i< getNumber(); i++)
      mItems[i] = pList.mItems[i];
}

template<class T>
inline GDynaList<T>::~GDynaList(void) {  
    gkillArray(mItems);  
}  
 
template<class T>
inline void GDynaList<T>::free(void) {  
    gkillArray(mItems);  
    
    mSize = 0;  
    setNumber(0);  
    mFree = 0;  
}  

template<class T>
inline void GDynaList<T>::reset(void) {  
    setNumber(0);  
    mFree = 0;  
}  

template<class T> 
inline void GDynaList<T>::swap(int iA, int iB) {  
    T temp = mItems[iA];  
    mItems[iA] = mItems[iB];  
    mItems[iB] = temp;  
}  
 
template<class T>
inline int GDynaList<T>::count(void) const { 
    return(getNumber()); 
}  

template<class T>
inline const T& GDynaList<T>::operator [](int index) const {  
    gcheck(index < 0 || index >= getNumber(),    
	   "Bad index in DynaList.");
    
    int i = index < mFree ? index : index + mSize - getNumber();  
    return(mItems[i]);  
}  

template<class T>
inline T& GDynaList<T>::operator [](int index) {  
    gcheck(index < 0 || index >= getNumber(), 
	   "Bad index in DynaList.");
    
    int i = index < mFree ? index : index + mSize - getNumber();  
    return(mItems[i]);  
}

template<class T>
inline void GDynaList<T>::setElement(int index, T& pElement) {
    gcheck(index < 0 || index >= getNumber(), 
	   "Bad index in DynaList.");
    
    int i = index < mFree ? index : index + mSize - getNumber();  
    mItems[i] = pElement;  
}

template<class T>
inline T& GDynaList<T>::getElement(int index) const {
    gcheck(index < 0 || index >= getNumber(), 
	   "Bad index in DynaList.");
    
    int i = index < mFree ? index : index + mSize - getNumber();  
    return(mItems[i]);  
}

template<class T>
inline void GDynaList<T>::append(const T& item) { 
    insert(getNumber(), item); 
}  

template<class T>
inline void GDynaList<T>::prepend(const T& item) { 
    insert(0, item); 
}  

template<class T>
inline void GDynaList<T>::insert(int index, const T& item) {  
    if (getNumber() == mSize) {  
        int size = gDynaListGetBestSize(mSize + 1, sizeof(T));  
        T* items = new T[size];  
        if (mItems != 0) {  
            register int i;  
            for (i = 0; i < mFree; ++i) {  
                items[i] = mItems[i];  
            }  
            for (i = 0; i < getNumber() - mFree; ++i) {  
                items[mFree + size - getNumber() + i] =  
                  mItems[mFree + mSize - getNumber() + i];  
            }  
            delete [] mItems;  
        }  
        mItems = items;  
        mSize = size;  
    }  
    if (index >= 0 && index <= getNumber()) {  
        if (index < mFree) {  
            for (register int i = mFree - index - 1; i >= 0; --i) {  
                mItems[index + mSize - getNumber() + i] = mItems[index + i];  
            }  
        } else if (index > mFree) {  
            for (register int i = 0; i < index - mFree; ++i) {  
                mItems[mFree + i] = mItems[mFree + mSize - getNumber() + i];  
            }  
        }  
        mFree = index + 1;  
        incNumber();  
        mItems[index] = item;  
    }  
}  

template<class T>
inline void GDynaList<T>::remove(int index) {  
    if (index >= 0 && index <= getNumber()) {  
        if (index < mFree) {  
            for (register int i = mFree - index - 2; i >= 0; --i) {  
                mItems[mSize - getNumber() + index + 1 + i] =  
                  mItems[index + 1 + i];  
            }  
        } else if (index > mFree) {  
            for (register int i = 0; i < index - mFree; ++i) {  
                mItems[mFree + i] = mItems[mFree + mSize - getNumber() + i];  
            }  
        }  
        mFree = index;  
        decNumber();  
    }  
}  
 
template<class T>
class GDynaListIterator {  
 public:  
    GDynaListIterator(const GDynaList<T> &);  
    
    int more(void) const;  
    const T& cur(void) const;  
    void next(void);  
 
 private:  
    const GDynaList<T>* mList;  
    int mCur;  
};  
 
template<class T>
inline int GDynaListIterator<T>::more(void) const { 
    return(mCur < mList->count()); 
}  

template<class T>
inline const T& GDynaListIterator<T>::cur(void) const { 
    return((*mList)[mCur]); 
}  

template<class T>
inline void GDynaListIterator<T>::next(void) { 
    mCur++; 
}  
 
template<class T>
inline GDynaListIterator<T>::GDynaListIterator(const GDynaList<T>& list) {  
    mList = &list;  
    mCur = 0;  
}  
 
#endif
