//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_MATRIX_H
#define G_MATRIX_H

#include "util/gvector.h"

// get the corrected values of sin(angle) and cos(angle)
extern void gSinCos(GFloat pAngle, GFloat& rSin, GFloat& rCos);

//@ManMemo: A homogeneous 4x4 matrix.
class GMatrix4 {
/*@Doc: 
A homogeneous 4x4 matrix. A flag may be used to inquire if the
matrix is different from the unity matrix.
*/
    friend ostream& operator<<(ostream&, const GMatrix4&);
  public:
    // Default constructor:
    GMatrix4(void) { 
	unity();
    }
    // copy CTor
    GMatrix4(const GMatrix4&);

    // contructor, init the matrix w/ 3 basis-vectors U, V, W
    GMatrix4(const GVector3& pU, const GVector3& pV, const GVector3& pW) {
	val[0][0] = pU.mX;
	val[1][0] = pU.mY;
	val[2][0] = pU.mZ;

	val[0][1] = pV.mX;
	val[1][1] = pV.mY;
	val[2][1] = pV.mZ;

	val[0][2] = pW.mX;
	val[1][2] = pW.mY;
	val[2][2] = pW.mZ;

	val[0][3] = val[1][3] = val[2][3] =
	    val[3][0] = val[3][1] = val[3][2] = 0;
	val[3][3] = 1;
    }
    
    // Set/Get element by index: 
    void set(int i, int j, GFloat d) {
	val[i][j] = d; 
	changed();
    } 
    GFloat get(int i, int j) const {
	return(val[i][j]);
    } 

    // Copy into 4x4 float array: 
    void get(GFloat f[4][4]) {
	for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 4; j++)
		f[i][j] = val[i][j];
    }

    
    GMatrix4 operator+(const GMatrix4&) const; 
    // Create a matrix by addition of a matrix.
    GMatrix4 operator-(const GMatrix4&) const; 
    // Create a matrix by subtraction of a matrix.
    GMatrix4 operator*(const GMatrix4&) const; 
    // Create a matrix by multiplication with a matrix.
    GVector3 operator*(const GVector3&) const; 
    // Create a vector by multiplication with a vector.
    GMatrix4 operator*(const GFloat&) const; 
    // Create a matrix by multiplication with a scalar.
    GMatrix4 inverse(void) const; 
    // Create the inverse matrix.
    GMatrix4 transpose(void) const; 
    // Create a transposed matrix.
    

    GMatrix4& operator=(const GMatrix4&); 
    GMatrix4& operator+=(const GMatrix4&); 
    GMatrix4& operator-=(const GMatrix4&); 
    GMatrix4& operator*=(const GMatrix4&); 
    GMatrix4& operator*=(const GFloat&); 

    GVector3 transformNormal(const GVector3&) const; 
    // Create a normal by multiplication with a normal.
    
    // Create a (inversely) scalary scaled matrix:
    GMatrix4 scaleS(const GFloat&) const;
    GMatrix4 invScaleS(const GFloat&) const;

    // Create a (inversely) scaled matrix:
    GMatrix4 scale(const GVector3&) const;
    GMatrix4 invScale(const GVector3&) const;

    // Create a (inversely) translated matrix:
    GMatrix4 translate(const GVector3&) const;
    GMatrix4 invTranslate(const GVector3&) const;

    // Create a (inversely) x-rotated matrix:
    GMatrix4 rotateX(const GFloat) const;
    GMatrix4 invRotateX(const GFloat) const;

    // Create a (inversely) y-rotated matrix:
    GMatrix4 rotateY(const GFloat) const;
    GMatrix4 invRotateY(const GFloat) const;

    // Create a (inversely) z-rotated matrix:
    GMatrix4 rotateZ(const GFloat) const;
    GMatrix4 invRotateZ(const GFloat) const;

    // Create a (inversely) x-y-z-rotated matrix:
    GMatrix4 rotate(const GVector3&) const;
    GMatrix4 invRotate(const GVector3&) const;

    // Create a general rotation matrix:
    GMatrix4 rotate(const GVector3& pAxis, GFloat pAngle) const;


    // Initialize matrix with unity matrix:
    void unity(void);

    // Check changes on matrix:
    bool isUnityMatrix(void) const {
	return(mChanged);
    }
  private:
    GFloat val[4][4];
    bool mChanged;
    // if false then matrix is unity matrix 
    void changed(void) {
	mChanged = true;
    }
};

extern const GMatrix4 gUnityMatrix4; 

typedef struct {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A matrix pair.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: For ray-tracers and other applications
    // is migth be useful to compute an inverse matrix together with
    // the modeling matrix.
    // CONSTANTS: 
    GMatrix4 mat, imat;
} GMatrix4Pair;

#endif
