//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_PARAMETRIC_TOPOLOGY_H
#define G_PARAMETRIC_TOPOLOGY_H

#include "util/gtempl.h"
#include "util/gtopolgy.h"
#include "util/ggenvec.h"
#include "model/gresol.h"

typedef GAggregate<GCurveVertex> GCurveVertexAggregate;
typedef GAggregate<GSurfaceVertex> GSurfaceVertexAggregate;

typedef GConstantVector<GCurveVertex> GCurveVertexVector;
typedef GConstantVector<GSurfaceVertex> GSurfaceVertexVector;

typedef GDLList<GCurveVertex> GCurveVertexList;
typedef GDLList<GSurfaceVertex> GSurfaceVertexList;

extern const int gMaxPoints;

//@ManMemo: Abstract base class for all control-polygon-based primitives.
class GCurveTopology: public GTopology1D, public GResolutionAttribute {
  public:
    GCurveTopology(int);
    GCurveTopology(const GCurveVertexAggregate&);
    
    ~GCurveTopology(void) { if(mCurv) delete mCurv; }

    void setVertex(int, GCurveVertex*);
    void setVertex(int, const GCurveVertex&);
    const GCurveVertex& getVertex(int) const;

    const GCurveVertexList& getVertexList(void) const {
	return (const GCurveVertexList&)GTopology1D::getVertexList();
    }

    virtual int getSegments(void) const { return getNumber(); }

    virtual int getCurveDim(void) const { return mCurv->getNumber(); }
    
    virtual const GVector& getParameterVector(void);
    virtual const GCurveVertex getCurveVertexOn(GFloat);

    virtual const GCurveVertex& getCurveVertex(int i) const {
	return(mCurv->getElement(i));
    }

    virtual const GCurveVertexVector& getAll() const {
	return(*mCurv);
    }

    void setResolution(const GResolution& pResolution) {
	mResolution = pResolution;
	GResolutionAttribute::setResolution(pResolution);
	
	geometryChanged();
    }

    void computeNormals(void);

    virtual const GResolution& getResolution(void) const { return mResolution; }
    GFloat getResolutionValue(void) const { return getResolution().getResolution(); }
  protected:
    virtual void setCurveVertex(int, GCurveVertex);

    virtual void resolutionChanged(void);
    virtual void newGeometry(int);
    // set new geometry to the curve, 
    // has to be called always when the curve changes
    GCurveTopology(void);
  private:
    GResolution mResolution;
    GCurveVertexVector* mCurv;
};

//@ManMemo: Internally used curve of ghost points.
class GPhantomCurveTopology: public GCurveTopology {
  public:
    GPhantomCurveTopology(int pNumber): GCurveTopology(pNumber) {}

    GPhantomCurveTopology(const GCurveVertexAggregate& pVertices): GCurveTopology(pVertices) {}
  protected:
    void resolutionChanged(void) {};

    GPhantomCurveTopology(void): GCurveTopology() {}
};

//@ManMemo: Base class for all lateral-mesh-based primitives.
class GSurfaceTopology: public GTopology2D, public GResolutionAttribute {
  public:
    GSurfaceTopology(int, int);
    GSurfaceTopology(int, int, const GSurfaceVertexAggregate&);

    ~GSurfaceTopology(void) { if(mSurf) delete mSurf; }

    virtual int getSegmentsX(void) const { return getX(); }
    virtual int getSegmentsY(void) const { return getY(); }

    void setVertices(const GSurfaceVertexAggregate&);

    GSurfaceVertex** getVertices(void) {
	return (GSurfaceVertex**)getBasePointer();
    }

    const GSurfaceVertexVector& getSurfaceVertices(void) const {
	return *mSurf;
    }

    void setVertex(int, int, GSurfaceVertex*);
    void setVertex(int, int, const GSurfaceVertex&);
    const GSurfaceVertex& getVertex(int, int) const;

    virtual const GVector& getParameterVectorU(void); 
    virtual const GVector& getParameterVectorW(void);
    // get parameter values in u or w direction
        
    virtual const GSurfaceVertex getSurfaceVertexOn(GFloat, GFloat); 
    // get surface values on parameter value u, w

    // get vertex on curve with that index
    const GSurfaceVertex& getSurfaceVertex(int, int) const;
	
    int getSurfX() const { return dimSurfX; }
    int getSurfY() const { return dimSurfY; }
    // get dimensions of the surface, depending on resolution

    virtual void setResolution(const GResolution& pResolution) {
	mResolution = pResolution;
	GResolutionAttribute::setResolution(pResolution);

	geometryChanged();
    }

    void computeNormals(void);

    virtual const GResolution& getResolution(void) const { return mResolution; }
    GFloat getResolutionValue(void) const { return getResolution().getResolution(); }
  protected:  
    virtual void setSurfaceVertex(int, int, GSurfaceVertex);

    virtual void resolutionChanged(void);    
    virtual void newGeometry(void);
    virtual void newGeometry(int, int);

    GSurfaceTopology(void): GTopology2D(), mResolution(0), mSurf(0) {}
    // create an empty topology
  private:
    int dimSurfX, dimSurfY;
    GResolution mResolution;
    GSurfaceVertexVector *mSurf;
};

//@ManMemo: Internally used surface of ghost points.
class GPhantomSurfaceTopology: public GSurfaceTopology {
  public:
    GPhantomSurfaceTopology(int pX, int pY): GSurfaceTopology(pX, pY) {}

    GPhantomSurfaceTopology
	(int pX, int pY, const GSurfaceVertexAggregate& pVertices): 
    GSurfaceTopology(pX, pY, pVertices) {}
  protected:
    void resolutionChanged(void) {};
    
    GPhantomSurfaceTopology(void): GSurfaceTopology() {}
};

#endif




