//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_CURVE_2_SURF_TOPOLOGY_H
#define G_CURVE_2_SURF_TOPOLOGY_H

#include "util/gparam2D.h"
#include "util/gmatrix.h"

extern const int gNumberOfFacets;

//@ManMemo: Base class for parametric sweeps and rotation curves.
class GCurve2SurfaceTopology: public GPhantomCurveTopology {
  public:
    GCurve2SurfaceTopology(const GCurveVertexAggregate&, const GResolution&,
			   const GSplineMode&, const GSplineMode&);

    ~GCurve2SurfaceTopology(void) { if(mSurf) delete mSurf; }

    int getX(void) const { return mSurf->getX(); }
    int getY(void) const { return mSurf->getY(); }

    int getSurfX(void) const { return mSurf->getSurfX(); }
    int getSurfY(void) const { return mSurf->getSurfY(); }

    const GSurfaceVertex& getNetVertex
	(int pX, int pY) const { return mSurf->getVertex(pX, pY); }

    void setNetVertex
	(int pX, int pY, const GSurfaceVertex&  pVertex) {
	    mSurf->setVertex(pX, pY, new GSurfaceVertex(pVertex));
	}

    const GSurfaceVertexVector& getSurfaceVertices
	(void) const { return mSurf->getSurfaceVertices(); }

    const GSurfaceVertex& getSurfaceVertex
	(int pX, int pY) const { return mSurf->getSurfaceVertex(pX, pY); }

    GSurfaceVertex getSurfaceVertexOn
	(GFloat pU, GFloat pW) {  return mSurf->getSurfaceVertexOn(pU, pW); }

    const GVector& getParameterVectorU(void) {
	return  mSurf->getParameterVectorU();
    }

    const GVector& getParameterVectorW(void) {
	return  mSurf->getParameterVectorW();
    }

    void setModeU(const GSplineMode& pModeU) {
	mModeU = pModeU;

	geometryChanged();
    }

    void setModeW(const GSplineMode& pModeW) {
	mModeW = pModeW;

	geometryChanged();
    }

    const GSplineMode& getModeU(void) const { return mSurf->getModeU(); }
    const GSplineMode& getModeW(void) const { return mSurf->getModeW(); }

    virtual const GSplineMode& getMode(void) const = 0;
  protected:
    virtual void newGeometry(int, int);

    void geometryChanged(void) {
	// change the geometry of the part surface finally:
	mSurf->geometryChanged();
    }

    GCurve2SurfaceTopology(void): GPhantomCurveTopology() {}
  private:
    GFittingSplineSurface* mSurf;
    GSplineMode mModeU, mModeW;
};

//@ManMemo: Generic parametric rotation surface.
class GRotationSurfaceTopology: public GCurve2SurfaceTopology {
  public:
    GRotationSurfaceTopology(const GVector3&, const GCurveVertexAggregate&, 
			     const GResolution&, const GSplineMode&);

    void setRotationAxis(const GVector3& pAxis) { 
	if(pAxis.abs() < gEpsilon) return;
	mAxis = pAxis;

	recompute();
	geometryChanged();
    }

    const GVector3& getRotationAxis(void) const { return mAxis; }

    void setMode(const GSplineMode& pMode) {
	setModeW(pMode);
    }

    const GSplineMode& getMode(void) const { return getModeW(); }

  protected:
    void recompute(void);
    GPoint3 wc2rc(const GPoint3&, GFloat pAng);
    GFloat getDistanceFromAxis(const GPoint3&);

    void resolutionChanged(void);
    void geometryChanged(void);
    GRotationSurfaceTopology(void): GCurve2SurfaceTopology() {}
  private:
    GVector3 mAxis; //GPoint3 mMin, mMax;
    GMatrix4 mB, mBInv, mC, mCInv, mR;
};

//@ManMemo: Generic sweeping surface.
class GSweepSurfaceTopology: public GCurve2SurfaceTopology {
  public:
    GSweepSurfaceTopology(GFloat, const GVector3&, const GCurveVertexAggregate&, 
			  const GResolution&, const GSplineMode&);

    void setLength(GFloat pLength) {
	mLength = pLength;
	geometryChanged();
    }

    const GFloat& getLength(void) const {
	return mLength;
    }

    void setSweepVector(const GVector3& pSweepVector) {
	mSweepVector = pSweepVector;
	geometryChanged();
    }

    const GVector3& getSweepVector(void) const {
	return mSweepVector;
    }

    void setMode(const GSplineMode& pMode) {
	setModeU(pMode);
    }

    const GSplineMode& getMode(void) const { return getModeU(); }
  protected:
    void resolutionChanged(void);
    void geometryChanged(void);
    GSweepSurfaceTopology(void): GCurve2SurfaceTopology() {}
  private:
    GFloat mLength;
    GVector3 mSweepVector;
};

#endif
