//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GENERAL_SPLINE_H__
#define __GENERAL_SPLINE_H__                                              

#include "util/ggenmat.h"
#include "util/ggenvec.h"

// enumerates which implement the base features of the spline:
typedef enum GSplineMode {
    G_OPEN = 0, // default
    G_CLOSED,
    G_NO_MODE
    };

typedef enum GSplineEnd {
    G_CLAMPED = 0, // default
    G_FREE,
    G_NO_END
    };

class GSplineBase {
  public:
    GSplineBase(const GMatrix pMatrix) {
	mMatrix = pMatrix;
	wanted = false;
	upToDate = false;
    }

    GSplineBase(void) {
	wanted = false;
	upToDate = false;
    }

    GMatrix mMatrix;

    bool& isWanted(void) { return wanted; }
    bool& isUpToDate(void) { return upToDate; }
  private:
    bool wanted, upToDate;
};

//@ManMemo: Basic spline values. Depending on these values will the approximation (interpolation) be carried out.
class GSplineValues {
friend class GSpline;    
  public:
    int numVal,order,dim;
    GSplineMode mode;
    GSplineEnd  end;

    GSplineValues(const GSplineValues& v) {
	numVal = v.numVal; dim = v.dim; order = v.order; mode = v.mode; end = v.end;
    }

    GSplineValues(void) { 
	numVal = dim = order = 0;
	mode = G_OPEN;
	end = G_CLAMPED;
    }
    
    GSplineValues
	(int a, int b, int c, GSplineMode d,GSplineEnd e) {
	    numVal = a; dim = b; order = c; mode = d; end = e;
	}

    GSplineValues& operator=(const GSplineValues& v) {
	numVal = v.numVal;
	order = v.order;
	dim = v.dim;
	mode = v.mode;
	end = v.end; 

	return *this;
    }
    
    bool operator==(const GSplineValues &v) {
	if(numVal == v.numVal && order == v.order &&
	   dim == v.dim && mode == v.mode && end == v.end)
	    return true;
	else
	    return false;
    }
};

//@ManMemo: An abstract base class with basic b-spline funcionality.
class GSpline {
/*@DOC:
An abstract base class, with basic b-spline funcionality
The class computes the blending functions for a given spline values and stores 
in matrices N,N1,N2
*/
  public:
    // methods:
    int getSegments(void) const;
  protected:
    void compBase(void);
    void compBase1(void);
    void compBase2(void);

    virtual GVector compKnotVector(void) = 0;
    virtual GVector compParVector(void);

    // set up knot and parameter vectors,
    // must be called always, if request for any base or base
    // translation occurs
    virtual void update(void);

    // set up matrices if necessary
    // called by any request for any kind of base matrix
    void updateMatrices(void);
    
    // the instances will be recomputed
    // must be called after each change
    bool& setChanged(void) { return changed; }
  public:
    // changed test    
    const bool isChanged(void) const { return changed; }
  protected:
    GFloat blend(int, int, GFloat);
    GFloat blend1(int, int, GFloat);
    GFloat blend2(int, int, GFloat);
    // inlines:
    void setKnotVector(const GVector& vec) { K = vec; }
    void setParVector(const GVector &vec)  { T = vec; }

    GSpline(void);
    GSpline(const GSplineValues&);
  public:
    const GVector& getParVector(void) {
	update();

	return getPrivParVector();
    }

    const GVector& getKnotVector(void) {
	update();

	return getPrivKnotVector();
    }

    GMatrix& getBaseMatrix(void);
    GMatrix& getBaseMatrix1(void);
    GMatrix& getBaseMatrix2(void);
    
    GVector getBaseTranslation(GFloat);
    GVector getBaseTranslation1(GFloat);
    GVector getBaseTranslation2(GFloat);
  public:
    GSplineValues& setSpline(void) { changed = true; return this->values; }
    const GSplineValues& getSpline(void) const { return this->values; }
  protected:
    const GVector& getPrivParVector(void)  { return T; } 
    const GVector& getPrivKnotVector(void) { return K; }
  private:
    GSplineValues values;
    bool changed;

    // matrices of base functions and their derivatives
    GSplineBase mBase, mBase1, mBase2;

    // current parameter vector
    GVector T;
    // knot vector
    GVector K;
    // methods:
    int checkValues(void);
};

//@ManMemo: Base class for approximating splines. 
class GApproxSpline : public GSpline {
  private:
    // flags to control the status of base matrices
    bool fullFlag, isRebuilt, isRebuilt1, isRebuilt2;
    
    void rebuildVector(GVector&);
    void rebuild(GMatrix&);
  protected:
    // methods:
    void update(void);

    GVector compKnotVector(void);
  public:
    GApproxSpline(void);
    GApproxSpline(const GSplineValues&);

    const GMatrix& getBaseMatrix(void);
    const GMatrix& getBaseMatrix1(void);
    const GMatrix& getBaseMatrix2(void);

    GVector getBaseTranslation(GFloat);
    GVector getBaseTranslation1(GFloat);
    GVector getBaseTranslation2(GFloat);
};

typedef struct {
    GVector mTranslation1, mTranslation2;
    GVector mValue1, mValue2;
} GSplineConstraints;

//@ManMemo: A fitting spline type.
class GNaturalFittingSpline: public GSpline {
  protected:
    // methods:
    void update(void);
    void updateVertexMatrix(void);

    GVector compKnotVector(void);
    GVector compParVector(void);

    virtual void computeConstraints(void);
  protected:
    GSplineConstraints& setConstraints(void) { return mConstr; }
    const GSplineConstraints& getConstraints(void) const { return mConstr; }
  protected:
    const GMatrix& getPrivateBase(void);
    const GMatrix& getPrivateBase1(void);
    const GMatrix& getPrivateBase2(void);

    // instance of approximating b-spline
    GApproxSpline mResult;
  public:
    GNaturalFittingSpline(void);
    GNaturalFittingSpline(const GSplineValues&);

    int getSegments(void) const {
	return mResult.getSegments();
    }

    GVector getBaseTranslation(GFloat);
    GVector getBaseTranslation1(GFloat);
    GVector getBaseTranslation2(GFloat);

    const GMatrix& getBaseMatrix(void);
    const GMatrix& getBaseMatrix1(void);
    const GMatrix& getBaseMatrix2(void);

    const GMatrix& getVertexMatrix(void);
    void setValuesMatrix(const GMatrix&);

    void setMode(const GSplineMode&);
    const GSplineMode& getMode(void) const;

    void setDimension(int);
    const int getDimension(void) const;
    
    const GVector& getParVector(void) {
	return mResult.getParVector();
    }

    const GVector& getKnotVector(void) {
	return mResult.getKnotVector();
    }
  private:
    GMatrix mOutput, mInput;
    GSplineConstraints mConstr;
};

//@ManMemo: A fitting spline type.
class GGeneralFittingSpline: public GNaturalFittingSpline {
  protected:
    void computeConstraints(void);
  public:
    GGeneralFittingSpline(void);
    GGeneralFittingSpline(const GSplineValues&, const GVector&, const GVector&);

    // overload, only the mode of must not mResult be set
    void setMode(const GSplineMode&);
    const GSplineMode& getMode(void) const;

    void setTangentBeg(const GVector&);
    void setTangentEnd(const GVector&);
};

#endif



