//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GTABLE_H
#define  GTABLE_H

template<class K, class V>
//@ManMemo: parametric hash table. 
class GTable {
friend class GTableIterator<K, V>;
  public:
    GTable(int pSize);
    ~GTable(void);
    
    // insert the key w/ given hash-value
    bool insert(const K& pKey, const V& pValue);

    // search the given key 
    bool find(const K& pKey);

    // find and replace/remove the given entry
    bool replace(const K& pKey, const V& pValue);
    bool remove(const K& pKey, V& rValue);
    void remove(const K& pKey);

    bool lookup(const K&, V&);
    
    // free up the table
    void free(void);

  private:
    int mSize;
    GTableEntry<K, V>** mFirst;
    GTableEntry<K, V>** mLast;

    GTableEntry<K, V>*& probe(const K&);
};


template<class K, class V>
class GTableEntry {
  friend class GTable<K, V>;
  friend class GTableIterator<K, V>;
  public:
    GTableEntry(const K& pK, const V& pV) 
    :mKey(pK), mValue(pV) {}
    
  private:
    K mKey;
    V mValue;
    
    GTableEntry * mChain;
};


template<class K, class V>
class GTableIterator {
  public:
    GTableIterator(const GTable<K, V>& pTable) {
	mLast = pTable.mLast;
	for (mEntry = pTable.mFirst; mEntry <= mLast; mEntry++) 
	    if((mCur = *mEntry) != 0)
		break;
    }
    

    K& curKey(void) { return (mCur->mKey); }
    V& curValue(void) { return (mCur->mValue); }
    
    bool more(void) { return (mEntry <= mLast);};
    
    bool next(void) {
	if ((mCur = mCur->mChain) != 0) 
	    return (true);
	for (++mEntry; mEntry <= mLast; mEntry++) 
	    if ((mCur = *mEntry) != 0) 
		return (true);
	
	return (false);
    }
  
  private:
    GTableEntry<K, V>*  mCur;
    GTableEntry<K, V>** mEntry;
    GTableEntry<K, V>** mLast;
};

template<class K, class V>
inline GTable<K, V>::GTable(int pSize) {
    // find the best size
    for (mSize = 32; mSize < pSize; mSize <<= 1);

    // allocate the table
    mFirst = new GTableEntry<K, V>* [mSize];
    --mSize;
    
    mLast = &mFirst[mSize];

    // initialize the chains\'s pointers
    for (register GTableEntry<K, V>** tPtr = mFirst; tPtr <= mLast; tPtr++) 
	*tPtr = 0;
    
}

template<class K, class V>
inline GTable<K, V>::~GTable(void) {
    gkillArray(mFirst);
}

template<class K, class V>
inline GTableEntry<K, V>*& GTable<K, V>::probe(const K& pKey) {
    // get the value associated w/ the hashed key.
    return (mFirst[pKey.hash() & mSize]);
}

template<class K, class V>
inline bool GTable<K, V>::insert(const K& pKey, const V& pValue) {
    // already in the table...
    if (find(pKey)) 
	return (false);

    // allocate the new entry.
    register GTableEntry<K, V>* tEntry = 
	new GTableEntry<K, V> (pKey, pValue);

    // get the old chain...
    register GTableEntry<K, V>** tChain = &probe(pKey);
    // and prepend the new entry to the chain
    tEntry->mChain = *tChain;
    *tChain = tEntry;

    return (true);
}

template<class K, class V>
inline bool GTable<K, V>::lookup(const K& pKey, V& rValue) {
    for (register GTableEntry<K, V>* tEntry = probe(pKey); 
	 tEntry; tEntry = tEntry->mChain) 
   	if (tEntry->mKey == pKey) {
	    rValue = tEntry->mValue;
	    return (true);
	}
   
    return (false);
}

template<class K, class V>
inline bool GTable<K, V>::find(const K &pKey) {
    for (register GTableEntry<K, V>* tEntry = probe(pKey); 
	 tEntry; tEntry = tEntry->mChain) 
    	if (tEntry->mKey == pKey) 
	    return (true);

    return (false);
}

template<class K, class V>
inline bool GTable<K, V>::replace(const K& pKey, const V& pValue) {
    for (register GTableEntry<K, V>* tEntry = probe(pKey); 
	 tEntry ; tEntry = tEntry->mChain) 
    	if (tEntry->mKey == pKey) {
	    tEntry->mValue = pValue;
	    return (true);
	}
    
    return (false);
}

template<class K, class V>
inline bool GTable<K, V>::remove(const K& pKey, V& rValue) {
    // find the chain
    GTableEntry<K, V>** tChain = &probe(pKey);

    // begin w/ the chain\'s first entry...
    register GTableEntry<K, V>* tEntry = *tChain;
    if (tEntry) {
	// got the right entry to delete!
	if (tEntry->mKey == pKey) {
	    rValue = tEntry->mValue;
	    *tChain = tEntry->mChain;

	    delete (tEntry);
	    return (true);
	} 
	// travers the chain..
	else {
	    register GTableEntry<K, V>* tPrev;
	    do {
		tPrev = tEntry;
		tEntry = tEntry->mChain;
	    } while (tEntry  && tEntry->mKey != pKey);
	    
	    if (tEntry ) {
		rValue = tEntry->mValue;
		tPrev->mChain = tEntry->mChain;

		delete (tEntry);
		return (true);
	    }
	}
    }
    return (false);
}

template<class K, class V>
inline void GTable<K, V>::remove(const K &pKey) {

    GTableEntry<K, V>** tChain = &probe(pKey);

    register GTableEntry<K, V>* tEntry = *tChain;
    if (tEntry) {
	if (tEntry->mKey == pKey) {
	    *tChain = tEntry->mChain;
	    
	    delete (tEntry);
	} 
	
	else {
	    register GTableEntry<K, V>* tPrev;
	    do {
		tPrev = tEntry;
		tEntry = tEntry->mChain;
	    } while (tEntry  && tEntry->mKey != pKey);

	    if (tEntry) {
		tPrev->mChain = tEntry->mChain;
		delete (tEntry);
	    }
	}
    }
}

template<class K, class V>
inline void GTable<K, V>::free(void) {
    for(GTableIterator<K, V> ti(self); ti.more(); ti.next())
	remove(ti.curKey());
}


#endif




