//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_TEMPLATES_H
#define G_TEMPLATES_H

#include <assert.h>
#include "ggeneric.h"
#include "util/gstring.h"

template<class T>
//@ManMemo: Basic abstract iterator.
class GIterator {
/*@Doc:
Derived types of this class may used to iterate over elements in
aggregates. They also may have a state. However, be carefully with
the usage of state-based iterators.
*/
  public:
    //@ManMemo: Abstract iterator function. Must return true to stop calling this function for further elements.
    virtual bool exec(T&) = 0;
};

template<class T>
//@ManMemo: Print iterator.
class GOutStreamIterator: public GIterator<T> {
/*@Doc:
Print an aggregate to the specified output stream.
The elements are spaced by a white space.
*/
  public:
    GOutStreamIterator(ostream& pStream): mStream(pStream) {}
    virtual bool exec(T& t) {
	mStream << t << ' ';
	return(false);
    }
  private:
    ostream& mStream;
};

template<class T>
//@ManMemo: Print iterator.
class GPrintIterator: public GIterator<T> {
/*@DOC:
An iterator for demonstration purposes. The functionality is to print
an aggregate to the output stream. The iterator counts the number of printed
elements.
*/
  public:
    GPrintIterator(void): i(0) {}
    virtual bool exec(T& t) {
	if (!i) cout << "printing an aggreate:\n";
	cout << '\t' << i++ << ": " << t << '\n';
	return(false);
    }
    void reset(void) { i = 0; }
  private:
    int i;
};

template<class T>
//@ManMemo: Basic aggregate in GENERIC.
class GAggregate {
  protected:
    GAggregate(void): mNumber(0) {}
  public:
    //@ManMemo: Set element by index.
    virtual void setElement(int, T&) = 0;
    //@ManMemo: Get element from index.
    virtual T& getElement(int) const = 0;

    //@ManMemo: Get element from index.
    T& operator[](int pIndex) const {
	return(getElement(pIndex));
    }

    //@ManMemo: Iterate over aggregate elements.
    virtual void doWithElements(GIterator<T>&) const = 0;

    //@ManMemo: Inquire number of elements.
    int getNumber(void) const { return(mNumber); }
  protected:
    //@ManMemo: Increment number after addition of an element.
    void incNumber(void) { mNumber++; }
    
    //@ManMemo: Decrement number after deletion of an element.
    void decNumber(void) { mNumber--; }

    //@ManMemo: Set number explicitely.  
    void setNumber(int pNumber) { mNumber = pNumber; }
  private:
    int mNumber;
};

template<class T>
//@ManMemo: A fix-sized aggregate.
class GConstantVector: public GAggregate<T> {
/*@DOC:
A ConstantVector holds a fix number of elements. Elements will be copied
by value. All elements will be deleted in destructor. The Number of elements
is fixed in constructor.
*/
  public:
    //@ManMemo: Initialize the aggregate with a fix size.
    GConstantVector(int pNumber) {
	assert(pNumber > 0);
	setNumber(pNumber);
	mVector = new T[getNumber()];
    }
    virtual ~GConstantVector(void) {
	delete [] mVector;
    }
    virtual void setElement(int pIndex, T& pElement) {
	mVector[pIndex] = pElement;
    }
    virtual T& getElement(int pIndex) const {
	return(mVector[pIndex]);
    }
    virtual void doWithElements(GIterator<T>& func) const {
	T* tVector = mVector;
	for (int i = 0; i < getNumber(); i++) {
	    if (func.exec(*tVector++)) break;
	} 
    }
  private:
    T* mVector;
};

template<class T>
//@ManMemo: An element for dynamic lists.
class GDLListElement {
/*@DOC:
An element for DL lists consisting of prev and next pointer and
an element pointer.
*/
  public:
    GDLListElement(T* pEntry): mEntry(pEntry) {}
    GDLListElement<T>* prev;
    GDLListElement<T>* next;

    // Set/get access:
    void setElement(T& pEntry) {
	delete mEntry;
	mEntry = &pEntry; 
    } 
    T& getElement(void) const { return(*mEntry); }
    T* getPointer(void) const { return(mEntry); }
  private:
    T* mEntry;
};

template<class T>
//@ManMemo: A generic DL list storing just references, not values. 
class GDLList: public GAggregate<T> {
/*@DOC:
A generic DL list storing just references, not values. 
In constructor an automatic deletion of the elements may be requested.
Default: off.
*/
  public:
    //@ManMemo: Initialize double-linked list.
    GDLList(bool pClear = false): root(0), last(0), mClear(pClear) {}

    inline virtual ~GDLList(void);

    //@ManMemo: Append an element at end.
    inline void append(T* pElement) { appendAtEnd(pElement); }

    //@ManMemo: Append an element at begin.
    inline void appendAtBegin(T*);

    //@ManMemo: Append an element at end.
    inline void appendAtEnd(T*);

    //@ManMemo: Remove an element, start search at begin.
    inline void remove(T* pElement){ removeFromBegin(pElement); }

    //@ManMemo: Remove an element, start search at begin.
    inline void removeFromBegin(T*);

    //@ManMemo: Remove an element, start search at end.
    inline void removeFromEnd(T*);

    //@ManMemo: Remove all elements.
    inline void reInitialize(void);

    inline virtual void setElement(int, T&);
    inline virtual T& getElement(int) const;
    
    inline void deleteElement(int);

    inline virtual void doWithElements(GIterator<T>&) const;
  private:
    GDLListElement<T>* root;
    GDLListElement<T>* last;
    bool mClear;
};

template<class T>
inline GDLList<T>::~GDLList(void) {
    reInitialize();
}

template<class T>
inline void GDLList<T>::appendAtBegin(T* pElement) {
    if (!root) {
	root = new GDLListElement<T>(pElement); 
	root->prev = 0;
	root->next = 0;
	last = root;
    }
    else {
	GDLListElement<T>* tmp = new GDLListElement<T>(pElement); 
	tmp->prev = 0;
	tmp->next = root;
	root->prev = tmp;
	root = tmp;
    }
    incNumber();
}

template<class T>
inline void GDLList<T>::appendAtEnd(T* pElement) {
    if (!last) {
	last = new GDLListElement<T>(pElement); 
	last->prev = 0;
	last->next = 0;
	root = last;
    }
    else {
	GDLListElement<T>* tmp = new GDLListElement<T>(pElement); 
	tmp->prev = last;
	tmp->next = 0;
	last->next = tmp;
	last = tmp;
    }
    incNumber();
}

template<class T>
inline void GDLList<T>::removeFromBegin(T* pElement) { 
    GDLListElement<T>* tmp = root;
    if (!tmp) return;
    do if (&tmp->getElement() == pElement) {
	if (tmp->prev) tmp->prev->next = tmp->next;
	else root = tmp->next;
	if (tmp->next) tmp->next->prev = tmp->prev;
	else last = tmp->prev;
	if (mClear) delete tmp->getPointer();
	delete tmp;
	decNumber();
	return;
    }
    while (tmp = tmp->next);
}

template<class T>
inline void GDLList<T>::removeFromEnd(T* pElement) { 
    GDLListElement<T>* tmp = last;
    if (!tmp) return;
    do if (&tmp->getElement() == pElement) {
	if (tmp->prev) tmp->prev->next = tmp->next;
	else root = tmp->next;
	if (tmp->next) tmp->next->prev = tmp->prev;
	else last = tmp->prev;
	if (mClear) delete tmp->getPointer();
	delete tmp;
	decNumber();
	return;
    }
    while (tmp = tmp->prev);
}

template<class T>
inline void GDLList<T>::setElement(int pIndex, T& pElement) {
    assert(pIndex >= 0);
    if (pIndex >= getNumber()) append(&pElement);
    else {
	GDLListElement<T>* tmp = root;
	for (int i = 0; i < getNumber(); i++) {
	    assert(tmp); 
	    if (i == pIndex) {
		tmp->setElement(pElement);
		break;
	    }
	    else tmp = tmp->next;
	}
    }
}

template<class T>
inline T& GDLList<T>::getElement(int pIndex) const {
    assert((0 <= pIndex) && (pIndex < getNumber()));
    GDLListElement<T>* tmp = root;
    for (int i = 0; i < getNumber(); i++) {
	assert(tmp); 
	if (i == pIndex) return(tmp->getElement());
	else tmp = tmp->next;
    }
    // never reached cos first assertion:
    return(tmp->getElement());
}

template<class T>
inline void GDLList<T>::deleteElement(int pIndex) { 
    assert((0 <= pIndex) && (pIndex < getNumber()));
    GDLListElement<T>* tmp = root;
    for (int i = 0; i < getNumber(); i++) {
	assert(tmp); 
	if (i == pIndex) {
	    if (tmp->prev) tmp->prev->next = tmp->next;
	    else root = tmp->next;
	    if (tmp->next) tmp->next->prev = tmp->prev;
	    else last = tmp->prev;
	    if (mClear) delete tmp->getPointer();
	    delete tmp;
	    decNumber();
	    return;
	}
	else tmp = tmp->next;
    }
}

template<class T>
inline void GDLList<T>::doWithElements(GIterator<T>& func) const {
    GDLListElement<T>* tmp = root;
    if (!tmp) return;
    do {
	T& tElement = tmp->getElement();
	tmp = tmp->next;
	if (func.exec(tElement)) break;
    } while(tmp);
}

template<class T>
inline void GDLList<T>::reInitialize(void) {
    GDLListElement<T>* tmp = root;
    GDLListElement<T>* x;
    if (tmp) do {
	x = tmp;
	if (mClear) delete tmp->getPointer();
	tmp = tmp->next;
	delete x;
    }
    while(tmp);
    root = last = 0;
}

typedef int GCacheValue;

template<class T>
//@ManMemo: Encapsulation of a cache value and a related DL list.
class GCacheTableEntry {
  public:
    GCacheTableEntry(GCacheValue pValue): mValue(pValue), mList(true) {}
    GCacheValue getCacheValue(void) const {
	return(mValue);
    }
    GDLList<T>* getCacheList(void) {
	return(&mList);
    }
  private:
    GCacheValue mValue;
    GDLList<T> mList;
};

template <class T>
//@ManMemo: A cache iterator.
class GFindCacheListIterator: public GIterator<T> {
/*@DOC:
Find the DL list related to the initial cache value.
*/
  public:
    GFindCacheListIterator(GCacheValue pValue): mValue(pValue), mList(0) {}
    virtual bool exec(T& t) {
	if (mValue == t.getCacheValue()) {
	    mList = &t;
	    return(true);
	}
	else return(false);
    }
    T* getList(void) const {
	return(mList);
    }
  private:
    GCacheValue mValue;
    T* mList;
};

template<class T>
//@ManMemo: Append iterator.
class GAppendAggregateIterator: public GIterator<T> {
/*@DOC:
Append all or a specified number of elements on an existing aggregate.
Use setNumber() to apply one and the same iterator object to different
aggregates in case you want to do a limited appending.
*/
  public:
    GAppendAggregateIterator(GDLList<T>& pList, bool pIndividualize = false): mList(pList), mIndividualize(pIndividualize), mNumber(0) {}
    virtual bool exec(T& t) {
	if (mIndividualize) mList.append(new T(t));
	else mList.append(&t);
	if (mNumber) {
	    mCounter++;
	    if (mNumber == mCounter) return(true);
	}
	return(false);
    }
    void setNumber(int pNumber) {
	assert(pNumber > 0);
	mNumber = pNumber;
	mCounter = 0;
    }
  private:
    GDLList<T>& mList;
    bool mIndividualize;
    int mNumber;
    int mCounter;
};

template<class T>
//@ManMemo: Append iterator.
class GAppendAggregateIterator2: public GIterator<T> {
/*@DOC:
Append all or a specified number of elements on an existing aggregate,
Use setNumber() to apply one and the same iterator object to different
aggregates in case you want to do a limited appending.
This version may be used for abstract element types
because the absence of the individualization.
*/
  public:
    GAppendAggregateIterator2(GDLList<T>& pList): mList(pList), mNumber(0) {}
    virtual bool exec(T& t) {
	mList.append(&t);
	if (mNumber) {
	    mCounter++;
	    if (mNumber == mCounter) return(true);
	}
	return(false);
    }
    void setNumber(int pNumber) {
	assert(pNumber > 0);
	mNumber = pNumber;
	mCounter = 0;
    }
  private:
    GDLList<T>& mList;
    int mNumber;
    int mCounter;
};

template<class T>
//@ManMemo: Copy iterator.
class GCopyAggregateIterator: public GIterator<T> {
/*@DOC:
Copy all aggregate elements into a classic C-like array.
Use setNumber() to apply one and the same iterator object to different
aggregates in case you want to do a limited copying.
*/
  public:
    GCopyAggregateIterator(T* pVector): mVector(pVector), mNumber(0) {}
    virtual bool exec(T& t) {
	*mVector++ = t;
	return(false);
    } 
    void setNumber(int pNumber) {
	assert(pNumber > 0);
	mNumber = pNumber;
	mCounter = 0;
    }
  private:
    T* mVector;
    int mNumber;
    int mCounter;
};

template<class T>
//@ManMemo: Copy Iterator.
class GCopyAggregateIterator2: public GIterator<T> {
/*@DOC:
Copy all aggregate elements into an another aggregate.
*/
  public:
    GCopyAggregateIterator2(GAggregate<T>& pAggr): mAggr(pAggr), mCounter(0) {}
    virtual bool exec(T& t) {
	mAggr.setElement(mCounter++, t);
	return(false);
    } 
  private:
    GAggregate<T>& mAggr;
    int mCounter;
};

template<class T>
//@ManMemo: Compare iterator.
class GCompareAggregateIterator: public GIterator<T> {
/*@DOC:
Compare two aggregate by checking the aquivalency of the elements.
*/
  public:
    GCompareAggregateIterator(const GAggregate<T>& pAggr): mAggr(pAggr), mCounter(0), mEqual(true) {}
    virtual bool exec(T& t) {
	if (t == mAggr.getElement(mCounter++)) return(false);
	else {
	    mEqual = false;
	    return(true);
	}
    } 
    bool wasEqual(void) const { return(mEqual); }
  private:
    const GAggregate<T>& mAggr;
    int mCounter;
    bool mEqual;
};

template <class T>
//@ManMemo: Search iterator.
class GFindByNameIterator: public GIterator<T> {
/*@DOC:
Look for an object by specified name.
*/
  public:
    GFindByNameIterator(const char* pName): mName(pName), mT(0) {}
    virtual bool exec(T& t) {
	if (mName == t.getName()) {
	    mT = &t;
	    return(true);
	}
	else return(false);
    }
    T* get(void) { return(mT); }
  private:
    const GString mName;
    T* mT;
};

template <class T>
//@ManMemo: Search iterator.
class GFindElementIterator: public GIterator<T> {
/*@DOC:
Find an element by checking address identity.
*/
  public:
    GFindElementIterator(const T* pElement): mElement(pElement), mFound(false) {}
    virtual bool exec(T& t) {
	if (&t == mElement) {
	    mFound = true;
	    return(true);
	}
	else return(false);
    }
    bool found(void) const { return(mFound); }
  private:
    const T* mElement;
    bool mFound;
};

#endif

