//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_TOPOLOGY_H
#define G_TOPOLOGY_H

#include "util/gvertex.h"
#include "util/gfacet.h"
#include "util/gtempl.h"

typedef GAggregate<GVertex> GVertexAggregate;
typedef GAggregate<GFacet> GFacetAggregate;
typedef GDLList<GVertex> GVertexList;
typedef GDLList<GFacet> GFacetList;

GNormal gCalcNormal(int n, GPoint3*);

//@ManMemo: Abstract geometry class.
class GGeometry {
/*@DOC:
The class defines an update hook method for changes of geometry.
*/
  protected:
    virtual void geometryChanged(void) {}
    // METHOD: To be called if geometry has changed.
};

//@ManMemo: One-dimensional topology.
class GTopology1D: public GGeometry  {
/*@DOC:
The class provides a DL list of vertices (1D toplogy) and may be used
as a base class for polygon, polyline, polymarker, etc.
The vertices are stored inside the primitive and are automatically deleted in
destructor.
*/ 
  public:
    GTopology1D(void): mList(true) {}
    // CTOR: Create an empty 1D topology.
    GTopology1D(const GVertexAggregate&);
    // CTOR: Create an 1D topology from a const list of vertices:

    virtual ~GTopology1D(void) {}

    virtual void setVertex(int, const GVertex&);
    // METHOD: Set a vertex value. If the specified index is larger
    // than the contained indexes, a new vertex will be appended. 

    const GVertex& getVertex(int) const;
    // METHOD: Supply a vertex value. An invalid number will cause an
    // assertion failure. Therefore, check the validity of the
    // index, e.g. by usage of getNumber().

    void deleteVertex(int);
    // METHOD: Delete a vertex from the list.

    int getNumber(void) const { return(mList.getNumber()); }
    // METHOD: Get number of vertices.

    const GVertexList& getVertexList(void) const {
	return(mList);
    }
    // METHOD: Get the complete vertex list.
    virtual void computeNormals(void);
    // METHOD: Compute normal vectors in clock-wise direction. 
  protected:
    void setVertex(int, GVertex*);
    
    // METHOD: Direct access to vertex list. $$Vertex$$ will be set as is.
  private:
    GVertexList mList;
};

//@ManMemo: Base class for planar polygons.
class GPolygon: public GTopology1D {
  protected:
    GPolygon(void): mNormal(0.0) {}
    // CTOR: Default constructor.
    GPolygon(const GVertexAggregate& pPoints): GTopology1D(pPoints), mNormal(0.0) {
	geometryChanged();
    } 
    // CTOR: Initialize polygon from a list of points.
  public:
    void setNormal(const GNormal& pNormal) {
	mNormal = pNormal.unitize();
	geometryChanged();
    }
    // METHOD: Set global normal vector.
    const GNormal& getNormal(void) const {
	return(mNormal);
    }
    // METHOD: Get global normal vector.
  private:
    GNormal mNormal;
};

//@ManMemo: Base class for two-dimensional topology.
class GTopology2D: public GGeometry  {
  protected:
    GTopology2D(void);
    // CTOR: Create an empty topology. The derived class is responsible
    // for the correct initialization.
    void setX(int pX) { mX = pX; }
    // METHOD: Set an x dimension for the 2D topology. The array of vertices
    // must be allocated in relation to this value.
    void setY(int pY) { mY = pY; }
    // METHOD: Set an y dimension for the 2D topology. The array of vertices
    // must be allocated in relation to this value.
    bool checkVertex(int, int);
    // check if there is a unset vertex on this position
    bool checkVertices(void);
    // check if there are unset vertices
  public:
    GTopology2D(int, int);
    // CTOR: Create an x*y sized static 2D topology with default
    // initialization of vertices.

    GTopology2D(int ,int, const GVertexAggregate&);

    virtual ~GTopology2D(void);
    // METHOD: Destruct object and remove dynamically created
    // vertices.

    int getX(void) const { return(mX); }
    // METHOD: Supply x dimension.
    int getY(void) const { return(mY); }
    // METHOD: Supply y dimension.

    virtual void setVertex(int, int, const GVertex&);
    // METHOD: Set a new $$vertex$$ at specified location ($$x$$, $$y$$).

    const GVertex& getVertex(int, int) const;
    // METHOD: Get a $$vertex$$ from specified location ($$x$$, $$y$$).
    
    virtual void computeNormals(void);
    // METHOD: Compute normal vectors.
  protected:
    void setVertex(int, int, GVertex*);
    // direct vertex access

    GVertex** getBasePointer(void) { return(mVertices); }
    // GVertex** getBasePointer(void) { return(mVertices); }
    // METHOD: Allow direct access to the vertex array for derived classes.

    bool checkIndices(int, int) const;
    // METHOD: Check validity of a 2D coordinates index.
  private:
    int mX, mY;
    GVertex** mVertices;
};

//@ManMemo: Base class for polyhedral topologies.
class GTopologyPh: public GGeometry  {
  public:
    GTopologyPh(void): mVList(true), mFList(true) {}
    // CTOR: Empty construction. 

    GTopologyPh(const GVertexAggregate&, const GFacetAggregate&);
    // CTOR: Create polyhedron from constant lists.
    
    virtual ~GTopologyPh(void) {}

    const GVertex& getVertex(int) const; 
    // METHOD: Get vertex by index in point list.
    const GVertex& getVertex(int FNR, int PNR) const; 
    // METHOD: Get vertex by index PNR in facet FNR.

    const GFacet& getFacet(int) const;
    // METHOD: Get facet by number.
    
    void appendVertices(const GVertexAggregate&);
    // METHOD: Append points to the end of list.
    void appendFacets(const GFacetAggregate&);
    // METHOD: Append facets to the end of list.

    int getNumberOfVertices(void) const { return(mVList.getNumber()); }
    // METHOD: Supply number of vertices.
    int getNumberOfFacets(void) const { return(mFList.getNumber()); }
    // METHOD: Supply number of facets.

    const GVertexList& getVertexList(void) { return(mVList); } 
    // METHOD: Supply constant vertex list.
    const GFacetList& getFacetList(void) { return(mFList); } 
    // METHOD: Supply constant facet list.

    void computeNormals(void);
    // METHOD: Compute normal vectors for each facet.
  private:
    GVertexList mVList;
    GFacetList mFList;    
};

#endif


