//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_VECTOR_H
#define G_VECTOR_H

#include <math.h>
#include <stdio.h>
#include "ggeneric.h"

enum { POS_X, NEG_X, POS_Y, NEG_Y, POS_Z, NEG_Z };

//@ManMemo: A type consisting of two scalars.
class GVector2 {
    friend ostream& operator<<(ostream&, const GVector2&);
  public:
    GVector2(void) { mX = mY = 0; }
    GVector2(GFloat pVal[2]): mX(pVal[0]), mY(pVal[1]) {}
    GVector2(GFloat pX, GFloat pY): mX(pX), mY(pY) {} 
    GVector2(GFloat pF): mX(pF), mY(pF) {} 

    void setX(GFloat pX) { mX = pX; }
    GFloat getX(void) const { return mX; }

    void setY(GFloat pY) { mY = pY; }
    GFloat getY(void) const { return mY; }

    bool operator==(const GVector2&) const;
    

    GVector2 operator+(const GVector2& v) const { 
	GVector2 c(mX + v.mX, mY + v.mY);
	return(c);
    }
    GVector2 operator-(const GVector2& v) const {
	GVector2 c(mX - v.mX, mY - v.mY);
	return(c); 
    }

    GVector2 operator*(const GFloat d) const {
	GVector2 c(mX * d, mY * d);
	return(c);  
    }

    
    GFloat sqdist(void) const {
	return(mX * mX + mY * mY);
    }

    
  private:
    GFloat mX, mY;
};

typedef GVector2 GPoint2;

//@ManMemo: Elementary 3D vector.
class GVector3 {
    friend class GMatrix4;
    friend ostream& operator<<(ostream&, const GVector3&);
    friend GVector3 gCalcVUP(const GVector3&, const GVector3&, GFloat);
 public:
    GVector3(void) { mX = mY = mZ = 0; }
    GVector3(GFloat pVal[3]): mX(pVal[0]), mY(pVal[1]), mZ(pVal[2]) {}
    GVector3(GFloat pX, GFloat pY, GFloat pZ): mX(pX), mY(pY), mZ(pZ) {} 
    GVector3(GFloat pF): mX(pF), mY(pF), mZ(pF) {} 

    void setX(GFloat pX) { mX = pX; }
    GFloat getX(void) const { return mX; }

    void setY(GFloat pY) { mY = pY; }
    GFloat getY(void) const { return mY; }

    void setZ(GFloat pZ) { mZ = pZ; }
    GFloat getZ(void) const { return mZ; }


    bool operator==(const GVector3&) const;

    GFloat abs(void) const {
	return(sqrt(mX * mX + mY * mY + mZ * mZ));
    }

    GFloat sqdist(void) const {
	return(mX * mX + mY * mY + mZ * mZ);
    }

    GVector3 scale(GFloat f) const {	
	return(*this * f);
    }

    GFloat dot(const GVector3& v) const {
	return(mX * v.mX + mY * v.mY + mZ * v.mZ);    
    }

    GVector3 cross(const GVector3& v) const {
	GVector3 c(mY * v.mZ - mZ * v.mY, 
		   mZ * v.mX - mX * v.mZ,
		   mX * v.mY - mY * v.mX);
	return(c);   
    }

    GVector3 eval(const GVector3& v) const {
	GVector3 c(mX * v.mX, mY * v.mY, mZ * v.mZ );
	return(c);  
    }

    GVector3 unitize(void) const {
	GFloat r = abs();
	assert(r != 0.0);
	GVector3 c(mX / r, mY / r, mZ / r);
	return(c);
    }

    // get into OpenGL/IRIS GL like float array:
    void get(float f[3]) const { 
	f[0] = mX; f[1] = mY; f[2] = mZ; 
    }

    GVector3 negate(void) const {
	GVector3 c(-mX, -mY, -mZ );
	return(c); 
    }

    GFloat& operator[] (int inx) {
	if (inx == 0) return(mX);
	else if (inx == 1) return(mY);
	else return(mZ);
    }

    const GFloat& operator[] (int inx) const {
	if (inx == 0) return(mX);
	else if (inx == 1) return(mY);
	else return(mZ);
    }

    GVector3& operator+=(const GVector3& A) {
	mX += A.mX;  
	mY += A.mY;  
	mZ += A.mZ;
	return(self);
    }

    GVector3& operator-=(const GVector3& A) {
	mX -= A.mX;  
	mY -= A.mY;  
	mZ -= A.mZ;
	return(self);
    }

    GVector3& operator*=(GFloat A) {
	mX *= A;  
	mY *= A;  
	mZ *= A;
	return(self);
    }

    GVector3& operator/=(GFloat A) {
	mX /= A;  
	mY /= A; 
	mZ /= A;
	return(self);
    }

    GVector3& operator*=(const GVector3& A) {
	mX *= A.mX; 
	mY *= A.mY; 
	mZ *= A.mZ;
	return(self);
    }

    GVector3& normalize(void) {
	GFloat r = abs();
	assert(r != 0.0);
	
	mX /= r; mY /= r; mZ /= r;
	
	return(self);
    }

    void setMin(const GVector3& a_) {
	// METHOD: Apply a minimum vector to this vector.
	if (a_.mX < mX) mX = a_.mX;
	if (a_.mY < mY) mY = a_.mY;
	if (a_.mZ < mZ) mZ = a_.mZ;
    }

    void setMax(const GVector3& a_) {
	// METHOD: Apply a maximum vector to this vector.
	if(a_.mX > mX) mX = a_.mX;
	if(a_.mY > mY) mY = a_.mY;
	if(a_.mZ > mZ) mZ = a_.mZ;
    }
    
    void project(GFloat& px, GFloat& py, int axis) const{
	// METHOD: Project the point to a plane.
	switch (axis) {
	  case 0:
	    px = mY;
	    py = mZ;
	    break;
	  case 1:
	    px = mX;
	    py = mZ;
	    break;
	  default:
	    px = mX;
	    py = mY;
	    break;
	}
    }

    int getDominantAxis(void) const {
	GFloat tX = gabs(mX);
	GFloat tY = gabs(mY);
	GFloat tZ = gabs(mZ);
	
	if(tX > tY && tX > tZ) 
	  return ((mX > 0) ? POS_X : NEG_X);
	else if ( tY > tZ )
	  return ((mY > 0) ? POS_Y : NEG_Y);
	else
	  return ((mZ > 0) ? POS_Z : NEG_Z);
    }
    
    //// friends
    friend GVector3 operator+(const GVector3& A, const GVector3& B) {
	return(GVector3(A.mX + B.mX, A.mY + B.mY, A.mZ + B.mZ));
    }

    friend GVector3 operator-(const GVector3& A, const GVector3& B) {
	return(GVector3(A.mX - B.mX, A.mY - B.mY, A.mZ - B.mZ));
    }

    friend GVector3 operator*(const GVector3& A, const GVector3& B) {
	return(GVector3(A.mX * B.mX, A.mY * B.mY, A.mZ * B.mZ));
    }

    friend GVector3 operator*(const GVector3& A, GFloat B) {
	return(GVector3(A.mX * B, A.mY * B, A.mZ * B));
    }

    friend GVector3 operator*(GFloat A, const GVector3& B) {
	return(B * A);
    }

    friend GVector3 operator/(const GVector3& A, const GVector3& B) {
	return(GVector3(A.mX / B.mX, A.mY / B.mY, A.mZ / B.mZ));
    }

    friend GVector3 operator/(const GVector3& A, GFloat B) {
	return(GVector3(A.mX / B, A.mY / B, A.mZ / B));
    }

    friend GVector3 Vec3Min(const GVector3& a, const GVector3& b) {
	// Return a vector containing the minimal components of the 
	// passed vectors.
	GVector3 tVec;
	tVec.mX = gmin(a.mX, b.mX);
	tVec.mY = gmin(a.mY, b.mY);
	tVec.mZ = gmin(a.mZ, b.mZ);
	return(tVec);
    }

    friend GVector3 Vec3Max(const GVector3& a, const GVector3& b) {
	// Return a vector containing the maximal components of the 
	// passed vectors.
	GVector3 tVec;
	tVec.mX = gmax(a.mX, b.mX);
	tVec.mY = gmax(a.mY, b.mY);
	tVec.mZ = gmax(a.mZ, b.mZ);
	return(tVec);
    }
 private:
    GFloat mX, mY, mZ;
};

typedef GVector3 GPoint3;
typedef GVector3 GBounds;
typedef GVector3 GNormal;

extern const GPoint3 gOrigin3;

//@ManMemo: Representation of a linear ray.
class GRay3 {
/*@DOC:
A simple representation of a linear ray. The length is optional.
Its interpretation depends on the scaling of the direction vector.
*/
    friend ostream& operator<<(ostream&, const GRay3&);
  public:
    // ray types 
    enum  {
	G_UNKNOW  = 0x00,
	G_EYE     = 0x01,
	G_SHADOW  = 0x02,
	G_REFLECT = 0x03,
	G_REFRACT = 0x04,
    };
    
    GRay3(void) : mLength(0), mTmin(gEpsilon), mTmax(HUGE), 
    mFlags(G_UNKNOW), mLevel(0) {
	//mID++;
    }
    
    GRay3(const GPoint3& pPoint, const GVector3& pVector, 
	  int pFlags = G_UNKNOW, int pLevel = 0)
    : mPoint(pPoint), mVector(pVector), mLength(0), 
      mTmin(gEpsilon), mTmax(HUGE), mFlags(pFlags), mLevel(pLevel) 
    {
	// increase the global counter
	if(mFlags) mID++;
	
	// compute the path for this ray
	if( pLevel >= G_REFLECT ) {   
	    mLevel = pLevel;
	    
	    unsigned int tMask = 0x01 << pLevel;
	    
	    // Turn off the reflection bit and turn on the refraction bit
	    (mFlags == G_REFLECT) ? (mPath &= ~tMask) : (mPath |= tMask);
	}
    }
    
    // set/get coordinate:
    void setPoint(const GPoint3& pPoint) { mPoint = pPoint; }
    const GPoint3 getPoint(void) const { return(mPoint); }

    // set/get direction vector:
    void setVector(const GVector3& pVector) { mVector = pVector; }
    const GVector3 getVector(void) const { return(mVector); }

    // set/get length:
    void setLength(GFloat pLength) { mLength = pLength; }
    GFloat getLength(void) const { return(mLength); }

    // compute 3D point for passed distance:
    GPoint3 compute(GFloat t) const {
	return(mPoint + mVector * t);
    }
    bool isValid(GFloat t) const {
	return((t >= mTmin) && (t <= mTmax));
    }
    bool isValid(GFloat bmin, GFloat bmax) const {
	// sort it:
	if (bmin > bmax) {
	    GFloat t = bmin; 
	    bmin = bmax; 
	    bmax = t;
	}
	return(!((bmax < mTmin) || (bmin > mTmax)));
    }
    
    void setTmax(GFloat pTmax) { mTmax = pTmax; }
    GFloat getTmax(void) const { return(mTmax); }
    
    void setTmin(GFloat pTmin) { mTmin = pTmin; }
    GFloat getTmin(void) const { return(mTmin); }
  
    // set/get the rays ID (used by mailbox-technique)
    long getID(void) const {
	return (mID);
    }
    
    void setID(long pID) {
	mID = pID;
    }
			
    // quick queries to rays type 
    bool isShadowRay(void) const {
	return(mFlags == G_SHADOW);
    }
    bool isReflectRay(void) const{
	return(mFlags == G_REFLECT);
    }
    bool isRefractRay(void)  const{
	return (mFlags == G_REFRACT);
    }
    bool isEyeRay(void)  const{
	return (mFlags == G_EYE);
    }
   
    // Compute the point w/ distance t to the ray origin 
    GPoint3 extrap(GFloat pT) const {
	return(mPoint + mVector * pT);
    }
    
    // generate a random ray within a solid angle along this ray.
    // -1 <= pR1, pR2 <= +1 : 2 random variables 
    // pAngle : the max. angle between the new ray and this ray
    GRay3 distribute(GFloat pAngle, GFloat pR1, GFloat pR2) const;

    // update max. and min. extensions 
    void minmax(GFloat pMin, GFloat pMax) {
	if (pMin > pMax) { 
	    GFloat t = pMin;
	    pMin = pMax;
	    pMax = t;
	}
	
	mTmin = gmax(pMin, mTmin);
	mTmax = gmin(pMax, mTmax);
    }
    
    // convert the ray to a 5D-tuple(X,Y,Z,U,V) and a dominant axis
    int compute5D(GFloat *r5D) const {
	int tAxis = mVector.getDominantAxis();
	
	// compute the UV-coord.
	int tDominantCoord = tAxis/2;
	GFloat tU = mVector[(tDominantCoord+1)%3];
	GFloat tV = mVector[(tDominantCoord+2)%3];
	
	r5D[G_D_U] = tU / gabs(mVector[tDominantCoord]);
	r5D[G_D_V] = tV / gabs(mVector[tDominantCoord]);
	r5D[G_D_X] = mPoint.getX();
	r5D[G_D_Y] = mPoint.getY();
	r5D[G_D_Z] = mPoint.getZ();
	
	return(tAxis);
    }


    int getPath(void) const {
	return (mPath);
    }
    int getLevel(void) const {
	return (mLevel);
    }
  private:
    GPoint3 mPoint;
    GVector3 mVector;
    GFloat mLength;
    
    // ray's interval
    GFloat mTmin, mTmax;
    
    // type of ray
    int mFlags;    
    
    // for cache's stuff
    int mLevel; 

    static int mPath;
    
    // for mailbox
    static long mID; 
};

#endif

