//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_VERTEX_H
#define G_VERTEX_H

#include "util/gvector.h"
#include "util/gcolor.h"
#include "util/gsurface.h"

//@ManMemo: Basic vertex type.
class GVertex {
/*@DOC:
Representation of vertices consisting of 3D geometry and an optional normal
vector.
*/
  public:
    //@ManMemo: Default initializer.
    GVertex(void): mNormal(0) {}
    //@ManMemo: Initialize geometry.
    GVertex(const GVector3& pPoint): mPoint(pPoint), mNormal(0) {}
    //@ManMemo: Initialize geometry and normal vector.
    GVertex(const GVector3& pPoint, const GVector3 &pNormal): mPoint(pPoint), mNormal(0) { 
	setNormal(&pNormal);
    }
    //@ManMemo: Copy constructor.
    GVertex(const GVertex& pVertex): mNormal(0) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
    }

    virtual ~GVertex(void) { if (mNormal) delete mNormal; }

    //@ManMemo: Assigment operator.
    GVertex& operator=(const GVertex& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	return(*this);
    }

    //@ManMemo: Set the point.
    void setPoint(const GVector3& pPoint) { mPoint = pPoint; }
    //@ManMemo: Get the point.
    const GVector3& getPoint(void) const { return(mPoint); }

    //@ManMemo: Set the normal vector.
    void setNormal(const GVector3* pNormal) {
	if (!pNormal) {
	    if (mNormal) {
		delete mNormal;
		mNormal = 0;
	    }
	}
	else {
	    if (!mNormal) mNormal = new GVector3;
	    *mNormal = *pNormal;
	}
    }
    //@ManMemo: Get the normal vector.
    const GVector3* getNormal(void) const { return(mNormal); }
  private:
    GVector3 mPoint;
    GVector3* mNormal;
};

ostream& operator<<(ostream&, const GVertex&);

//@ManMemo: Extended vertex type.
class GVertexWithColorRGB: public GVertex {
/*@DOC:
A vertex additionally providing an RGB color (optionally).
*/
  public:
    // Default initializer:
    GVertexWithColorRGB(void): mColor(0) {}
    // Initialize geometry:
    GVertexWithColorRGB(const GVector3& pPoint): GVertex(pPoint), mColor(0) {}
    // Initialize geometry and normal vector:
    GVertexWithColorRGB(const GVector3& pPoint, const GVector3 &pNormal): GVertex(pPoint, pNormal), mColor(0) {}
    // Initialize geometry and color:
    GVertexWithColorRGB(const GVector3& pPoint, const GColorRGB& pColor): GVertex(pPoint), mColor(0) {
	setColor(&pColor);
    }
    // Initialize all components:
    GVertexWithColorRGB(const GVector3& pPoint, const GVector3 &pNormal, const GColorRGB& pColor): GVertex(pPoint, pNormal), mColor(0) {
	setColor(&pColor);
    }

    // Copy constructor:
    GVertexWithColorRGB(const GVertexWithColorRGB& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	setColor(pVertex.getColor());
    }

    virtual ~GVertexWithColorRGB(void) { if (mColor) delete mColor; }

    // Assigment operator:
    GVertexWithColorRGB& operator=(const GVertexWithColorRGB& pVertex) {
	(GVertex&)*this = pVertex;
	setColor(pVertex.getColor());
	return(*this);
    }

    // Set/get color value:
    void setColor(const GColorRGB* pColor) {
	if (!pColor) {
	    if (mColor) {
		delete mColor;
		mColor = 0;
	    }
	}
	else {
	    if (!mColor) mColor = new GColorRGB;
	    *mColor = *pColor;
	}
    }
    const GColorRGB* getColor(void) const { return(mColor); }
  private:
    GColorRGB* mColor;
};

ostream& operator<<(ostream&, const GVertexWithColorRGB&);

//@ManMemo: An extended vertex.
class GVertexWithSurface: public GVertex {
/*@DOC:
A vertex additionally providing a surface (optionally).
*/
  public:
    // Default initializer:
    GVertexWithSurface(void): mSurface(0) {}
    // Initialize geometry:
    GVertexWithSurface(const GVector3& pPoint): GVertex(pPoint), mSurface(0) {}
    // Initialize geometry and normal vector:
    GVertexWithSurface(const GVector3& pPoint, const GVector3 &pNormal): GVertex(pPoint, pNormal), mSurface(0) {}
    // Initialize geometry and surface:
    GVertexWithSurface(const GVector3& pPoint, const GSurface& pSurface): GVertex(pPoint), mSurface(0) {
	setSurface(&pSurface);
    }
    // Initialize all components:
    GVertexWithSurface(const GVector3& pPoint, const GVector3 &pNormal, const GSurface& pSurface): GVertex(pPoint, pNormal), mSurface(0) {
	setSurface(&pSurface);
    }

    // Copy constructor:
    GVertexWithSurface(const GVertexWithSurface& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	setSurface(pVertex.getSurface());
    }

    virtual ~GVertexWithSurface(void) { if (mSurface) delete mSurface; }

    // Assigment operator:
    GVertexWithSurface& operator=(const GVertexWithSurface& pVertex) {
	(GVertex&)*this = pVertex;
	setSurface(pVertex.getSurface());
	return(*this);
    }

    // Set/get surface pointer:
    void setSurface(const GSurface* pSurface) {
	if (!pSurface) {
	    if (mSurface) {
		delete mSurface;
		mSurface = 0;
	    }
	}
	else {
	    if (!mSurface) mSurface = new GSurface;
	    *mSurface = *pSurface;
	}
    }
    const GSurface* getSurface(void) const { return(mSurface); }
  private:
    GSurface* mSurface;
};

ostream& operator<<(ostream&, const GVertexWithSurface&);

//@ManMemo: An extended vertex.
class GCurveVertex: public GVertex {
/*@DOC:
A vertex additionally providing a surface (optionally), a gradient
and a tangent.
*/
  public:
    GCurveVertex(void): mTangent(0), mGradient(0), mT(0) {}
    // default init.

    GCurveVertex(const GVector3& pPoint, GFloat pT = 0): GVertex(pPoint), mTangent(0), mGradient(0), mT(pT) {}
    // initialize with point value

    GCurveVertex(const GVector3& pPoint, const GVector3& pNormal, GFloat pT = 0): GVertex(pPoint, pNormal), mTangent(0), mGradient(0), mT(pT) {}
    // initialize with point and normal values

    GCurveVertex(const GVector3& pPoint, const GVector3& pNormal, const GVector3& pTangent, const GVector3& pGradient, GFloat pT = 0): GVertex(pPoint, pNormal), mTangent(0), mGradient(0), mT(pT) {
	setTangent(&pTangent);
	setGradient(&pGradient);
    }
    // initialize with point, normal, tangent and gradient values
    
    GCurveVertex
	(const GCurveVertex& pVertex): mTangent(0), mGradient(0), mT(0) {
	    setPoint(pVertex.getPoint());
	    setNormal(pVertex.getNormal());
	    setTangent(pVertex.getTangent());
	    setGradient(pVertex.getGradient());

	    mT = pVertex.mT;
	}
    // Copy constructor
    
    virtual ~GCurveVertex(void) {
	if(mTangent) delete mTangent; if(mGradient) delete mGradient;
    }

    GCurveVertex& operator=(const GCurveVertex& pVertex) {
	(GVertex&)*this = pVertex;
	// set base class values

	setTangent(pVertex.getTangent());
	setGradient(pVertex.getGradient());

	mT = pVertex.mT;
	// the rest
	
	return *this;
    }

    bool operator==(const GCurveVertex&) const;

    void setTangent(const GVector3* pTangent) {
	if(!pTangent) {
	    if(mTangent) {
		delete mTangent;
		mTangent = 0;
	    }
	}
	else
	    {
		if(!mTangent) mTangent = new GVector3;
		*mTangent = *pTangent;
	    }
    }

    void setGradient(const GVector3* pGradient) {
	if(!pGradient) {
	    if(mGradient) {
		delete mGradient;
		mGradient = 0;
	    }
	}
	else
	    {
		if(!mGradient) mGradient = new GVector3;
		*mGradient = *pGradient;
	    }
    
    }

    const GVector3* getTangent(void) const { return mTangent; }
    const GVector3* getGradient(void) const { return mGradient; }
    // set/get tangent, gradient and surface values

    void setT(GFloat pT) { mT = pT; }
    GFloat getT(void) const { return mT; }
  private:
    GVector3 *mTangent, *mGradient;
    GFloat mT;
};

ostream& operator<<(ostream&, const GCurveVertex&);
    
//@ManMemo: An extended vertex.
class GSurfaceVertex: public GVertex {
/*@DOC:
A vertex additionally providing a surface (optionally) and tangent
vectors.
*/
  public:
    GSurfaceVertex(void): mTangentU(0), mTangentW(0), mU(0), mW(0) {}
    // default init.

    GSurfaceVertex(const GVector3& pPoint, GFloat pU = 0, GFloat pW = 0): GVertex(pPoint), mTangentU(0), mTangentW(0), mU(pU), mW(pW) {}
    // initialize with point value

    GSurfaceVertex(const GVector3& pPoint, const GVector3& pTangentU, const GVector3& pTangentW, GFloat pU = 0, GFloat pW = 0): GVertex(pPoint), mTangentU(0), mTangentW(0), mU(pU), mW(pW) {
	setTangentU(&pTangentU);
	setTangentW(&pTangentW);
    }
    // initialize with point, normal, tangentU and tangentW values

    GSurfaceVertex(const GVector3& pPoint, const GVector3& pNormal, const GVector3& pTangentU, const GVector3& pTangentW, GFloat pU = 0, GFloat pW = 0): GVertex(pPoint, pNormal), mTangentU(0), mTangentW(0), mU(pU), mW(pW) {
	setTangentU(&pTangentU);
	setTangentW(&pTangentW);
    }
    // initialize with all values

    GSurfaceVertex
	(const GSurfaceVertex& pVertex): mTangentU(0), mTangentW(0) {
	    setPoint(pVertex.getPoint());
	    setNormal(pVertex.getNormal());
	    setTangentU(pVertex.getTangentU());
	    setTangentW(pVertex.getTangentW());

	    mU = pVertex.mU;
	    mW = pVertex.mW;
	}
    // Copy constructor
    
    virtual ~GSurfaceVertex(void) {
	if(mTangentU) delete mTangentU; if(mTangentW) delete mTangentW;
    }

    GSurfaceVertex& operator=(const GSurfaceVertex& pVertex) {
	(GVertex&)*this = pVertex;
	// set base class values

	setTangentU(pVertex.getTangentU());
	setTangentW(pVertex.getTangentW());

	mU = pVertex.mU;
	mW = pVertex.mW;
	// the rest

	return *this;
    }

    bool operator==(const GSurfaceVertex&) const;

    void setTangentU(const GVector3* pTangentU) {
	if(!pTangentU) {
	    if(mTangentU) {
		delete mTangentU;
		mTangentU = 0;
	    }
	}
	else
	    {
		if(!mTangentU) mTangentU = new GVector3;
		*mTangentU = *pTangentU;
	    }
    }

    void setTangentW(const GVector3* pTangentW) {
	if(!pTangentW) {
	    if(mTangentW) {
		delete mTangentW;
		mTangentW = 0;
	    }
	}
	else
	    {
		if(!mTangentW) mTangentW = new GVector3;
		*mTangentW = *pTangentW;
	    }
    }

    const GVector3* getTangentU(void) const { return mTangentU; }
    const GVector3* getTangentW(void) const { return mTangentW; }
    // set/get tangentU and tangentW values

    void setU(GFloat pU) { mU = pU; }
    void setW(GFloat pW) { mW = pW; }

    GFloat getU(void) const { return mU; }
    GFloat getW(void) const { return mW; }
  private:
    GVector3 *mTangentU, *mTangentW;
    GFloat mU, mW;
};

ostream& operator<<(ostream&, const GSurfaceVertex&);

#endif






