//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <stdlib.h>
#include "render/graytrac.h"
#include "util/gheap.h"
#include "util/gstack.h"

#include "gx/gx_defs.h"
#include "gx/gx_bbox.h"

// KayKaijya BVH-Traversing 
class KKNode   {
  public:
    GFloat mT;
    GXObject *mObj;
    KKNode(void)   {}
    KKNode(GXObject *pObj, GFloat pT) {
	mObj = pObj;
	mT = pT;
    }
    
  friend int operator <(const KKNode& x, const KKNode& y) {
      return (x.mT < y.mT);
  }
  friend int operator >(const KKNode& x, const KKNode& y) {
      return (x.mT > y.mT);
  }
  friend int operator ==(const KKNode& x, const KKNode& y) {
      return (x.mT == y.mT);
  }
};
typedef GHeap<KKNode> __KKHeap;



class StkNode {
  public:  
    GXObjectList *mList;
    long mN;
};
typedef GStack<StkNode> ListStk;
#define GET_OBJ(node) (*((node).mList))[(node).mN]


// 
// Calculate the cost for a bounding volume.
//
static void __computeCostOfNode(GXBBox *rBound, GFloat& rCost) {
    GXObjectList *tList = rBound->getListOfChildren();
    rCost += tList->count() * rBound->getCost();
   
    for (GXObjectItr _Itr(*tList); _Itr.more(); _Itr.next()) {
	GXObject *tObj =  _Itr.cur();
	if(tObj->isBound())
	    __computeCostOfNode( (GXBBox *)tObj, rCost);
    }
}

class SortNode { 
 public:
    GXObject *mObj;
    GFloat    mCost;
};
int __compareNodes(const void *p0,const void *p1 )  {
    SortNode *tNode0  = (SortNode *) p0;
    SortNode *tNode1  = (SortNode *) p1;
    
    if ( tNode0->mCost == tNode1->mCost )  
	return(0);
    else 
	return( tNode0->mCost < tNode1->mCost ? -1 : 1 ) ;
}


// 
// Sort a list of GXObjects. The sorting criteria is the intersection
// cost of the objects. 
//
static void __sortList(GXObjectList *rList) {
    long tNum = rList->count();
    SortNode *tNodes = new SortNode[tNum];
    
    for(register int i = 0; i < tNum; i++) {
	GXObject *tObj  = (*rList)[i];
	tNodes[i].mObj  = tObj;
	tNodes[i].mCost = 0.0;
       
	if(tObj->isBound())
	    __computeCostOfNode((GXBBox *)tObj, tNodes[i].mCost);
	else
	    tNodes[i].mCost = tObj->getCost();
    }
    
    qsort( tNodes, tNum, sizeof( SortNode ), __compareNodes) ;
   
    for(i = 0; i < tNum; i++) 
	(*rList)[i] = tNodes[i].mObj;
   
    delete [] tNodes;
}

static void __optimizeNode(GXBBox *rBound) {
    GXObjectList *tList = rBound->getListOfChildren();
    
    __sortList(tList);
   
    for (GXObjectItr _Itr(*tList); 
	 _Itr.more(); 
	 _Itr.next())
      {
	  GXObject *tObj =  _Itr.cur();
	  if(tObj->isBound())
	    __optimizeNode( (GXBBox *)tObj);
      }
}


// 
// Calculate the cost of a BV-hierarchy.  
//
static GFloat __computeCostOfHierarchy(const GXObjectList& pHier) {
    
    // If root is empty or its head isnt a bound, then exit! 
    if (pHier.count() == 0)
	return 0;
   
    GXObject *tRoot = pHier[0];
    if (! tRoot->isBound())
	return 0;
   
    GFloat tRet = 0;
    __computeCostOfNode((GXBBox *)tRoot, tRet);
   
    tRet = tRet / ((GXBBox *)tRoot)->getCost();
   
   return tRet;
}



//
// Find the best methode to connect a new object to a local position
// in the BVH.
//
static int __findLocalMethod(GXObject*       pOldParent,
			     const GBounds3&   pNewChildBound,
			     GFloat&         rCost,
			     const ListStk&  pStk) 
{

    int    tNum = pOldParent->getNumOfChildren();
    GBounds3 tOldParentBound(pOldParent->getWorldBounds());
   
   
    //////////////////////////////////////////////
    // First thing: calculate the local cost!
    //////////////////////////////////////////////
   
    // Calculate the cost for the new parent bound 
    GFloat tCost_NewParentBound = 
	(tOldParentBound | pNewChildBound).getSurfaceArea();
    
    /* 
     * Methode 0:
     * ------------------------------------------------
     * Add the new object direct to the parent bbox,
     * wich may increase in size.
     * 
     * New_Parent_Bound
     * |                
     * +-----------------------+<---------------------+
     * |                       |                      |
     * Old_Child[0] ... Old_Child[tNum-1]            New_Child
     * 
     */
    GFloat tCost_M0 = tCost_NewParentBound * (tNum + 1);
    
    /* 
     * Methode 1: 
     * --------------------------------------------------
     * Create a new bound, wich contains the old bound 
     * and the new object.
     * 
     *                      New_Parent_Bound
     *                              |
     *             +----------------+<----------------+
     *             |                                  |
     *      Old_Parent_Bound                       New_Child
     *             |
     * +-----------+-----------+
     * |                       |
     * Old_Child[0] ... Old_Child[tNum-1]
     * 
     */
    GFloat tCost_M1 = 
	2 * tCost_NewParentBound + tNum * tOldParentBound.getSurfaceArea();
   
    int tBestMethode;
    if (tCost_M0 < tCost_M1)  {
	rCost = tCost_M0;
	tBestMethode = 0;
    }
    else  {
	rCost = tCost_M1;
	tBestMethode = 1;
    }
   
    ////////////////////////////////////////////////
    // OK, time to find the inheritance cost... 
    ////////////////////////////////////////////////
   
    // Dont destroy my STACK Plzzzzz! 
    ListStk tStkUp(pStk);
   
    while (!tStkUp.empty())  {
	StkNode tStkNode = tStkUp.pop();
      
	GXObject *tObj = GET_OBJ(tStkNode);
	tNum = tObj->getNumOfChildren();
	rCost += 
	    tNum * (tObj->getWorldBounds() | pNewChildBound).getSurfaceArea();
    }
    return tBestMethode;
}
 
//
// Find in the list for the best object to combine w/ the new object  
//
static  long __findBestLocalObj(GXObjectList*  pTestList, 
				const GBounds3&  pNewBound)
{
    GFloat  tMinCost = MAXFLT;
    long tRet = 0, tIndex = 0;
   
    for(GXObjectItr _Itr(*pTestList);
	_Itr.more(); _Itr.next(), tIndex++)
    {
	GFloat tCost = 
	    (_Itr.cur()->getWorldBounds()| pNewBound).getSurfaceArea();
      
	if (tCost < tMinCost)  {
	    tMinCost = tCost;
	    tRet = tIndex;
	}
    }
    
    return tRet;
}


static void __connectObject(GXObjectList* rList, 
			    GXObject*     pNewObj) 
{
    StkNode        tStkNode;
    static ListStk sSTK;
   
    GBounds3      tNewChildBound(pNewObj->getWorldBounds());
    GXObjectList* tCurList = rList;
    GXObject*     tObj;
   
    ///////////////////////////////////////////////////////////// 
    // Make a DOWN-path by searching in the hierarchy for 
    // smallest bounds when adding the new object to them.
    /////////////////////////////////////////////////////////////
    for(;;)  {
	tStkNode.mN    = __findBestLocalObj(tCurList, tNewChildBound);
	tStkNode.mList = tCurList;
	sSTK.push(tStkNode);
      
	tObj = GET_OBJ(tStkNode);
	if (tObj->isBound())
	    tCurList = ((GXBBox *)tObj)->getListOfChildren();
	else
	    break;
    } 
   
   
    //////////////////////////////////////////////////////
    // Traverse the path, UP to the root. For each node 
    // in the path, calculate the new intersection cost
    // when the new object is connected to it. The best
    // node and methode will be saved.
    //////////////////////////////////////////////////////
    
   
    // backup the stack 
    ListStk tStkUp(sSTK);

    int tBestMethod, tMethod;
    GFloat tMinCost, tCost;
   
    StkNode tMinNode = sSTK.pop();
   
    tObj = GET_OBJ(tMinNode);
    tBestMethod = __findLocalMethod(tObj, 
				    tNewChildBound, 
				    tMinCost, 
				    sSTK);
    while (! sSTK.empty())  {
	
	tStkNode = sSTK.pop();
      
	tObj = GET_OBJ(tStkNode);
	tMethod = __findLocalMethod(tObj, 
				    tNewChildBound, 
				    tCost, 
				    sSTK);
      
	if (tCost < tMinCost)  {
	    tMinNode = tStkNode;
	    tMinCost = tCost;
	    tBestMethod = tMethod;
	 
	    // reconfigure the old stack pointer to expand the
	    // bounds of the parents
	    while(tStkUp.top().mList != tStkNode.mList)
		tStkUp.delTop();
	}
    }
      
    tObj = GET_OBJ(tMinNode);

    ///// METHODE 0
    if (!tBestMethod && tObj->isBound())  {
	((GXBBox *)tObj)->add(pNewObj);
    }
    ///// METHODE 1
    else  {
	GXBBox *tNewBound = new GXBBox();
	tNewBound->add(tObj);
	tNewBound->add(pNewObj);
	(*(tMinNode.mList))[tMinNode.mN] =  tNewBound;
    }
   
   
    //////////////////////////////////////////////////////
    // expand the parents bounds to fit the new object 
    //////////////////////////////////////////////////////
    while (! tStkUp.empty())  {
	tStkNode = tStkUp.pop();
	tObj = GET_OBJ(tStkNode);
	if (tObj->isBound())  
	    *((GXBBox *)tObj) += tNewChildBound;
    }
}



static void __buildHierarchy(GXObjectList*       rOptiList, 
			     const GXObjectList& pInputList) 
{
    GXObjectItr _Itr(pInputList);
   
    // remember, item 0 is the root box! 
    rOptiList->append(_Itr.cur());
   
    for(_Itr.next(); _Itr.more(); _Itr.next()) 
	__connectObject(rOptiList, _Itr.cur());
}


//
// Shuffle a list of objects to ensure a better tree
// After G&S.
//
static void __shuffleList(GXObjectList& rList) {
    long tNum = rList.count();
    for(register int i = 0; i < tNum; i++) {
	int tR = rand() % tNum;
	rList.swap(i, tR);
    }
}

void GXBBox::buildHierarchy(void) {
   
    // We dont need to make the BVH when this bound containts
    // only one object!
    if (mListChildren.count() <= 1)
	return;

    if(!GXOptions::ABVHFlag) {
	
	GXObject *tBox = new GXBBox();
	for (GXObjectItr _Itr(mListChildren); _Itr.more(); _Itr.next()) {
	    ((GXBBox *)tBox)->add(_Itr.cur());
	}
	mListChildren.reset();
	mListChildren.append(tBox);
    
    }
    else {
	
    
	// Make a copy for the hierarchy 
	GXObjectList tTestList; 
	{
	    for (GXObjectItr _Itr(mListChildren); 
		 _Itr.more(); _Itr.next())
		tTestList.append(_Itr.cur());
	}
   
	GFloat tBestCost = MAXFLT;
	
	// @WARNING@ : 
	int tNumTrees = 9; 
	//(int)ceil(log(mListChildren.count())/log(2));
	
	int i;
	for (i = 0; i < tNumTrees; i++)  {
      
	    // randomize the list...
	    __shuffleList(tTestList);
      
	    // build the BVH...
	    GXObjectList tOptiList;
	    __buildHierarchy(&tOptiList, tTestList);
      
	    // Is the cost for this BVH better than the actual 
	    // best cost?
	    GFloat tTestCost = __computeCostOfHierarchy(tOptiList);
	    if (tTestCost < tBestCost)  {
		tBestCost = tTestCost;
	 
		// update the BVH for this node 
		mListChildren.free();
		{	 
		    for (GXObjectItr _Itr(tOptiList);
			 _Itr.more(); _Itr.next()) 
			mListChildren.append(_Itr.cur());
		}
	    }
	    
	}
	// @DEBUG@
	cerr << "BVH\t" << ((GXBBox *)mListChildren[0])->mListChildren.count() 
	     << "\telements\t" << "best-cost\t" << tBestCost << "\n";  
	
    }
    
    // Sort the list of children 
    __optimizeNode((GXBBox *)(mListChildren[0]));
}



bool GXBBox::intersectShading(GRay3& rRay, GInterSectionList&  rInter)  
{
    // KayKajia heap
    static  __KKHeap  sKK_HEAP;

    // last nearest caching
    static  GXObject*  sLastNearest = 0;  
   
    GXObject *tObj;
    GFloat  tTmin, tTmax, tMinT;
    bool tHIT = false;

    // opps, nothing to intersect with..

      
    // get the root-node
    tObj = mListChildren[0];
      
    rRay.setTmin(gEpsilon);
    rRay.setTmax(HUGE);

    
    // bound
    if (tObj->isBound()) {
	if(((GXBBox *)tObj)->findMinMax(rRay, tTmin, tTmax)) {
	    // update min/max intervall
	    rRay.setTmin(tTmin);
	    rRay.setTmax(tTmax);
	    
	    sKK_HEAP.insert(KKNode(tObj, tTmin));
	    tMinT = tTmax;
	} else
	  goto L_EXIT;
    }
    // primitive
    else if(tObj->intersect(rRay, rInter)){
	tMinT = rInter.getFirst().getT();
	sLastNearest = tObj;
	tHIT = true;
    }
    
    while (sKK_HEAP.count())  {
      
	KKNode tNode = sKK_HEAP.remove();
      
	// There is no need for intersecting w/ the rest of bounds
	// when the nearest object was found.
	if ((tNode.mT - tMinT) > gEpsilon )  
	    break;             

	GXObjectList *tList = 
	    ((GXBBox *)tNode.mObj)->getListOfChildren();
      
	for(GXObjectItr _Itr(*tList); _Itr.more(); _Itr.next()) {
	    
	    tObj = _Itr.cur();
	   
	    // If the object is a bound and it intersects
	    // the ray, insert it to the KK-Heap w/ _tmin
	    // as the key.
	      
	    // bound
	    if (tObj->isBound()) {
		if(((GXBBox*)tObj)->findMinMax(rRay, tTmin, tTmax)) 
		    sKK_HEAP.insert(KKNode(tObj, tTmin));
	    }
	    // primitive
	    else  {
		if(!tObj->checkMailBox(tHIT, rRay, rInter) &&
		   tObj->intersect(rRay, rInter)) 
		{
		    tHIT = true;
		    tMinT = rInter.getFirst().getT();
		    sLastNearest = tObj;
		}
	    }
	}
    }
   
    // L_FREE_HEAP:
    sKK_HEAP.reset();
   
    L_EXIT:
    return tHIT;
}



bool GXBBox::intersectShadow(const GRay3& pRay, GInterSectionList &rInter) {

    static GStack<GXObjectList *> sSTK;
    static GXObjectList sTransList;
    
    bool tHIT = false;
    GFloat  tTmin, tTmax;
      
    // get the root-node
    GXObject *tObj = mListChildren[0];
    
    // bound
    if (tObj->isBound()) {
	if(((GXBBox *)tObj)->findMinMax(pRay, tTmin, tTmax))
	    sSTK.push(&(((GXBBox *)tObj)->mListChildren));
	else
	    return (false);
    }
    
    // primitive
    else if(tObj->intersect(pRay, rInter) && !tObj->getTransmission())
	return (true);

    while ( ! sSTK.empty() && !tHIT)  {
	
	for(GXObjectItr _Itr(* sSTK.pop()); 
	    !tHIT && _Itr.more(); _Itr.next()) 
	{
	    
	    tObj = _Itr.cur();

	    // bound
	    if (tObj->isBound()) {
		if(((GXBBox*)tObj)->findMinMax(pRay, tTmin, tTmax)) 
		    sSTK.push(&(((GXBBox *)tObj)->mListChildren));
	    }
	    // primitive
	    else  {
		if(!tObj->checkID(pRay)) {
		    if(tObj->getTransmission()) {
			sTransList.append(tObj);
			continue;
		    }
	
		    tHIT = tObj->intersect(pRay, rInter);
		    tObj->updateID(pRay);
		}
		else
		    gstat(GXStatistics::MailBoxHits++);
	    }
	} // for
    } // while
   
    
    // There was no opaque-object which occludes the light.
    // Begin to intersect w/ transparent object to find the
    // max. attenuation.
    if(!tHIT) {
	GOneInterSectionList tInter;

	for(GXObjectItr Itr(sTransList); Itr.more(); Itr.next()) {
	    GXObject *tPrim = Itr.cur();
	
	    tHIT = tPrim->intersect(pRay, tInter);
	    tPrim->updateID(pRay);
	}
	
	if(tHIT) { 
	    rInter.updateTransmission(tInter.getTransmission());
	    tHIT = false;
	}
    }
    
    sSTK.reset();
    sTransList.reset();
    return (tHIT);


}


// Dtor
GXBBox::~GXBBox(void) 
{
    for(GXObjectItr _Itr(mListChildren);
	_Itr.more(); _Itr.next()) {
	GXObject *tObj = _Itr.cur();
	if(tObj->isBound())
	  delete tObj;
    }
}

 

