//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <stdlib.h>
#include <fstream.h>

#include "input/gtime.h"

#include "gx/gx_cam.h"



GXRayTracer::~GXRayTracer(void) {
    gkillPointer(mPrimaryAccel);
    gkillPointer(mShadowAccel);
    gkillPointer(mSecondaryAccel);
}


/*
 * %FUNCT: GXRayTracer::trace_ShadowRay
 * 
 * %DESCR: trace a shadow ray and return the light attenuation.
 *         If an opaque object was found, save it in pLastOpaque
 *         for the shadow cache.
 * 
 * %NOTES: Many optimizations (and simplifications, too:) are impl.
 *         here: We save all transparent objects in a list for later
 *         testing. For the worse case, if no opaque object was found, 
 *         check this list and multiple all attenuation factors of 
 *         transparent-objects along the ray. This simplification is 
 *         neccessary for shadow-calculation, because refraction 
 *         calculation is expensive for this task. 
 * 	   [Shadow-computation in the future: caustic! I'm working
 *         on it...]
 */
GColorRGB GXRayTracer::traceShadowRay(GRay3&          rRay,
				      GRayTraceable*& rLastOpaque,
				      GXAccelCell*& rLastVoxel)
{
    GOneInterSectionList tInterList;
    
    if (mShadowAccel->intersectShadow(rRay, tInterList, 
				      rLastOpaque, rLastVoxel)) {
	
	
	return (GColorRGB(0, 0, 0));
    }
    else {  
	
	return (GColorRGB(1.0, 1.0, 1.0) * tInterList.getTransmission());
    }
    
}


GColorRGB GXRayTracer::traceShadingRay(int       pLevel, 
				       GFloat    pWeight, 
				       GRay3&    rRay)
{
    GOneInterSectionList tInterList;
    GColorRGB tColor;

    GXAccelMethode *tAccel = (rRay.isEyeRay()) ? mPrimaryAccel : mSecondaryAccel;
    

    // find the nearest intersection, if anyone... 
    if (tAccel->intersectShading(rRay, tInterList))  {

	gstat(GXStatistics::ShadingHits++);

	// shade the nearest object 
	GInterSection tInter = tInterList.getFirst();
	tColor = shade(pLevel,  pWeight, rRay, tInter);
    }
    else 
	tColor = getBackground();
   
    // clip color to [0,1]:
    tColor.correct();
    
    return tColor;
}

GColorRGB GXRayTracer::trace(int    pLevel, 
			     GFloat pWeight, 
			     GRay3& rRay)
{
    return(traceShadingRay(pLevel, pWeight, rRay)); 
}
 

GColorRGB GXRayTracer::shade(int    pLevel, 
			     GFloat pWeight, 
			     GRay3& rRay, 
			     const GInterSection& pInter)
{

    GVector3       tPoint(rRay.extrap(pInter.getT()));
    GRayTraceable& tPrim = pInter.getPrimitive();
    GSurface       tSurf(tPrim.getSurface(tPoint));
    GVector3       tNormal(tPrim.getNormal(tPoint));
      
   
    GVector3 tView(rRay.getVector());
    
    // negate the normal if necessary
    if(tNormal.dot(tView) > 0)
	tNormal = tNormal.negate();
    
    // compute the reflection\'s vector
    GVector3 tReflect(tView + tNormal * -2 * tNormal.dot(tView));
   
    GXShadingParams tShadeP(tPoint, 
			    tNormal, 
			    GVector3(0, 0, 0), 
			    tReflect, 
			    tView.negate(),  
			    pInter.enter(), 
			    &tSurf, 
			    &tPrim, 
			    this);
   
    GColorRGB tColor(0, 0, 0);
    //GColorRGB tColor = tSurf.getAmbient();
   

    //// DIFFUSE && SPECULAR 
    
    // compute all contributions for direct illumination 
    for (GXLightItr _Itr(mLightList); _Itr.more(); _Itr.next())  {
	tColor += _Itr.cur()->shade(tShadeP);
    }

   
    if(pLevel + 1 >= gMaxRecursion)
	goto L_EXIT;


    
    
    //// REFLECTION
    
    // recurse on reflection ray if significant 
    if (tSurf.getSpecular().abs() * pWeight > gMinWeight)  {
	// construct a refl. ray...
	GRay3 tReflRay(tPoint, tReflect, GRay3::G_REFLECT);
	gstat(GXStatistics::ReflectionRays++);
	
	// Planar-objects cannot intersect the ray 2times.
	// Also, if the prim. is convex and the ray was going into it, then
	// the reflection ray can not intersect it again. 
	if(tPrim.isPlanar() || (tPrim.isConvex() && pInter.enter()))
	    tPrim.updateID(tReflRay);
	
	// and let our camera do the hard job: Trace me this ray!
	tColor += tSurf.getSpecular() * traceShadingRay(pLevel + 1, 
							tSurf.getSpecular().abs() * pWeight, 
							tReflRay);
    }
   
    //// TRANSMISSION
    
    if (tSurf.getTransparency() * pWeight > gMinWeight)  {
	
	// cos M1
	GFloat tCos1 = -tNormal.dot(tView);

	// comming from air or going into air ? 
	GFloat tGamma  = tSurf.getRefraction();
	if (pInter.enter()) 
	    tGamma = 1 / tGamma;
	
	// cos M2
	GFloat tCos2_2 = 1.0 - tGamma * tGamma * (1.0 - tCos1 * tCos1);
	
	// check for total internal refraction 
	if (tCos2_2 > 0)  {
	    // construct the refracted ray
	    GRay3 tTransRay(tPoint, 
			    tNormal * (tGamma * tCos1 - sqrt(tCos2_2)) + tView * tGamma,
			    GRay3::G_REFRACT, pLevel);
	    gstat(GXStatistics::RefractionRays++);
	    
	    // Planar-objects cannot intersect the ray 2times.
	    // Also, if the prim. is convex and the ray was going out from it, then
	    // the refraction ray can not intersect it again. 
	    if(tPrim.isPlanar() || (tPrim.isConvex() && !pInter.enter()))
		tPrim.updateID(tTransRay);
	
	    // get the contribution of transmission
	    tColor += traceShadingRay(pLevel + 1, tSurf.getTransparency() * pWeight, 
				      tTransRay) * tSurf.getTransparency();
	}
    }
   
    L_EXIT:
    return tColor;
}


extern void sortObjects(const GXPrimitiveList& pScene, GXObjectList *rObjLists[6]);

////
//// @GXRayTracer@rendering&
////
void GXRayTracer::rendering(const GXPrimitiveList& pScene)  {
    
    ///// Preparing...
    GTime tPreTime;

    // build the list of lights 
    mLightList.reset();
    class GXConvertLightsIter: public GIterator<GXLight>  {
	GXLightList *mLights;
     public:
	GXConvertLightsIter( GXLightList *pLights)  {
	    mLights = pLights; 
	}
	bool exec(GXLight& pL)  {
	    mLights->append(&pL);
	    return (false);
	}
    } tIter( &mLightList );
    gGXLights->doWithElements(tIter );
   
    //// sort the scene
    GXObjectList *tObjLists[6];
    sortObjects(pScene, tObjLists);
    
    //// build the lightbuffers 
    for (GXLightItr _Itr(mLightList); _Itr.more(); _Itr.next())  {
	_Itr.cur()->preprocess(tObjLists);
    }


    //// acceleration scheme selection
   

    cout << "Number of primitives:\t" << pScene.count() <<"\n";
    
    GXAccelMethode *tBSPAccel=0, *t5DAccel=0;
    
    if(GXOptions::PrimaryAcceleration      == GX_BSP 
       || GXOptions::SecondaryAcceleration == GX_BSP 
       || GXOptions::ShadowAcceleration    == GX_BSP)
    {
	tBSPAccel = new GX_BSP_Tree(pScene);

	((GX_BSP_Tree *)tBSPAccel)->buildHierarchy();
    }
    
    if(GXOptions::PrimaryAcceleration      == GX_HOT 
       || GXOptions::SecondaryAcceleration == GX_HOT 
       || GXOptions::ShadowAcceleration    == GX_HOT)
    {
	t5DAccel = new GXHyperOctree(tObjLists, getViewpoint());
    }
    
    if(GXOptions::PrimaryAcceleration == GX_BUF)
	mPrimaryAccel = new GXViewBuffer(tObjLists, 
					 GBounds3(getViewpoint() - GVector3(gEpsilon),
						  getViewpoint() + GVector3(gEpsilon)));
    else
    	mPrimaryAccel = (GXOptions::PrimaryAcceleration == GX_BSP) ? 
	    tBSPAccel : t5DAccel;
    
    mSecondaryAccel = (GXOptions::SecondaryAcceleration == GX_BSP) ? 
	tBSPAccel : t5DAccel;
    
    mShadowAccel = (GXOptions::ShadowAcceleration == GX_BSP) ? 
	tBSPAccel : t5DAccel;
    
    cout << "Preparing time:\t" << GTime::getTime() - tPreTime << " secs.\n";
    

    
    ///// rendering...
    GTime tRenderTime;
    
    if(GXOptions::DisplayFlag) {
	mRealization->activate();
	mRealization->clear(GColorRGB(0, 0, 0));
	mRealization->swapBuffers();
    }

    int pw = GXOptions::PixWidth;
    int ph = GXOptions::PixHeight;
    
    
    GFloat wc = 2.0/(pw - 1);
    GFloat hc = 2.0/(ph - 1);
    
    GFloat    _xx, _yy;
	
    GFloat tAspect = (GFloat)pw / (GFloat)ph;
    GFloat tViewDistance  = 1/tan(getZoom()/2.0 * M_PI/180.0);

    GColorRGB tColor(0, 0, 0);
    
    GVector2 *tSubPix = new GVector2[GXOptions::MaxSamples];

    GX_PoissonDiskSampling tSampling;
    GMatrix4 tWXForm = getMatrix();
    register int i;
    register int j;
    for (j = ph-1, _yy = 1.0; j >= 0; j--, _yy -= hc) {
	
	// Flip the direction of scanline 
	// This tricky methode is very efficient for
	// caching ( applied image-coherence!)
	int tIStart, tIInc, tIEnd;
	GFloat tXStart, tXInc;
	if(j%2) {
	    tIStart = 0; tIInc = 1; tIEnd = pw;
	    tXStart = 1.0; tXInc = -wc;
	}
	else {
	    tIStart = pw-1; tIInc = -1; tIEnd = -1;
	    tXStart = -1.0; tXInc = wc;
	}
	
	for (i = tIStart, _xx = tXStart; 
	     i != tIEnd; 
	     i+=tIInc, _xx += tXInc) {
	    
	    
	    tColor = GColorRGB(0.0, 0.0, 0.0);
	    
	    //// adaptive supersampling
	    if(GXOptions::Antialiasing) {
		GVector3 tImin(MAXFLT, MAXFLT, MAXFLT);
		GVector3 tImax(-MAXFLT, -MAXFLT, -MAXFLT);

		tSampling.generateSamples();
		for(register int k = 0; k < tSampling.getNumber(); k++) {

		    GPoint2 tSample = tSampling.getSample(k);
		    GPoint3 tTarget((_xx + wc*tSample.getX()) * tAspect, 
				    _yy + hc*tSample.getY(), 
				    tViewDistance);
		    
		    GRay3  tRay(getViewpoint(), 
				( tWXForm * tTarget - getViewpoint()).unitize(), 
				GRay3::G_EYE);
		    gstat(GXStatistics::EyeRays++);
		
		    GColorRGB tI = traceShadingRay( 0, 1, tRay);
		    tColor += tI;
		
		    GVector3 tAct(tI.getR(), tI.getG(), tI.getB());
		    tImin.setMin(tAct);
		    tImax.setMax(tAct);
		    if(k >= tSampling.getMinS()) {
			GVector3 tContrast = (tImax - tImin) / (tImax + tImin);
			if(tContrast.getX() < GXOptions::ContrastThreshold.getX() &&
			   tContrast.getY() < GXOptions::ContrastThreshold.getY() &&
			   tContrast.getZ() < GXOptions::ContrastThreshold.getZ() )
			  {
			      k++;
			      break;
			  }
			
		    }
		}
		tColor /= k;
		tColor.correct();
	    }
	    //// one sample per pix 
	    else {
		
		// pix-coord. in screen space
		GPoint3 tTarget(_xx * tAspect, _yy, tViewDistance);
		// transform the pix to world-space and contruct an 
		// EYE-ray
		GRay3  tRay(getViewpoint(), 
			    (getMatrix() * tTarget - getViewpoint()).unitize(), 
			    GRay3::G_EYE);
	    
		gstat(GXStatistics::EyeRays++);
		tColor = traceShadingRay( 0, 1, tRay);
	    
	    
	    }
	    
	    if(GXOptions::DisplayFlag)
	      mRealization->putPixel(i, j, tColor); 
		
	    if(GXOptions::OutputFlag) 
	      mOutputImage->setColor(i, j, tColor);
	    
	} // row
	  
	  //// etc.
	  if(GXOptions::VerboseFlag)
	  fprintf(stderr, "Lines remaining:%4d\r", j );
    } // column
    
    delete[] tSubPix;
    gkillPointer(mPrimaryAccel);
    gkillPointer(mShadowAccel);
    gkillPointer(mSecondaryAccel);
    
 
    
    if(GXOptions::VerboseFlag)
      cerr << "\nDone.\n";
    
    if(GXOptions::OutputFlag) { 	
	if(mOutputImage) { 
	    if(GXOptions::VerboseFlag)
	      cerr << "Write to file " << GXOptions::OutFileName << ".\n";
	    mOutputImage->write();
	    if(GXOptions::VerboseFlag)
	      cerr << "\nDone.\n";
	}
    }
    
    
    cout << "-------------+---------------------------------------\n";
    cout << "Tracing time |\t" << GTime::getTime() - tRenderTime << " secs.\n";
    cout << "-------------+---------------------------------------\n";

#ifdef GSTAT 
    
    cout << "Ray Type     | Number of rays\n";
    cout << "-------------+---------------------------------------\n";
    cout << "Eye          |\t" << GXStatistics::EyeRays << "\n";
    cout << "Reflection   |\t" << GXStatistics::ReflectionRays << "\n";
    cout << "Refraction   |\t" << GXStatistics::RefractionRays << "\n";
    cout << "Shadow       |\t" << GXStatistics::ShadowRays << "\n";
    cout << "-------------+---------------------------------------\n";
    cout << "Cache Hits(%)|\tObject\tVoxel\n";
    cout << "-------------+---------------------------------------\n";
    cout << "Shadow       |" 
	 << "\t" << (GFloat)GXStatistics::ShadowLastObjectHits * 100.0 
	 / (GFloat)GXStatistics::ShadowHits 
	 << "\t" << GXStatistics::ShadowLastVoxelHits * 100.0
	 / (GFloat)GXStatistics::ShadowHits << "\n";
    
    cout << "Shading      |" 
	 << "\t" << (GFloat)GXStatistics::ShadingLastObjectHits * 100 
	 / (GFloat) (GXStatistics::ShadingHits)
	 << "\t" << GXStatistics::ShadingLastVoxelHits * 100
	 / (GFloat) (GXStatistics::ShadingHits)
	 << "\n";;
    
    cout << "-------------+---------------------------------------\n";
    cout << " Mailbox     |\t" << GXStatistics::MailBoxHits << " hits.\n"; 
    
    cout << "-------------+---------------------------------------\n";
    cout << " Object-Type |\tI-Test\tHit\tTime\n";
    cout << "-------------+---------------------------------------\n";
    cout << "BBoxes       |\t" << GXStatistics::BBox.mTests << "\t"
      << GXStatistics::BBox.mHits << "\t" 
      << GXStatistics::BBox.mTime << "\n";
    cout << "Spheres      |\t" << GXStatistics::Sphere.mTests << "\t"
      << GXStatistics::Sphere.mHits << "\t" 
      << GXStatistics::Sphere.mTime << "\n";

    cout << "Polygons     |\t" << GXStatistics::Polygon.mTests << "\t"
      << GXStatistics::Polygon.mHits << "\t" 
      << GXStatistics::Polygon.mTime << "\n";

    cout << "Cones        |\t" << GXStatistics::Cone.mTests << "\t"
      << GXStatistics::Cone.mHits << "\t" 
      << GXStatistics::Cone.mTime << "\n";
    cout << "-------------+---------------------------------------\n";
 
  
    cout << "-----------------------------------------------------\n";
    cout << "Subdivision statistics\n";
    cout << "-----------------------------------------------------\n";
   
    cout << "Objects/Hyper-Cube";
    if(GXStatistics::HOT_NumShadingVoxels )
	cout << "\tShading: " 
	     << (GFloat)GXStatistics::HOT_NumObjsInShadingVoxels 
	     / (GFloat)GXStatistics::HOT_NumShadingVoxels ;
    if(GXStatistics::HOT_NumShadowVoxels)
	cout << "\tShadow : "
	     << (GFloat)GXStatistics::HOT_NumObjsInShadowVoxels 
	     / (GFloat)GXStatistics::HOT_NumShadowVoxels ;
    
    cout << "\n";

    cout << "Objects/BSP-Voxel ";
    if(GXStatistics::BSP_NumShadingVoxels )
	cout << "\tShading: " 
	     << (GFloat)GXStatistics::BSP_NumObjsInShadingVoxels 
	     / (GFloat)GXStatistics::BSP_NumShadingVoxels ;
    if(GXStatistics::BSP_NumShadowVoxels)
	cout << "\tShadow : "
	     << (GFloat)GXStatistics::BSP_NumObjsInShadowVoxels 
	     / (GFloat)GXStatistics::BSP_NumShadowVoxels ;
    cout << "\n";

    cout << "BSP-Voxels/shading-ray\t" 
	 << (GFloat)GXStatistics::BSP_NumShadingVoxels  
	 / (GFloat) (GXStatistics::EyeRays + 
		     GXStatistics::ReflectionRays + 
		     GXStatistics::RefractionRays) 
	 << "\n";
    cout << "BSP-Voxels/shadow-ray \t" 
	 << (GFloat)GXStatistics::BSP_NumShadowVoxels 
	 / (GFloat) GXStatistics::ShadowRays 
	 << "\n";
#endif

}
 



