//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "render/graytrac.h"
#include "gx/gx_csg.h"

//////////////////////////////////////////////////////////////////////////////
// @GX_CSG_UNION@
//////////////////////////////////////////////////////////////////////////////

GX_CSG_Union::GX_CSG_Union(const GXObjectList& pObjList) 
: GXPrimitive()
{
    for(GXObjectItr _Itr(pObjList); _Itr.more(); _Itr.next())
      mChildrenList.append(_Itr.cur());
        
    // compute a bounding box for BVH  
    computeBounds();
}

GX_CSG_Union::~GX_CSG_Union(void) {
    ;;;      
}
    

const GBounds3&  GX_CSG_Union::computeBounds(void) {
    setWorldBounds() = GBounds3(GPoint3(MAXFLT, MAXFLT, MAXFLT),
			   GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));

    if(mChildrenList.count()) 
      for(GXObjectItr _Itr(mChildrenList); _Itr.more(); _Itr.next()) {
	  // Union
	  setWorldBounds() += (_Itr.cur()->getWorldBounds());
      }
    
    return(getWorldBounds());
}

GVector3 GX_CSG_Union::getNormal(const GPoint3& pPoint) {
    gstat( cerr << "CSG_Union::getNormal called!\n" );
    return (GVector3(0, 0, 0));
}


// If CX_CSG_Union is derived from GXBBox, one can use 
// GXBBox::buildHierachy() and GXBBox::intersect() to 
// optimize this routine. 
bool GX_CSG_Union::intersect(const GRay3& pRay, 
			     GInterSectionList& rInter) 
{
    bool tHIT = false;
    GFloat tMin, tMax;
    if(!getWorldBounds().findMinMax(pRay, tMin, tMax))
      return (tHIT);
    
    // ugly and slow!
    for(GXObjectItr _Itr(mChildrenList); _Itr.more(); _Itr.next()) {
	if(_Itr.cur()->intersect(pRay, rInter))
	  tHIT = true;
    }
    
    return (tHIT);
}



//////////////////////////////////////////////////////////////////////////////
// @GX_CSG_Intersection@
//////////////////////////////////////////////////////////////////////////////

GX_CSG_Intersection::GX_CSG_Intersection(const GXObjectList& pObjList) 
: GXPrimitive(GRayTraceable::CSG)
{
    mLastChild = 0;
    for(GXObjectItr _Itr(pObjList); _Itr.more(); _Itr.next())
      mChildrenList.append(_Itr.cur());
        
    // compute a bounding box for BVH  
    computeBounds();
}

GX_CSG_Intersection::~GX_CSG_Intersection(void) {
    ;;;      
}
    
bool GX_CSG_Intersection::intersect(const GRay3& pRay, 
				    GInterSectionList& rInter) 
{
    static GAllInterSectionList sIHeap;
    bool tHIT = false;
    GFloat tMin, tMax;
    if(!getWorldBounds().findMinMax(pRay, tMin, tMax))
      return (false);
    
    GRay3 tRay(pRay.getPoint(), pRay.getVector());
    tRay.setTmin(-MAXFLT);
    tRay.setTmax(MAXFLT);
    
    int tNum = mChildrenList.count();
    bool tInside;
    register int tIntervalCount = 0;
    mLastChild = 0;
    
    for(GXObjectItr _Itr(mChildrenList); _Itr.more(); _Itr.next()) {
	if(!_Itr.cur()->intersect(tRay, sIHeap))
	  goto L_EXIT;
    }
    
    while(sIHeap.count()) {
	GInterSection tI = sIHeap.remove();
	
	tInside = (tI.enter()) ? 
	  (++tIntervalCount == tNum) : (--tIntervalCount == tNum -1);
	
	// a very bad example for aggreate objects! 
	if( tInside && pRay.isValid(tI.getT())) {
	    rInter.add(GInterSection(self, tI.getT(), tI.enter()));
	    mLastChild = &(tI.getPrimitive());
	    tHIT = true;
	    break;
	}
	
    }
    
    
    L_EXIT:
    
    sIHeap.reset();
    return (tHIT);
}


GVector3 GX_CSG_Intersection::getNormal(const GPoint3& pPoint) {
    gcheck(!mLastChild, "CSG_Intersection::getNormal() w/o mLastChild called!");
    return (mLastChild->getNormal(pPoint));
}

const GSurface& GX_CSG_Intersection::getSurface(const GPoint3& pPoint) {
    gcheck(!mLastChild, "CSG_Intersection::getSurface() w/o mLastChild called!");
    return (mLastChild->getSurface(pPoint));
}
    
const GBounds3& GX_CSG_Intersection::computeBounds(void) {
    
    setWorldBounds() = GBounds3(GPoint3(-MAXFLT, -MAXFLT, -MAXFLT),
			   GPoint3(MAXFLT, MAXFLT, MAXFLT));
    
    for(GXObjectItr _Itr(mChildrenList); _Itr.more(); _Itr.next()) {
	setWorldBounds() = getWorldBounds() & (_Itr.cur()->getWorldBounds());
    }
    return (getWorldBounds());
}


//////////////////////////////////////////////////////////////////////////////
// @GX_CSG_Difference@
//////////////////////////////////////////////////////////////////////////////


GX_CSG_Difference::GX_CSG_Difference(const GXObjectList& pObjList) 
: GXPrimitive(GRayTraceable::CSG)
{
    mLastChild = 0;
    gcheck(pObjList.count() < 2, "GX_CSG_Difference w/ less than 2 children!");

    for(GXObjectItr _Itr(pObjList); _Itr.more(); _Itr.next())
      mChildrenList.append(_Itr.cur());
        
    // compute a bounding box for BVH  
    computeBounds();
}

GX_CSG_Difference::~GX_CSG_Difference(void) {
    ;;;      
}
    
bool GX_CSG_Difference::intersect(const GRay3& pRay, 
				  GInterSectionList& rInter) 
{
    static GAllInterSectionList sIHeap;
    static GAllInterSectionList sLocalIHeap;
    bool tHIT = false;
    
    GFloat tMin, tMax;
    if(!getWorldBounds().findMinMax(pRay, tMin, tMax))
      return (false);

    
    GRay3 tRay(pRay.getPoint(), pRay.getVector());
    tRay.setTmin(-MAXFLT);
    tRay.setTmax(MAXFLT);
    
    
    mLastChild = 0;
    int tNum = mChildrenList.count();
    bool tInside;
    register int tIntervalCount = 0;
    
    // head
    GXObjectItr _Itr(mChildrenList);
    if(!_Itr.cur()->intersect(tRay, sIHeap))
      goto L_EXIT;
    
    for(_Itr.next(); _Itr.more(); _Itr.next()) {
	if(_Itr.cur()->intersect(tRay, sLocalIHeap)) {
	    
	    // haha, a little tricky here..We generate
	    // the inverse intervalls of object line-segments and
	    // apply the same methode as w/ GX_CSG_Intersection !!
	    sIHeap.add(GInterSection(*(_Itr.cur()), -MAXFLT, true));
	    sIHeap.add(GInterSection(*(_Itr.cur()), MAXFLT, false));
	    
	    while(sLocalIHeap.count()) {
		GInterSection tI = sLocalIHeap.remove();
		sIHeap.add(GInterSection(tI.getPrimitive(), 
					 tI.getT(), 
					 !tI.enter()));

	    }
	}
	else {
	    tNum--;
	}
	
    }
    
    while(sIHeap.count()) {
	GInterSection tI = sIHeap.remove();
	
	tInside = (tI.enter()) ? 
	  (++tIntervalCount == tNum) : (--tIntervalCount == tNum -1);
	
	// a very bad example for aggreate objects! 
	if( tInside && pRay.isValid(tI.getT())) {
	    rInter.add(GInterSection(self, tI.getT(), tI.enter()));
	    mLastChild = &(tI.getPrimitive());
	    tHIT = true;
	    break;
	}
	
    }
    
    
    L_EXIT:
    
    sIHeap.reset();
    return (tHIT);
}


GVector3 GX_CSG_Difference::getNormal(const GPoint3& pPoint) {
    gcheck(!mLastChild, "CSG_Intersection::getNormal() w/o mLastChild called!");
    return (mLastChild->getNormal(pPoint));
}

const GSurface& GX_CSG_Difference::getSurface(const GPoint3& pPoint) {
    gcheck(!mLastChild, "CSG_Intersection::getSurface() w/o mLastChild called!");
    return (mLastChild->getSurface(pPoint));
}
    
const GBounds3& GX_CSG_Difference::computeBounds(void) {
    setWorldBounds() = mChildrenList[0]->getWorldBounds();
    return (getWorldBounds());
}



