//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "gx/gx_defs.h"
#include "gx/gx_shade.h"
#include "gx/gx_cam.h"

#include "gx/gx_light.h"
#include "gx/gx_obj.h"
#include "gx/gx_hyper.h"
#include "gx/gx_bsp.h"



GDLList<GXLight>* gGXLights = 0;


//////////////////////////////////////////////////////////////////////////////
// @GXAmbientLight@
//////////////////////////////////////////////////////////////////////////////

GColorRGB GXAmbientLight::shade(GXShadingParams &pSP) { 
    if (!isOn()) 
	return (GColorRGB( 0, 0, 0 ));
    else
	return (getColor() * pSP.mSurface->getAmbient());
}




//////////////////////////////////////////////////////////////////////////////
// @GXPointLight@
//////////////////////////////////////////////////////////////////////////////

GXPointLight::GXPointLight(void)
: GXLight(), GPointLight() 
{
    mShadowCache = new  GXCache();
}

GXPointLight::~GXPointLight(void) {
    gkillPointer(mShadowCache);
}

void GXPointLight::preprocess(GXObjectList *pObjLists[6]) {
    GPoint3 tMin = getPosition() - GPoint3(gEpsilon, gEpsilon, gEpsilon);
    GPoint3 tMax = getPosition() + GPoint3(gEpsilon, gEpsilon, gEpsilon);
    
    mLightBuffer = new GXViewBuffer(pObjLists, GBounds3(tMin, tMax));
    
}


/*
 * @FUNCT@: @GXPointLight::shade@
 * 
 * @DESCR@: shade a point by "first order"-illumination
 * 
 * @NOTES@: Shadow-cache (Haines & Co.:-)
 * 
 */

GColorRGB GXPointLight::shade(GXShadingParams &rSP)
{
    if (!isOn()) 
	return(GColorRGB( 0, 0, 0 ));

    return (getLight(getPosition(), rSP));
}

GColorRGB GXPointLight::getLight(const GPoint3& pTarget, GXShadingParams &rSP)
{
   
    GColorRGB tColor(0, 0, 0);
	
    // calculate the shadow rays direction 
    GVector3 tDir(pTarget - rSP.mP);   

    // calculate the distance from rays origin to the given point 
    GFloat   tDist = tDir.abs(); 
   
    // normalize the rays direction
    tDir /= tDist;

    // the point faces away from light! There\'re no lights comming
    // from the other side, if the object is fully opaque.
    GFloat tDot = rSP.mN.dot(tDir);
    if(tDot < 0.0 && !rSP.mSurface->getTransparency())
	return (tColor);
      
    

    // update the light vector 
    rSP.mL = tDir;


    // We are going to fire a shadow ray! 
    GRay3    tShadowRay(rSP.mP, tDir, GRay3::G_SHADOW);
    gstat(GXStatistics::ShadowRays++);
    
    // update tmin and tmax. Significant interval is [0, _dis] 
    tShadowRay.minmax(gEpsilon, tDist - gEpsilon);

    // if lightbuffer is ON, revers the ray
    if(GXOptions::ShadowAcceleration == GX_BUF) {
	tShadowRay.setPoint(pTarget);
	tShadowRay.setVector(tDir.negate());
    }

    
    // the shadow-test w/ the primitive containing the intersection point
    // is unneccessary if this primitive is planar, or if it is convex and
    // the last intersection was going into it.
    if(rSP.mPrimitive->isPlanar() ||
       (tDot > 0.0 && rSP.mPrimitive->isConvex()))
	rSP.mPrimitive->updateID(tShadowRay);
    
    


    //// finding the last-object and the last-voxel hited by this ray\'s level
    GXCache *tCache = mShadowCache->getNode(tShadowRay); 
    
 
    // First, check the last opaque object for intersection 
    if (tCache->mLastObject)  {
	if (tCache->mLastObject->hitRay(tShadowRay))  {
	    gstat(GXStatistics::ShadowLastObjectHits++);
	    gstat(GXStatistics::ShadowHits++);

	    // Full shadowing!
	    return tColor;
	}
	// last opaque object missed!
	tCache->mLastObject->updateID(tShadowRay);
	tCache->mLastObject = 0;
    }
    
    

    if (tCache->mLastVoxel) { 
	    
	// if BSP is used, intersect the shadowray w/ the last recently 
	// voxel. If HIT, return full shadowing, else update the mailbox
	// of this voxel. 
	if(GXOptions::ShadowAcceleration == GX_BSP) {
	    GOneInterSectionList tInterList;
	    if(tCache->mLastVoxel->intersectShadow(tShadowRay, tInterList)) {
		tCache->mLastObject = &(tInterList.getFirst().getPrimitive());
		gstat(GXStatistics::ShadowLastVoxelHits++);
		gstat(GXStatistics::ShadowHits++);
    
		// Full shadowing!
		return tColor;
	    }
	    //@WARNING@
	    ((GXBBox *)(tCache->mLastVoxel))->updateID(tShadowRay);
	    tCache->mLastVoxel = 0;
		
	} 
	// If ray-classification is used, check the 5D-tuple of the ray
	// w/ the last hyper-cube for intersecting. 
	else if(GXOptions::ShadowAcceleration == GX_HOT ||
		GXOptions::ShadowAcceleration == GX_BUF) 
	{
	    GXBeam *tBeam = (GXBeam *)tCache->mLastVoxel;
		
	    GFloat t5D[5];
	    int tAxis = tShadowRay.compute5D(t5D);
		
	    if(tAxis == tBeam->getDominantAxis() && tBeam->isInside(t5D)) {
		
		GOneInterSectionList tInterList;
		if (tBeam->intersectShadow(tShadowRay, tInterList)) {
		    tCache->mLastObject = &(tInterList.getFirst().getPrimitive());
		    
		    // Full occludding
		    gstat(GXStatistics::ShadowHits++);
		    gstat(GXStatistics::ShadowLastVoxelHits++);
		    return tColor;
		}
		
		// There was no opaque-object, which occludes the light.
		tColor += shadeDiffuse(rSP);
		tColor += shadeSpecular(rSP);
		return (tColor 
			* getColor()  
			* tInterList.getTransmission() 
			* findDistAtten(tDist));
		
	    }
	    else
		tCache->mLastVoxel = 0;
	

	}// Acceleration
	
    }

    


    
    GColorRGB tAtten;
    
    // shadow cache missed! Test for shadow w/ all objects in the scene
    if(GXOptions::ShadowAcceleration == GX_BUF) {
	GOneInterSectionList tInterList;
	 
	if (mLightBuffer->intersectShadow(tShadowRay,tInterList, 
					  tCache->mLastObject,
					  tCache->mLastVoxel)) 
	    
	    tAtten = GColorRGB(0, 0, 0);
	else   
	    tAtten = GColorRGB(1.0, 1.0, 1.0) * tInterList.getTransmission();
    }
    else {
	tAtten = rSP.mRayTracer->traceShadowRay(tShadowRay, 
						tCache->mLastObject,
						tCache->mLastVoxel);
    }
   
    // a new opaque object intersects the ray! 
    if (tCache->mLastObject) { 
	gstat(GXStatistics::ShadowHits++);

	return tColor;
    }
    
    // multiple the light-color w/ all attenuation factors. We are done! 
    tColor += shadeDiffuse(rSP);
    tColor += shadeSpecular(rSP);
    return (tColor * getColor() * tAtten * findDistAtten(tDist));
}
 

GColorRGB GXPointLight::shadeDiffuse(const GXShadingParams &pSP)
{
    GFloat tDiffAtten = pSP.mN.dot(pSP.mL);
    if(tDiffAtten > 0.0)
	return (pSP.mSurface->getDiffuse() * 
		tDiffAtten); 
    else
	return (pSP.mSurface->getDiffuse() * 
		-tDiffAtten * 
		pSP.mSurface->getTransparency()); 
}

GColorRGB GXPointLight::shadeSpecular(const GXShadingParams &pSP)
{
//     GFloat tSpecCosine = pSP.mR.dot(pSP.mL);
//     return (tSpecCosine > 0) ? 
//       pSP.mSurface->getSpecular() * 
//       pow(tSpecCosine, pSP.mSurface->getShininess())
//       : GColorRGB(0, 0, 0);
    
    // light comming from the same side with the viewer
    if(pSP.mN.dot(pSP.mL) > 0.0) {
	// blinn\'s halfway vector
	GVector3 tH = (pSP.mV + pSP.mL).unitize();
	
	GFloat tSpecCosine = pSP.mN.dot(tH);
	return (pSP.mSurface->getSpecular() * 
		pow(tSpecCosine, pSP.mSurface->getShininess()));
		
    }
    // light comming from the other side w/ the viewer, so
    // the object must be transparent to let some lights
    // visible.
    else {
	// comming from air or going into air ? 
	GFloat tGamma  = pSP.mSurface->getRefraction();
	
	// blinn\'s halfway vector for transmission
	GVector3 tH = (gabs(tGamma-1)>gEpsilon) 
	    ? ((pSP.mV - tGamma * pSP.mL) / (tGamma -1)).unitize()
		: (pSP.mV - tGamma * pSP.mL).unitize();
	
	GFloat tSpecCosine = pSP.mN.dot(tH);
	return (pSP.mSurface->getSpecular() * 
		pow(tSpecCosine, pSP.mSurface->getShininess()) *
		pSP.mSurface->getTransparency());
	
    }
    
}



//////////////////////////////////////////////////////////////////////////////
// @GXRectangularLight@
//////////////////////////////////////////////////////////////////////////////



void GXRectangularLight::computeModelling(void) {
    GMatrix4 tMat;
    
    //// scalling
    tMat = tMat.scale(GVector3(mWidth, mHeight, 1));
    
    //// Rotating
    GVector3 tNormal = mLookAt - getPosition();
    tNormal = tNormal.unitize();
    
    GVector3 tRotAxis;
    if (1. - fabs(tNormal.getZ()) < gEpsilon) 
	tRotAxis = GVector3(1, 0, 0);
    else 
	tRotAxis = GVector3(tNormal.getY(), -tNormal.getX(), 0);
    
    tRotAxis = tRotAxis.unitize();
    GFloat n1 = tRotAxis.getX(); 
    GFloat n2 = tRotAxis.getY(); 
    GFloat n3 = tRotAxis.getZ();
    
    GFloat costheta = tNormal.getZ();
    GFloat sintheta = -sqrt(1 - costheta*costheta);

    GMatrix4 m1;
    m1.set(0,0, (n1*n1 + (1. - n1*n1)*costheta));
    m1.set(1,0, (n1*n2*(1 - costheta) + n3*sintheta));
    m1.set(2,0, (n1*n3*(1 - costheta) - n2*sintheta));
    m1.set(0,1, (n1*n2*(1 - costheta) - n3*sintheta));
    m1.set(1,1, (n2*n2 + (1 - n2*n2)*costheta));
    m1.set(2,1, (n2*n3*(1 - costheta) + n1*sintheta));
    m1.set(0,2, (n1*n3*(1 - costheta) + n2*sintheta));
    m1.set(1,2, (n2*n3*(1 - costheta) - n1*sintheta));
    m1.set(2,2, (n3*n3 + (1 - n3*n3)*costheta));
    
    //// translating
    m1.set(0,3, getPosition().getX());
    m1.set(1,3, getPosition().getY());
    m1.set(2,3, getPosition().getZ());
    
    
    
    tMat = m1 * tMat;
    mModel.setMatrix(tMat);

}

GXRectangularLight::GXRectangularLight(const GPoint3& pPosition, 
			 const GPoint3& pLookAt, 
			 GFloat pWidth, 
			 GFloat pHeight,
			 int pXRes,
			 int pYRes,
			 int pMinSamples,
			 int pMaxSamples,
			 int pNumCandidates)
: GXPointLight(), mLookAt(pLookAt), 
mWidth(pWidth), mHeight(pHeight),
mXRes(pXRes), mYRes(pYRes),
mSampling(pMinSamples, pMaxSamples, pNumCandidates)
{
    setPosition(pPosition);
    
    //@WARNING@ :fixed sampling pattern!
    //mSampling.generateSamples();
}

GXRectangularLight::~GXRectangularLight(void) {

}


void GXRectangularLight::preprocess(GXObjectList *pObjLists[6]) {
    GBounds3 tB(GVector3(MAXFLT, MAXFLT, MAXFLT),
		GVector3(-MAXFLT, -MAXFLT, -MAXFLT));
    
    tB += mModel.mc2wc(GPoint3(-1, -1, 0));
    tB += mModel.mc2wc(GPoint3(1, -1, 0));
    tB += mModel.mc2wc(GPoint3(1, 1, 0));
    tB += mModel.mc2wc(GPoint3(-1, 1, 0));
    
    
    mLightBuffer = new GXViewBuffer(pObjLists, tB);
    
}


void GXRectangularLight::setPosition(const GPoint3& pPosition) {
    GLight::setPosition(pPosition);
    computeModelling();
}

    
GColorRGB GXRectangularLight::shade(GXShadingParams& rSP) {
    
    // light\'s surface normal
    GVector3 tNormal = (mLookAt - getPosition()).unitize();

    
    //////////////////////////////////////////////////////////////////////
    //// First, fire a shadow ray to the center of the light...
    //////////////////////////////////////////////////////////////////////
	
    // calculate the shadow rays direction 
    GVector3 tDir(rSP.mP - getPosition());   

    // calculate the distance from rays origin to the given point 
    GFloat   tDist = tDir.abs(); 
   
    // normalize the rays direction
    tDir /= tDist;

    // We fire a dummy-shadow ray to get the actual leaf volume. 
    GRay3    tShadowRay(getPosition(), tDir, GRay3::G_SHADOW);
    gstat(GXStatistics::ShadowRays++);
    
    // update tmin and tmax. Significant interval is [0, _dis] 
    tShadowRay.minmax(gEpsilon, tDist - gEpsilon);

    
    //// finding the last-object and the last-voxel hited by this ray\'s level
    GXCache *tCache = mShadowCache->getNode(tShadowRay); 
    GXBeam *tBeam = (GXBeam *)tCache->mLastVoxel;
		
    //// classify the right volume of the light-buffer
    GFloat t5D[5];
    int tAxis = tShadowRay.compute5D(t5D);
    if(!tBeam || tAxis != tBeam->getDominantAxis() || !tBeam->isInside(t5D)) {
	tBeam = mLightBuffer->classify(tShadowRay);
	tCache->mLastVoxel = tBeam;
    }
    



    //// max. DIFFUSE
    GVector3 tL[4];
    tL[0] = (mModel.mc2wc(GPoint3(-1, -1, 0)) - rSP.mP);
    tL[1] = (mModel.mc2wc(GPoint3( 1, -1, 0)) - rSP.mP);
    tL[2] = (mModel.mc2wc(GPoint3( 1,  1, 0)) - rSP.mP);
    tL[3] = (mModel.mc2wc(GPoint3(-1,  1, 0)) - rSP.mP);
    
    // the shading point lies behind the light-plane
    bool tFlip = (tL[0].dot((tL[1] - tL[0]).cross(tL[2] - tL[0])) > 0);
    
    tL[0] /= tL[0].abs();
    tL[1] /= tL[1].abs();
    tL[2] /= tL[2].abs();
    tL[3] /= tL[3].abs();
    

    // clipping planes for the light-pyramide
    // @@@@@@@@@@@@@@@@@@ WARNING!!! @@@@@@@@@
    GXBeamPlane *tClipPlanes[5];
    tClipPlanes[4] = new GXBeamPlane((tFlip) ? tNormal : tNormal.negate(), 
				     getPosition());

    // compute the diffuse constribution using contour integral
    GFloat tSum = 0;
    
    
    for(register int i=0; i<4; i++) {
	int j = (i+1)%4;
	    
	GFloat tBeta = acos(tL[i].dot(tL[j]));
	GVector3 tN = (tFlip) ? 
	    tL[i].cross(tL[j]).unitize() : tL[j].cross(tL[i]).unitize();
	
	tClipPlanes[i] = new GXBeamPlane(tN.negate(), rSP.mP);
	
	GFloat tCosDelta = tN.dot(rSP.mN);
	
	tSum += (tCosDelta * tBeta);
    }
    
    GFloat tVisibleFraction = 1.0;
    GColorRGB tMaxColor = getColor() * rSP.mSurface->getDiffuse() * tSum * 0.5;
    




    
    //////////////////////////////////////////////////////////////////////
    //// get the list of occluding candidates
    //////////////////////////////////////////////////////////////////////
    static GXObjectList tClipList;
    static GXObjectList tOccludingList;
    tBeam->getClippedCandidatesList(tShadowRay, tClipList);
    


    //// if no shadow can occurs, simple integrate over the light\' area
    // @WARNING@ : more conditions required !
    if(tClipList.count() == 1) 
	goto L_EXIT;
    
    //// Extra work required, if this beam contains more than one candidate!
    else {

	// First, construct a shadow-culler from the shading point to the
	// lightsource...
	tOccludingList.reset();
	for(GXObjectItr _Itr(tClipList); _Itr.more(); _Itr.next()) {
	    GXObject *tPrim = _Itr.cur();
	
	    GBounds3 tB = tPrim->getWorldBounds();
	    GFloat tExts[6];
	    tExts[0] = tB.getMin().getX(); tExts[1] = tB.getMax().getX();
	    tExts[2] = tB.getMin().getY(); tExts[3] = tB.getMax().getY();
	    tExts[4] = tB.getMin().getZ(); tExts[5] = tB.getMax().getZ();

	    if(tClipPlanes[0]->isOverlap(tExts) && 
	       tClipPlanes[1]->isOverlap(tExts) &&
	       tClipPlanes[2]->isOverlap(tExts) && 
	       tClipPlanes[3]->isOverlap(tExts) &&
	       tClipPlanes[4]->isOverlap(tExts))
		tOccludingList.append(_Itr.cur());
	}

	if(tOccludingList.count() == 1) {
	    //cerr << "clipping successed!\n";
	    
	    goto L_EXIT;
	}
	
	GFloat tNumSamples = 0.0;
	GFloat tNumVisible = 0.0;
	
	GFloat wc = 2.0 / mXRes ;
	GFloat hc = 2.0 / mYRes ;

	GFloat _xx, _yy;
	register int i;
	register int j;
	for (j = 0, _yy = -1.0 + 0.5*hc; j < mYRes; j++, _yy += hc) {
	    for (i = 0, _xx = -1.0 + 0.5*wc; i < mXRes; i++, _xx += wc) {
	    
		GVector3 tImin(MAXFLT, MAXFLT, MAXFLT);
		GVector3 tImax(-MAXFLT, -MAXFLT, -MAXFLT);
		mSampling.generateSamples();
	    
		for(register int k = 0; k < mSampling.getNumber(); k++) {
		    GPoint2 tSample = mSampling.getSample(k);
		    GPoint3 tOrigin = mModel.mc2wc(GPoint3(_xx + wc*(tSample.getX() - 0.5), 
							   _yy + hc*(tSample.getY() - 0.5), 
							   0));
	
		    GVector3 tDir(rSP.mP - tOrigin);   
		    GFloat   tDist = tDir.abs(); 
		    tDir /= tDist;
		    
		    
		    // update the light vector
		    rSP.mL = tDir.negate();

		    // the point faces away from light! There\'re no lights comming
		    // from the other side, if the object is fully opaque.
		    GFloat tDot = rSP.mN.dot(rSP.mL);
		    if(tDot < 0.0 && !rSP.mSurface->getTransparency())
			continue;
		    
		    
		    GRay3    tShadowRay(tOrigin, tDir, GRay3::G_SHADOW);
		    gstat(GXStatistics::ShadowRays++);
    
		    // update tmin and tmax. Significant interval is [0, _dis] 
		    tShadowRay.minmax(gEpsilon, tDist - gEpsilon);
		    
		    // the shadow-test w/ the primitive containing the intersection point
		    // is unneccessary if this primitive is planar, or if it is convex and
		    // the last intersection was going into it.
		    if(rSP.mPrimitive->isPlanar() ||
		       (tDot > 0.0 && rSP.mPrimitive->isConvex()))
			rSP.mPrimitive->updateID(tShadowRay);
		    
		    // First, check the last opaque object for intersection 
		    if (tCache->mLastObject)  {
			if (tCache->mLastObject->hitRay(tShadowRay))  {
			    gstat(GXStatistics::ShadowLastObjectHits++);
			    gstat(GXStatistics::ShadowHits++);

			    // Full shadowing!
			    continue;
			}
			// last opaque object missed!
			tCache->mLastObject->updateID(tShadowRay);
			tCache->mLastObject = 0;
		    }
    
    
		    static GXObjectList sTransList;
		    
		    GOneInterSectionList tInterList;
		    bool tInShadow = false;
		    GColorRGB tI(0,0,0);
		    for(GXObjectItr _Itr(tOccludingList); _Itr.more(); _Itr.next()) {

			GXObject *tPrim = _Itr.cur();
	
			GBounds3 tB = tPrim->getWorldBounds();
			
			// check the mailbox and the bbox, before I-testing w/ the prim. 
			if(tPrim->checkID(tShadowRay)) {
			    gstat(GXStatistics::MailBoxHits++);
			    continue;
			}
	
			if(tPrim->isBounded()) {
			    gstat(GXStatistics::BBox.mTests++);
			    GFloat tMin, tMax;
			    // miss the bbox, continuing w/ the next prim.
			    if(!tB.findMinMax(tShadowRay, tMin, tMax))
				continue;
	    
			    gstat(GXStatistics::BBox.mHits++);
			}

			// If this prim is transparent, append it to a temporal list for
			// later processing.
			if(tPrim->getTransmission()) {
			    sTransList.append(tPrim);
			    continue;
			}

			if(tPrim->intersect(tShadowRay, tInterList)) {
			    tCache->mLastObject = tPrim;
		    
			    // Full occludding
			    gstat(GXStatistics::ShadowHits++);
			    gstat(GXStatistics::ShadowLastVoxelHits++);
			    tInShadow = true;
			    
			    break;
			}
		    }

		    // no opaque object was hit, so continue w/ transparent-objects
		    if(!tInShadow) {
			if(sTransList.count()) {
			    for(GXObjectItr Itr(sTransList); Itr.more(); Itr.next()) {
				GXObject *tPrim = Itr.cur();
	    
				tPrim->intersect(tShadowRay, tInterList);
			    }
			    
			    sTransList.reset();
			}
			
			tI += shadeDiffuse(rSP);
			//tI += shadeSpecular(rSP);
			tI *= (tInterList.getTransmission() * findDistAtten(tDist));

			tNumVisible += tInterList.getTransmission();
		    }
		    
		    // multiplies w/ cos(phi)
		    tI *= gabs(tNormal.dot(tDir));
		    
		    GVector3 tAct(tI.getR(), tI.getG(), tI.getB());
		    tImin.setMin(tAct);
		    tImax.setMax(tAct);
		    if(k >= mSampling.getMinS()) {
			GVector3 tContrast = (tImax - tImin) / (tImax + tImin);
			if(tContrast.getX() < GXOptions::ContrastThreshold.getX() &&
			   tContrast.getY() < GXOptions::ContrastThreshold.getY() &&
			   tContrast.getZ() < GXOptions::ContrastThreshold.getZ() )
			{
			    k++;
			    break;
			}
		    }
		}
		tNumSamples += k;
	    }
	}
    
	tVisibleFraction = tNumVisible / tNumSamples;
    }
    

  L_EXIT:
    
    for(i=0; i<5; i++)
	gkillPointer(tClipPlanes[i]);
    
    
    return (tMaxColor * tVisibleFraction);
}





//////////////////////////////////////////////////////////////////////////////
// @GXSphericalLight@
//////////////////////////////////////////////////////////////////////////////


GXSphericalLight::GXSphericalLight(const GPoint3& pPosition, 
				   GFloat pRad,
				   int pXRes, int pYRes) : 
GXPointLight()  
{
    mRad = pRad;
    GLight::setPosition(pPosition);

    mXRes = pXRes;
    mYRes = pYRes;
    
}

    
GXSphericalLight::~GXSphericalLight(void) {
///
}


void GXSphericalLight::preprocess(GXObjectList *pObjLists[6]) {
    GPoint3 tMin = getPosition() - GPoint3(mRad, mRad, mRad);
    GPoint3 tMax = getPosition() + GPoint3(mRad, mRad, mRad);
    
    mLightBuffer = new GXViewBuffer(pObjLists, GBounds3(tMin, tMax));
    
}


void GXSphericalLight::setPosition(const GPoint3& pPosition) {
    GLight::setPosition(pPosition);
}

GColorRGB GXSphericalLight::shade(GXShadingParams& rSP) {
    GColorRGB tColor(0, 0, 0);

    // direction from shading point to the center of luminare
    GVector3 tDir(getPosition() - rSP.mP);   
    GFloat   tDist = tDir.abs(); 
    tDir /= tDist;
    
    // compute the max angle between any new distributed ray and the
    // vector from shading point to sphericallight\'s center.
    // assuming the shading point is outside the spherical luminare
    GFloat tMaxAngle = asin(mRad / tDist);
    
    // distribute the shadow ray inside this solid angle
    GRay3 tMainRay(rSP.mP, tDir, GRay3::G_SHADOW);

    // We got a N-rooks here...
    GX_NRooksSampling tSampling(mXRes, mYRes);
    tSampling.generateSamples();
    
    int tNumS = mXRes * mYRes;
    
    for (register int i=0; i < tNumS; i++) {
	
	GVector2 tSample = tSampling.getSample(i);
	GRay3 tDistRay = tMainRay.distribute(tMaxAngle, 
					     tSample.getX()*2. - 1., 
					     tSample.getY()*2. - 1.);

	
	// find the intersection of this ray w/ our luminare
	GFloat tPrjDist = tDir.dot(tDistRay.getVector()) * tDist;
	GFloat tDisc = tPrjDist*tPrjDist - tDist * tDist + mRad * mRad;
	if (tDisc < gEpsilon)
	    continue;

	GFloat tTmin = tPrjDist - sqrt(tDisc);
	GPoint3 tTarget = tDistRay.extrap(tTmin);
    
	// fire the distributed shadow ray to the luminare...
	GColorRGB tRadiance = getLight(tTarget, rSP);

	// Normal at point on the light
	GVector3 tNormal = tTarget - getPosition();
	tNormal = tNormal.unitize();
	
	GFloat tCosTheta =  gabs(tNormal.dot(tDistRay.getVector()));
	tRadiance *= tCosTheta;
	
// 	tRadiance *= ( M_PITIMES2 * (1 - cos(tMaxAngle)) * tTmin*tTmin / tCosTheta); 
	
	tColor += tRadiance;
    }
    
    //// Monte Carlo w/ N = 9, distr. density func. = 1 / solid
    
    return (tColor /= tNumS);
}

 








