//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [shading] for black and white shading
// OpenGL 1.0

// BUG 02/04/96: Solid shading does not work properly for quadmesh.

#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gevent.h"

#define GPFILL_COLOR 0.9, 0.9, 0.9
#define GPLINE_COLOR 0.0, 0.0, 0.0

// class GShadable:

GShadable::~GShadable(void) {
    if (getID()) {
	if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
	gFreeID(getID());
    }
}

void GShadable::setMatrix(const GMatrix4& pMat) {
    GLfloat m[16];
    int ind = 0;
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    m[ind++] = pMat.get(j, i);
    glPushMatrix();
    glMultMatrixf(m);
}

void GShadable::unsetMatrix(void) { 
    glPopMatrix(); 
}

void GShadable::setSurface(const GSurface& pSurf) {
    glColorMaterial(GL_FRONT, GL_AMBIENT);
    if (&pSurf == &gRedSurface || &pSurf == &gBlueSurface || &pSurf == &gGreenSurface) 
	glColor3f(pSurf.getAmbient().getR(), pSurf.getAmbient().getG(), pSurf.getAmbient().getB());
    else glColor3f(GPLINE_COLOR);
}

void GShadable::openDisplayList(void) {
    if (!mID) mID = gGetID();
    if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
    glNewList((GLuint)getID(), GL_COMPILE);
}

void GShadable::closeDisplayList(void) { 
    glEndList(); 
}

void GShadable::callDisplayList(void) const {
    if (!getID()) gOutput->fatal("Attempt to call non-existing display list.");
    glCallList((GLuint)getID());
} 

// class GShadableQuadmesh:

void GShadableQuadmesh::shade(void) {
    if (fillstyle) {
	glClear(GL_STENCIL_BUFFER_BIT);

	int rows = getX();
	int cols = getY();
	float f[4];
	int i, j;
	
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(GPLINE_COLOR);
	
	for (i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for (int j = 0; j < cols; j++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (i = 0; i < rows; i++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	
	glColor3f(GPFILL_COLOR);
	glStencilFunc(GL_EQUAL, 0, 1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	
	for (j = 0; j < cols - 1; j++) {
	    glBegin(GL_QUAD_STRIP);
	    for (i = 0; i < rows - 1 ; i++) {
		for(int p = 0; p < 2; p++)
		    for(int r = 0; r < 2; r++) {
			const GVertex& vt = getVertex(i + p, j + r);
			vt.getPoint().get(f);
			glVertex3fv(f);
		    }
	    }
	    glEnd();
	}
	
	glColor3f(GPLINE_COLOR);
	glStencilFunc(GL_ALWAYS, 0, 1);
	glStencilOp(GL_INVERT, GL_INVERT, GL_INVERT);
	
	for (i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for (int j = 0; j < cols; j++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (i = 0; i < rows; i++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
    } 
    else {
	int rows = getX();
	int cols = getY();
	float f[4];
	int i, j;
	
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(GPLINE_COLOR);
	
	for (i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for (int j = 0; j < cols; j++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (i = 0; i < rows; i++) {
		const GPoint3& vt = getVertex(i, j).getPoint();  
		vt.get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
    }
}

template<class V>
class GShadeTopology1DIterator: public GIterator<V> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Shading iterator for subclasses
    // of 1D topology.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(V& v) { 
	GVector3 po = v.getPoint();
	GLfloat p[4];
	p[0] = po.getX();
	p[1] = po.getY();
	p[2] = po.getZ();
	glVertex3fv(p);
	return(false);
    }
  private:
    const GVector3* mNormal;
};

void GShadablePolyline::shade(void) {
    glBegin(GL_LINE_STRIP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

void GShadablePolygon::shade(void) {
    GShadeTopology1DIterator<GVertex> tIter;

    if (fillstyle) {
	glClear(GL_STENCIL_BUFFER_BIT);
	
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(GPLINE_COLOR);
	
	glBegin(GL_LINE_LOOP);
	getVertexList().doWithElements(tIter);
	glEnd();
	
	glColor3f(GPFILL_COLOR);
	glStencilFunc(GL_EQUAL, 0, 1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	
	glBegin(GL_POLYGON); 
	getVertexList().doWithElements(tIter);
	glEnd();
	
	glColor3f(GPLINE_COLOR);
	glStencilFunc(GL_ALWAYS, 0, 1);
	glStencilOp(GL_INVERT, GL_INVERT, GL_INVERT);
	
	glBegin(GL_LINE_LOOP);
	getVertexList().doWithElements(tIter);
	glEnd();
    } 
    else {
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(GPLINE_COLOR);

	glBegin(GL_LINE_LOOP);
	getVertexList().doWithElements(tIter);
	glEnd();
    }
}

template<class F>
class GShadeTopologyPhIterator: public GIterator<F> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator for shading polyhedron-compatible
    // shadable primitives.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The iterator is only useable to render
    // exactly one polyhedron.
    // CONSTANTS: 
  public:
    GShadeTopologyPhIterator(const GTopologyPh& pPolyhedron): mPolyhedron(pPolyhedron), mNumber(0) {}
    virtual bool exec(F& f) { 
	GLfloat sFloat[4]; 
   	glBegin(GL_LINE_LOOP);	
	for (int i = 0; i < f.getNumber(); i++) {
	    const GVertex& v = mPolyhedron.getVertex(mNumber, i);
	    v.getPoint().get(sFloat);
	    glVertex3fv(sFloat);
	}
	glEnd();
	mNumber++;
	return(false);
    }
  private:
    const GTopologyPh& mPolyhedron;
    int mNumber;
};

void GShadablePolyhedron::shade(void) {
    GShadeTopologyPhIterator<GFacet> tIter(*this);
    getFacetList().doWithElements(tIter);
}

void gShaderSetLookatMatrix(const GLookatCamera& r, const GPixmapDisplay& p) {
    glDisable(GL_LIGHTING);
    glShadeModel(GL_FLAT);
    glDrawBuffer(GL_BACK);
    glFrontFace(GL_CW);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    if (GShadable::fillstyle) {
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_STENCIL_TEST);
	glClearStencil(0x0);
	glStencilFunc(GL_ALWAYS, 0, 1);
	glStencilOp(GL_INVERT, GL_INVERT, GL_INVERT);
	glClear(GL_STENCIL_BUFFER_BIT);
    }
    else glDisable(GL_DEPTH_TEST);

    glPushMatrix();
    // setup projection:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (r.isPerspectivic()) {
	gluPerspective((GLdouble)r.getZoom(), (GLdouble)p.getWidth()/(GFloat)p.getHeight(), (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }
    else {
	GFloat tWidth = tan((GLdouble)r.getZoom() * M_PI_2 / 90.0);
	GFloat tHeight = tWidth;
	if (p.getWidth() > p.getHeight()) {
	    tWidth *= (GLdouble)p.getWidth()/(GFloat)p.getHeight();
	}
	if (p.getHeight() > p.getWidth()) {
	    tHeight *= (GLdouble)p.getHeight()/(GFloat)p.getWidth();
	}

	glOrtho(-tWidth, tWidth, -tHeight, tHeight, (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }

    GVector3 vup = gCalcVUP(r.getViewpoint(), r.getRefpoint(), r.getTwist() * M_PI / 180.0);
    gluLookAt(r.getViewpoint().getX(), r.getViewpoint().getY(), r.getViewpoint().getZ(),
	      r.getRefpoint().getX(), r.getRefpoint().getY(), r.getRefpoint().getZ(), 
	      -vup.getX(), -vup.getY(), -vup.getZ()
	      );    
    
    // switch in model mode:
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
} 

int gGetID(void) {
    static sNumber = 1;
    return(sNumber++);
}

void gFreeID(int) {}

void gShaderUnsetMatrix(const GLookatCamera&, const GPixmapDisplay&) {
    glPopMatrix();
}

void GShadePointLight::shade(void) {}

void GShadeAmbientLight::shade(void) {}

GColorRGB GPixmapDisplay::getPixel(int pX, int pY) {
    GLfloat c[3];
    if (pX < 0 || pX >= getWidth() || pY < 0 || pY >= getHeight() )
	gOutput->error("PixmapDisplay::getPixel(): Coordinates out of range.");
    else glReadPixels(pX, pY, 1, 1, GL_RGB, GL_FLOAT, c);
    return(GColorRGB(c[0], c[1], c[2]));
}

void GShadeLight::disable(void) {}

