//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [window handling and interaction]
// X11 + OpenGL 1.0

#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gevent.h"

#ifdef GWITH_TCL
#include "util/gtcl.h"
#endif

static int waitForNotify(Display*, XEvent *e, char *arg) { 
    return (e->type == MapNotify ) && (e->xmap.window == (Window)arg );
}

// the first pixmap will create a new graphics context
// all other link this
static GLXContext sGlobalContext = 0;

// class GPixmapDisplay:

class GPixmapDisplayData {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: OpenGL specific data
    // for pixmap displays.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GPixmapDisplayData(void): first(true) {}
  public:
    Display *display;
    Window window;
    int screen;
    XVisualInfo *visualInfo;
    Colormap colorMap;
    GLXContext context;
    int left, middle, right, width, height;
    bool first;
};

GPixmapDisplay::GPixmapDisplay(const GString& pName, int pWidth, int pHeight): mWidth(pWidth), mHeight(pHeight) {
    // config for doublebuffer mode
    static int sDbConfiguration[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_DOUBLEBUFFER,
        GLX_DEPTH_SIZE, 1,
        (int)None,
    };
    // config for singlebuffer mode
    static int sSbConfiguration[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_DEPTH_SIZE, 1,
        (int)None,
    };
    XSetWindowAttributes wa;
    XEvent e;
    int erb, evb; // error base and event base
    unsigned long mask;

    mData = new GPixmapDisplayData;
    mData->width = mWidth;
    mData->height = mHeight;
    mData->left = mData->middle = mData->right = 0; 
    
    // open a connection to the X server 
    mData->display = XOpenDisplay(0);
    if (!mData->display) gOutput->fatal("Can't connect to display.");

    // is the GLX extension present ?
    if (!glXQueryExtension(mData->display, &erb, &evb)) 
	gOutput->fatal ("X server hasn't an OpenGL GLX extension.");
    mData->screen = DefaultScreen(mData->display);
    mSglBuf = false;
    mData->visualInfo = glXChooseVisual( mData->display, mData->screen, sDbConfiguration);
    if (!mData->visualInfo ) {
        mSglBuf = true;
        mData->visualInfo = glXChooseVisual(mData->display, mData->screen, sSbConfiguration);
        if (!mData->visualInfo) 
	    gOutput->fatal("No appropriate RGB visual with depth buffer.");
    }

    // create OpenGL context:
    mData->context = glXCreateContext( mData->display, mData->visualInfo, sGlobalContext, GL_TRUE );
    if(!mData->context) gOutput->fatal("Can't create graphics context."); 
    if (!sGlobalContext) sGlobalContext = mData->context;

    // set standard properties:
    mData->colorMap = XCreateColormap(mData->display, RootWindow(mData->display, mData->screen), mData->visualInfo->visual, AllocNone);
    wa.colormap = mData->colorMap;
    wa.border_pixel = 0;
    wa.background_pixmap = None;
    wa.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
    mask = CWColormap | CWBorderPixel | CWEventMask | CWBackPixmap;
    mData->window = XCreateWindow(mData->display, RootWindow(mData->display, mData->visualInfo->screen), 0, 0, mWidth, mHeight, 0, mData->visualInfo->depth, InputOutput, mData->visualInfo->visual, mask, &wa);
    XSetStandardProperties(mData->display, mData->window, (char*)pName, (char*)pName, None, 0, 0, 0);

    // request the X window to be displayed on the screen:
    XMapWindow(mData->display, mData->window);
    XIfEvent(mData->display, &e, waitForNotify, (char*)mData->window);

    // connect the context to the window:
    activate();
}	
	  
GPixmapDisplay::~GPixmapDisplay(void) {
    if (mData->display) {
	glFlush();
	glFinish();
	XDestroyWindow(mData->display, mData->window);
	glXDestroyContext(mData->display, mData->context);
	XFreeColormap(mData->display, mData->colorMap);
	XFree((char*)mData->visualInfo);
    }
    delete(mData);
}  

void GPixmapDisplay::activate() {
    if (!glXMakeCurrent(mData->display, mData->window, mData->context)) 
	gOutput->fatal("Can't connect context and window.");
}

void GPixmapDisplay::putPixel(int x, int y, const GColorRGB& colr) {
    GLfloat f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB();
    glColor3fv(f);
    glRecti(x, y, x + 1, y + 1);
    glFlush();
}

void GPixmapDisplay::clear(const GColorRGB& colr) {
    glDrawBuffer(GL_BACK);
    float f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB(); 
    glClearColor(f[0], f[1], f[2], 0.0);
    glClear((GLbitfield)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));

    // setup OpenGL for a pixel-oriented output:

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    if (mSglBuf == false) glDrawBuffer(GL_FRONT_AND_BACK);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, (GLdouble)mData->width , 0.0, (GLdouble)mData->height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

GEvent* GPixmapDisplay::getEvent(void) { 
    if (mData->first) {
	mData->first = false; 
	return(new GUpdatePixmapEvent(this));
    }

    XEvent ev;
    if (XPending(mData->display)) {
	XNextEvent(mData->display, &ev);
	if(ev.xany.window != mData->window) {
	    XPutBackEvent(mData->display, &ev);
	    return(0);
	} 
	int tPress = false;
	switch(ev.type) {
	  case ButtonPress:
	    tPress = true;
	  case ButtonRelease: { 
	      XButtonEvent* evb = (XButtonEvent*)&ev;
	      GMouseButtonEvent* pev = new GMouseButtonEvent;
	      if ((evb->button == 1) && Button1Mask && tPress) pev->left = true; 
	      if ((evb->button == 2) && Button2Mask && tPress) pev->middle = true; 
	      if ((evb->button == 3) && Button3Mask && tPress) pev->right = true; 
	      if (evb->state & ShiftMask) pev->shift = true;
	      if (evb->state & ControlMask) pev->ctrl = true;
	      if (evb->state & Mod1Mask) pev->alt = true;
	      pev->w = mWidth;
	      pev->h = mHeight;
	      pev->x = evb->x;  
	      pev->y = mHeight-evb->y;
	      return(pev); 
	  }
	  case MotionNotify: {
	      XPointerMovedEvent* evm = (XPointerMovedEvent*)&ev;
	      GMouseMotionEvent* pev = new GMouseMotionEvent;
	      if (evm->state & Button1Mask) pev->left = true; 
	      if (evm->state & Button2Mask) pev->middle = true; 
	      if (evm->state & Button3Mask) pev->right = true; 
	      if (evm->state & ShiftMask) pev->shift = 1;
	      if (evm->state & ControlMask) pev->ctrl = 1;
	      if (evm->state & Mod1Mask) pev->alt = 1;
	      pev->w = mWidth;
	      pev->h = mHeight;
	      pev->x = evm->x;  
	      pev->y = mHeight-evm->y;
	      return(pev);
	  }
	  case Expose: 
	    return(new GUpdatePixmapEvent(this));    
	  case ConfigureNotify: {
	    mData->width = ev.xconfigure.width;
	    mData->height = ev.xconfigure.height;
	    setWidth(mData->width);
	    setHeight(mData->height);
	    return(new GResizePixmapEvent(this));
	}
       }
    } 
    return(0);
}

void GPixmapDisplay::swapBuffers(void) {
    if (inSingleBufferMode()) glFlush();
    else glXSwapBuffers(getData().display, getData().window);
}

bool gInitialize(void) { 
    signal(SIGFPE, SIG_IGN);

#ifdef GWITH_TCL
    if (!gGlobalInterp) gGlobalInterp = Tcl_CreateInterp();

    Tcl_SetVar(gGlobalInterp, "tcl_interactive", isatty(0) ? "1" : "0", TCL_GLOBAL_ONLY);
#ifdef GSYSTEM_V
    Tcl_SetVar(gGlobalInterp, "gSystem", "System V", TCL_GLOBAL_ONLY);
#endif
    
    if (Tcl_Init(gGlobalInterp) == TCL_ERROR) gOutput->fatal ("Can't initialize Tcl.");
#endif
    return(true);
}
