//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// FILE: gPCXImg.C
//
// CONTENT: Implementation of PCX image class.
//

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include "render/gpcximg.h"

GPCXImage::GPCXImage(const char* pFname): GImage(pFname) {
    mBase = 0; 
    
    // in following, a virtual method will be called
    // thus, a subclass of PCXImage shouldn't overload the read() method
    gOutput->messageVar("Trying to autoload image ", getFilename(), ".", 0);

    valid = read();
}

GPCXImage::GPCXImage(GPCXFileStruct fileData): GImage(fileData.name) {
    mXSize = fileData.width;
    mYSize = fileData.height;
    newGeometry();

    // fill header information
    memset(&image, 0, sizeof(GPCXImageStruct));

    image.id          = 10;
    image.version     = 5;
    image.compressed  = 1;
    image.bitppix     = 8;

    image.xmax        = fileData.width-1;
    image.ymax        = fileData.height-1;

    image.colorplanes = fileData.colorplanes;
    image.byteperline = ((image.xmax & 1) == 0) ? image.xmax + 2 : image.xmax + 1; 
   
    image.paltype     = 1;

    if ((image.xmax > 0) && (image.ymax > 0) && 
        ((image.colorplanes == 1) || (image.colorplanes == 3))) valid = true;
    else valid = false; 

    mapcnt = 0;  // reset counter for colormap		
}

bool GPCXImage::getValid(void) {
    return valid;
}

void GPCXImage::putShort(FILE* outf, unsigned short val) {
    byte buf[2];

    buf[0] = (byte)(val >> 0);
    buf[1] = (byte)(val >> 8);
    
    fwrite(buf, 2, 1, outf); 
}

unsigned short GPCXImage::getShort(FILE* inf) {
    byte buf[2];

    fread(buf, 2, 1, inf);
    return ((buf[1]<<8) + (buf[0]<<0));
}

//
// PCX Header read and write
//

void GPCXImage::readHeader(FILE* inf) {
    unsigned char buf[1];

    memset(&image, 0, sizeof(GPCXImageStruct));
    memset(&colormap, 0, sizeof(GPCXVGAColorMap));

    fread(buf, 1, 1, inf);
    image.id = buf[0];
    
    fread(buf, 1, 1, inf);
    image.version = buf[0];

    fread(buf, 1, 1, inf);
    image.compressed = buf[0];

    fread(buf, 1, 1, inf);
    image.bitppix = buf[0];

    image.xmin = getShort(inf);
    image.ymin = getShort(inf);    
    image.xmax = getShort(inf);
    image.ymax = getShort(inf);

    image.hdpi = getShort(inf);
    image.vdpi = getShort(inf);

    fread(image.colormap, 48, 1, inf);
    
    fread(buf, 1, 1, inf);
    image.dummy = buf[0];
    
    fread(buf, 1, 1, inf);
    image.colorplanes = buf[0];

    image.byteperline = getShort(inf);
    image.paltype     = getShort(inf);
    
    fread(image.dummys, 58, 1, inf); 
}    	

int GPCXImage::writeHeader(FILE* outf) {
    fwrite(&image.id, 1, 1, outf);
    fwrite(&image.version, 1, 1, outf);
    fwrite(&image.compressed, 1, 1, outf);
    fwrite(&image.bitppix, 1, 1, outf);

    putShort(outf,image.xmin);
    putShort(outf,image.ymin);
    putShort(outf,image.xmax);
    putShort(outf,image.ymax);

    putShort(outf,image.hdpi);
    putShort(outf,image.vdpi);

    fwrite(image.colormap, 48, 1, outf);
    fwrite(&image.dummy, 1, 1, outf);
    fwrite(&image.colorplanes, 1, 1, outf);

    putShort(outf,image.byteperline);
    putShort(outf,image.paltype);
    
    fwrite(image.dummys, 58, 1, outf);

    return(0);
}

// inf specifies a for reading open file
// buf is an buffer of byte, its size : mXSize * image.colorplanes
void GPCXImage::uncompressLine(FILE* inf, byte* buf) {
  int pos = 0,j;
  byte b1[2],b2[2];

  while (pos < mXSize * image.colorplanes) {
      fread(b1,1,1,inf);	

      if (b1[0] < 0xC0)       // controlbyte ?
      {                       // < C0 --> no controlbyte
	  buf[pos] = b1[0];
	  pos++;
      } else                  
      {
          if (image.compressed != 1)  // since uncompressed -> no controlbyte
	  {
	      buf[pos] = b1[0];
	      pos++;
	  } else                      // it is a controlbyte 
	  {
	      fread(b2,1,1,inf);      // read second byte         

	      for (j=0; j < (b1[0] - 0xC0); j++) // copy second byte
	      {
	          buf[pos] = b2[0];
		  pos++;
              }
	  }
      }
  }		 
}


// outf specifies a for writing open file
// buf is an buffer of byte with values,
//     size : mXSize * image.colorplanes 
void GPCXImage::compressLine(FILE* outf, byte* buf) {
  int  pos = 0;
  byte b1,b2,c,j;
  long size = mXSize * image.colorplanes;

  while (pos  < size) {
      b1 = buf[pos];
      
      if (pos+1 < size)
      {
          b2 = buf[pos+1];

	  if (b1==b2)
	  {
	      for (j=2; (pos+j < size) && (j<63) && (buf[pos+j] == b1); j++) ;

              c = 0xC0 + j;
	      fwrite(&c,1,1,outf);
	      fwrite(&b1,1,1,outf); 
	      pos += j;
	  } else
	  {
	      if (b1<0xC0) fwrite(&b1,1,1,outf);  else
	      {
                  c = 0xC1;
		  fwrite(&c,1,1,outf);
		  fwrite(&b1,1,1,outf);
	      }
	      pos++;
	  }
      } else
      {
	  if (b1<0xC0) fwrite(&b1,1,1,outf);  else
	  {
	      c = 0xC1;
	      fwrite(&c,1,1,outf);
	      fwrite(&b1,1,1,outf);
	  }
	  pos++;
      }
  }
}

bool GPCXImage::read(void) {   
   
    FILE* inf = fopen((char*)getFilename(),"r");
    int   i,j;
    byte  *buf,id; 

    if (!inf) {
        gOutput->warningVar("Can't open image file ",(char*)getFilename(),".",0);
        return(false);
    }
 
    readHeader(inf);
 
    if (image.id != 10) {
      gOutput->errorVar((char *)getFilename()," is not a PCX File !",0);  
      fclose(inf);
      return(false);
    }

    if ((image.colorplanes != 1) && (image.colorplanes != 3)) {
      gOutput->errorVar("Unsupported old PCX format in ", 
			(char*)getFilename(), ".", 0);
      fclose(inf);
      return(false);
    }    

    mXSize = image.xmax - image.xmin + 1;
    mYSize = image.ymax - image.ymin + 1;

    if (mBase) delete mBase;
    mBase = new unsigned long[mXSize * mYSize];

    if (image.colorplanes == 1) {
        // read colormap
	if ((image.version == 2) || (image.version == 5)) {
	    fseek(inf,-1* sizeof(colormap) -1, SEEK_END);  // SEEK_END is 2
            fread(&id, 1, 1, inf);
            
            if (id != 0x0C) {
      		gOutput->errorVar("Colormap not found !",0);
      		fclose(inf);
      		return(false);
    	    }
	    
	    fread(colormap.map, sizeof(colormap), 1, inf);
	    fseek(inf, 128, SEEK_SET);		     // SEEK_SET is 0
	} 
      
        buf = new byte[image.byteperline];
	for (i=0; i < mYSize; i++) {
	    uncompressLine(inf, buf);
	    for (j = 0; j < image.byteperline; j++)
		mBase[i* mXSize + j] = buf[j];
	}
        delete buf;
                    
    } // end if
     
    if (image.colorplanes == 3) {
        buf = new byte[image.byteperline * 3];
      
        for (i = 0; i < mYSize; i++) {
	    uncompressLine(inf, buf);
            
	    for (j = 0; j < mXSize; j++) 
		mBase[i* mXSize + j] = (buf[j] << 16) + (buf[j+mXSize] << 8) + (buf[j+mXSize+mXSize] << 0);
	}

        delete buf;        
    }

    fclose(inf);
    return (true);
}

bool GPCXImage::write(void) {
    FILE* outf = fopen((char*)getFilename(),"w+"); 
    int  i, j;
    byte* buf, id;

    if (!outf) {
      gOutput->warningVar("Can't open image file", (char*)getFilename(), ".", 0);
      return(false);
    }

    writeHeader(outf);

    if (image.colorplanes == 1) {
	buf = new byte[image.byteperline];

	for (i=0; i < mYSize; i++) {
	    for (j = 0; j < image.byteperline; j++)
		buf[j] = (byte)(mBase[i* mXSize + j] & 255);
	    compressLine(outf, buf);
	}

	delete buf;
	id = 0x0C;   // colormap id
        fwrite(&id, 1, 1, outf);
	fwrite(colormap.map, sizeof(colormap), 1, outf);

    } // end if

    if (image.colorplanes == 3) {
	buf = new byte[image.byteperline*3];

	for (i = 0; i < mYSize; i++) {

	    for (j = 0; j < image.byteperline; j++) {
	      buf[j]          = (byte)((mBase[i* mXSize + j] >> 16) & 255);
	      buf[j + mXSize]   = (byte)((mBase[i* mXSize + j] >> 8)  & 255);
	      buf[j + 2*mXSize] = (byte)((mBase[i* mXSize + j] >> 0)  & 255);
	    }
	    compressLine(outf, buf);
	}

        delete buf;
    }

    fclose(outf);
    return (true);
}

GColorRGB GPCXImage::getColor(int x, int y) const {
     GColorRGB c;
    
     if (image.colorplanes == 1) {
         long      l = mBase[(long)(y * mXSize) + x];
	 c.setR(((GFloat)colormap.map[l][0])/255.0);
         c.setG(((GFloat)colormap.map[l][1])/255.0);
	 c.setB(((GFloat)colormap.map[l][2])/255.0);        
     }

     if (image.colorplanes == 3) { 
         long  l = mBase[(long)(y * mXSize) + x];
	 byte  r = (l >> 16) & 255;
         byte  g = (l >> 8 ) & 255;
         byte  b = (l >> 0 ) & 255;  

	 c.setR(((GFloat)r)/255.0);
	 c.setG(((GFloat)g)/255.0);
	 c.setB(((GFloat)b)/255.0);	 
     }

     return (c);
}

void GPCXImage::setColor(int x, int y, const GColorRGB& c) {
     if (image.colorplanes == 1) {
	 byte  r = byte(c.getR() * 255);
         byte  g = byte(c.getG() * 255);
	 byte  b = byte(c.getB() * 255);
	 
	 int   i;

	 for (i = 0;
	      (i < mapcnt) && ((r != colormap.map[i][0]) || (g != colormap.map[i][1]) || (b != colormap.map[i][2]));
	      i++) ;

	 if (i == mapcnt) { 
	     // color not in map yet
             if (mapcnt < 16) {
                 image.colormap[mapcnt][0] = r;   // insert color in map
	     	 image.colormap[mapcnt][1] = g;
	         image.colormap[mapcnt][2] = b;
             } 
	     
	     colormap.map[mapcnt][0] = r;   // insert color in map
	     colormap.map[mapcnt][1] = g;
	     colormap.map[mapcnt][2] = b;
             
	
	     i = mapcnt;
             mapcnt++;
         };

	 mBase[y* mXSize + x] = i;	 
     }

     if (image.colorplanes == 3) { 
         byte  r = byte(c.getR()*255);
         byte  g = byte(c.getG()*255);
         byte  b = byte(c.getB()*255);  

	 mBase[y * mXSize + x] = (r << 16) + (g << 8) + (b << 0);
     }

}
