//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "render/gray2D.h"

GTrianglePatch::GTrianglePatch(void): mCutMat(3, 3), mCoorMat(3, 2), isBadTriangle(true) {}

GTrianglePatch::GTrianglePatch(const GSurfaceVertex& pX, const GSurfaceVertex& pY, const GSurfaceVertex& pZ): mX(pX), mY(pY), mZ(pZ), mCutMat(3, 3), mCoorMat(3, 2), mCutM(0), mCutN(0), mCutT(0), mTrianHit(false), isBadTriangle(false) {

    mXZ = pZ.getPoint()-pX.getPoint();
    mXY = pY.getPoint()-pX.getPoint();
    
    mN = mXY.cross(mXZ);
    if(mN.sqdist() < gEpsilon)
	isBadTriangle = true;
    else
	mN = mN.unitize();
    
    // matrix of side vectors to compute the cuts
    mCutMat.setCol(0, mXY);
    mCutMat.setCol(1, mXZ);
    
    // transformation matrix wc <-> tc
    mCoorMat.setCol(0, mXY);
    mCoorMat.setCol(1, mXZ);

    // store the transpose
    mCoorMatTrans = mCoorMat.transpose();
    mCoorMat = mCoorMatTrans*mCoorMat;

    computeBounds();
}

GVector3 GTrianglePatch::getNormal(const GPoint3& P) {
    return(normal2wc(P));
}

GFloat GTrianglePatch::getDistance(const GPoint3& P) {
    return mN.dot(P) - mN.dot(mX.getPoint());
}

bool GTrianglePatch::interpolate(const GPoint3& P, GVector3& k) {
/*
    if(gabs(getDistance(P)) > gEpsilon)
	return false;
*/
    // point not on surface
    
    // compute factors
    GFloat tA = mXY.cross(mXZ).abs()/2.0;
    k[0] = (mY.getPoint() - P).cross(mZ.getPoint() - P).abs()/2.0/tA;
    k[1] = (mZ.getPoint() - P).cross(mX.getPoint() - P).abs()/2.0/tA;
    k[2] = (mX.getPoint() - P).cross(mY.getPoint() - P).abs()/2.0/tA;

    return true;
}

GVector3 GTrianglePatch::normal2wc(GFloat m, GFloat n) {
    if(isBadTriangle)
	return mN;

    if(!validate(m, n))
	return mN;

    if(mX.getNormal() && mY.getNormal() && mZ.getNormal()) {
	GVector3 k;
	if(interpolate(tc2wc(m, n), k))
	    return(*mX.getNormal()*k[0] + *mY.getNormal()*k[1] + *mZ.getNormal()*k[3]);
    }

    return mXY.cross(tc2wc(m, n) - mX.getPoint());
}

GVector3 GTrianglePatch::normal2wc(const GPoint3& P) {
    if(isBadTriangle)
	return mN;

    if(mX.getNormal() && mY.getNormal() && mZ.getNormal()) {
	GVector3 k;
	if(interpolate(P, k))
	    return(*mX.getNormal()*k[0] + *mY.getNormal()*k[1] + *mZ.getNormal()*k[3]);
    }

    return mN;
}

GPoint3 GTrianglePatch::tc2wc(GFloat m, GFloat n) {
    return mX.getPoint() + m*mXY + n*mXZ;
}

void GTrianglePatch::computeBounds() {
    setBounds() = GBounds3
	(GPoint3(+HUGE), GPoint3(-HUGE));

    setBounds() += mX.getPoint();
    setBounds() += mY.getPoint();
    setBounds() += mZ.getPoint();
    
    for(int  i = 0; i < 3; i++)
	while(gabs(getBounds().min()[i] - getBounds().max()[i]) < gEpsilon) {
	    setBounds().min()[i] -= gEpsilon;
	    setBounds().max()[i] += gEpsilon;
	}
}

bool GTrianglePatch::intersect(const GRay3& pRay, GInterSectionList& pList) {	    
    if(isBadTriangle) return false;

    GFloat tRel = mN.dot(pRay.getVector());
    GFloat tDist = getDistance(pRay.getPoint());

    if(gabs(tRel) < gEpsilon)
	return false;
    if(tRel < -gEpsilon && tDist < -gEpsilon)
	return false;
    if(tRel > gEpsilon && tDist > gEpsilon)
	return false;
    if(gabs(tRel) < gEpsilon && gabs(tDist) > gEpsilon)
	return false;

    GFloat m, n, t;
    mTrianHit = false;

    GFloat tMin, tMax;
    if(!getBounds().findMinMax(pRay, tMin, tMax))
	return false;

    // ray is OK
    GVector tCut;
    bool sing;

    mCutMat.setCol(2, -1.0*pRay.getVector());
	
    tCut = mCutMat.solve
	(pRay.getPoint() - mX.getPoint(), sing);
    assert(!sing);
    
    m = tCut[0]; n = tCut[1]; t = tCut[2];

    if(validate(m, n) && validate(t) && pRay.isValid(t)) {
	mCutM = m; mCutN = n; mCutT = t;
	mCutPoint = pRay.compute(t);
	
	mTrianHit = true;
    }
    
    if(mTrianHit)
	pList.add(GInterSection(self, mCutT, tRel < 0));

    // everything failed
    return mTrianHit;
}

bool GTrianglePatch::tc2uw
(GFloat m, GFloat n, GFloat& u, GFloat& w) {
    if(isBadTriangle) return false;

    if(!validate(m, n)) {
	u = w = 0;
	return false;
    }
    
    GVector3 k;
    if(interpolate(tc2wc(m, n), k)) {
	u = k[0]*mX.getU() + k[1]*mY.getU() + k[2]*mZ.getU();
	w = k[0]*mX.getW() + k[1]*mY.getW() + k[2]*mZ.getW();

	return true;
    }

    return false;
}
// triangle object space coordinates to uw object coordinates

bool GTrianglePatch::wc2tc(const GPoint3& P, GFloat& m, GFloat& n) {
    if(isBadTriangle) return false;

    if(gabs(getDistance(P)) > gEpsilon)
	return false;
    // point does not lay in plane
    
    bool sing;
    GVector mn = mCoorMat.solve(mCoorMatTrans*(P - mX.getPoint()), sing);
    assert(!sing);
    // solve by LSD
    
    m = mn[0];
    n = mn[1];
    
    return validate(m, n);
}





