//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <ctype.h>

#include "goutput.h"
#include "util/gconstr.h"
#include "util/gvariabl.h"

GDbConstraint::GDbConstraint(int variableCount, GDbStrength pStrength) {
    if (variableCount > gDbMaxVar) gOutput->error("Too much variables for constraint.");
    inputFlag = false;
    strength = pStrength;
    whichMethod = gDbNoMethod;
    methodCount = 0;
    for (int i = 0; i < gDbMaxVar; i++) methodOuts[i] = 0;
    varCount = variableCount;
}

void GDbConstraint::add(void) {
    for (int i = varCount; i > 0; i--) variables[i-1]->constraints.add(this);
    whichMethod = gDbNoMethod;
    incrementalAdd();
}

int GDbConstraint::chooseMethod(void) {
    int bestOutStrength = strength;
    int best = gDbNoMethod;
    for (int m = methodCount; m > 0; m--) {
	GDbVariable* mOut = variables[methodOuts[m-1]];
	if ((mOut->mark != gDbCurrentMark) &&
	    (gDbWeaker(mOut->walkStrength, bestOutStrength))) {
	    best = m - 1;
	    bestOutStrength = mOut->walkStrength;
	}
    }
    return(best);
}

GDbConstraint* GDbConstraint::satisfy(void) {
    whichMethod = chooseMethod();
    if (satisfied()) {
	// mark inputs to allow cycle detection in addPropagate 
	int outIndex = methodOuts[whichMethod];
	for (int i = varCount; i > 0; i--) 
	    if ((i-1) != outIndex) variables[i-1]->mark = gDbCurrentMark;
	GDbVariable* out = variables[outIndex];
	GDbConstraint* overridden = out->determinedBy;
	if (overridden ) overridden->whichMethod = gDbNoMethod;
	out->determinedBy = this;
	if (!addPropagate()) {
	    gOutput->error("Cycle encountered.");
	    return(0);
	}
	out->mark = gDbCurrentMark;
	return(overridden);
    } else {
	if (strength == G_REQUIRED) gOutput->error("Could not satisfy a required constraint.");
	return(0);
    }
}

bool GDbConstraint::addPropagate(void) {
    delete gDbTodo1;
    gDbTodo1 = new GDbConstraintList;
    GDbConstraint* nextC = this;
    while (nextC) {
	GDbVariable* out = gDbOutVar(nextC);
	if (out->mark == gDbCurrentMark) { 
	    // remove the cycle-causing constraint 
	    incrementalRemove();
	    return(false);
	}
	nextC->recalculate();
	nextC = out->nextDownstreamConstraint(gDbTodo1);
    }
    return(true);
}

template<class P>
class GDbIncAddIterator: public GIterator<P> {
  public:
    virtual bool exec(P& p) { 
	p.incrementalAdd();
	return(false);
    }  
};

void GDbConstraint::incrementalRemove(void) {
    GDbVariable* out = gDbOutVar(this);
    whichMethod = gDbNoMethod;
    for (int i = varCount; i > 0; i--)  
	variables[i-1]->constraints.remove(this);
    GDbConstraintList* unsatisfied = new GDbConstraintList;
    out->removePropagateFrom(unsatisfied);

    i = 0;
    GDbIncAddIterator<GDbConstraint> tIter;
    for (gDbStrength = gDbStrengths[0]; gDbStrength <= G_WEAKEST; gDbStrength = gDbStrengths[++i])
	unsatisfied->doWithElements(tIter);
    delete unsatisfied;
}

GDbConstraint::~GDbConstraint(void) {
    if (satisfied()) incrementalRemove();
    for (int i = varCount; i > 0; i--) 
	variables[i-1]->constraints.remove(this);
}

void GDbConstraint::recalculate(void) {
    GDbVariable* out = gDbOutVar(this);
    out->walkStrength = outputWalkStrength();
    out->stay = constantOutput();
    if (out->stay) this->execute();
}

GDbStrength GDbConstraint::outputWalkStrength(void) {
    int mOutIndex;
    GDbStrength minStrength = strength;
    int outIndex = methodOuts[whichMethod];
    for (int m = methodCount; m > 0; m--) {
    	mOutIndex = methodOuts[m-1];
	if ((mOutIndex != outIndex) &&
	    (gDbWeaker(variables[mOutIndex]->walkStrength, minStrength))) {
	    minStrength = variables[mOutIndex]->walkStrength;
	}
    }
    return(minStrength);
}

bool GDbConstraint::constantOutput(void) {
    if (inputFlag) return(false);
    int outIndex = methodOuts[whichMethod];
    for (int i = varCount; i > 0; i--) {
	if ((i-1) != outIndex) {
	    if (!variables[i-1]->stay) return(false);
	}
    }
    return(true);
}

void GDbConstraint::incrementalAdd(void) {
    gDbCurrentMark++; 
    GDbConstraint* overridden = satisfy();
    while (overridden) overridden = overridden->satisfy();
}

GDbConstraintList* GDbConstraint::extractPlanFromConstraint(void) {
    delete gDbHot;
    gDbHot = new GDbConstraintList;
    if (inputFlag && satisfied()) gDbHot->add(this);
    return(GDbConstraintList::makePlan());
}

bool GDbConstraint::inputsKnown(void) {
    int outIndex = methodOuts[whichMethod];
    for (int i = varCount; i > 0; i--) {
	if ((i-1) != outIndex) {
	    GDbVariable* in = variables[i-1];
	    if ((in->mark != gDbCurrentMark) && (!in->stay) && in->determinedBy )
		return(false);
	}
    }
    return(true);
}

const char* gDbStrength2string(GDbStrength strength) {
    if (strength <= G_WEAKEST) return(gDbStrengthNames[strength]);
    gOutput->error("Bad strength.");
    return(0);
}

GDbStrength gDbString2strength(const char* s) {
    int i = 0;
    GDbStrength st = gDbStrengths[i];
    char* ss = new char[strlen(s) + 1];
    strcpy(ss, s);
    while (*ss) {
	*ss = tolower(*ss);
	ss++;
    }
    const char** tmp = gDbStrengthNames;
    while (*tmp) {
	if (!strcmp(s, *tmp)) {
	    delete ss;
	    return(st); 
	}
	st = gDbStrengths[++i];
	tmp++;
    }
    delete ss;
    return((GDbStrength)-1);
}

GDbStrength gDbStrength;
long gDbCurrentMark;

GDbConstraintList* gDbHot;
GDbConstraintList* gDbTodo1;
GDbConstraintList* gDbTodo2;

const char* gDbStrengthNames[] = {
    "required", 
    "strongpreferred", 
    "preferred",
    "strongdefault",
    "default",
    "weakdefault",
    "weakdefault",
    "weakest",
    0
};

const GDbStrength gDbStrengths[] = {
    G_REQUIRED,
    G_STRONG_PREFERRED,
    G_PREFERRED,
    G_STRONG_DEFAULT,
    G_DEFAULT,
    G_WEAK_DEFAULT,
    G_WEAKEST
};

template<class P>
class GDbExecuteIterator: public GIterator<P> {
  public:
    virtual bool exec(P& p) { 
	p.execute();
	return(false);
    }  
};

void GDbConstraintList::executePlan(void) {
    GDbExecuteIterator<GDbConstraint> tIter;
    doWithElements(tIter);
}

GDbConstraintList* GDbConstraintList::makePlan(void) {
    gDbCurrentMark++; 
    GDbConstraintList* plan = new GDbConstraintList;
    GDbConstraint* nextC = gDbHot->removeFirst();
    while (nextC) {
        GDbVariable* out = gDbOutVar(nextC);
        if ((out->mark != gDbCurrentMark) && nextC->inputsKnown()) {
            plan->add(nextC);
            out->mark = gDbCurrentMark;
            nextC = out->nextDownstreamConstraint(gDbHot);
        }
        else nextC = gDbHot->removeFirst();
    }
    return(plan);
}

template<class P>
class GDbAddIfSatisfiedIterator: public GIterator<P> {
  public:
    virtual bool exec(P& p) { 
	if (p.inputFlag && p.satisfied()) gDbHot->add(&p);
	return(false);
    }  
};

template<class P>
class GDbCollectSatisfiedInputsIterator: public GIterator<P> {
  public:
    virtual bool exec(P&); 
};

template<class P>
bool GDbCollectSatisfiedInputsIterator<P>::exec(P& p) { 
    static GDbAddIfSatisfiedIterator<GDbConstraint> sIter;
    p.constraints.doWithElements(sIter);
    return(false);
}  

GDbConstraintList* GDbConstraintList::extractPlan(void) {
    delete gDbHot;
    gDbHot = new GDbConstraintList;
    GDbCollectSatisfiedInputsIterator<GDbVariable> tIter;
    gDbAllVariables->doWithElements(tIter);
    return(makePlan());
}

GDbConstraintList* GDbConstraintList::extractPlanFromConstraints(void) {
    delete gDbHot;
    gDbHot = new GDbConstraintList;
    GDbAddIfSatisfiedIterator<GDbConstraint> tIter;
    doWithElements(tIter);
    return(makePlan());
}

GDbConstraint* GDbConstraintList::removeFirst(void) {
    if (mList.getNumber()) {
	GDbConstraint* tConstr = getFirst();
	mList.removeFromBegin(tConstr);
	return(tConstr);
    }
    else return(0);
}  

