//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gmatrix.h"
#include "math.h"

const GMatrix4 gUnityMatrix4; 

void gSinCos(GFloat pAngle, GFloat& rSin, GFloat& rCos) {
    rSin = sin(pAngle);
    rCos = cos(pAngle);

    if (rSin > 1 - gEpsilon) {
	rCos = 1; 
	rSin = 0;
    }
    else if (rCos < -1 + gEpsilon) {
	rCos = -1;
	rSin = 0;
    }

    if (rSin > 1 - gEpsilon) {
	rCos = 0;
	rSin = 1;
    }
    else if (rSin < -1 + gEpsilon) {
	rCos = 0;
	rSin = -1;
    }
}


GMatrix4 GMatrix4::scaleS(const GFloat& d) const {
    GMatrix4 m = *this * d; 
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::invScaleS(const GFloat& d) const {
    GMatrix4 m = *this * -d; 
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::scale(const GVector3& v) const {
    GMatrix4 m1, m2;
    m1.set(0, 0, v.mX);
    m1.set(1, 1, v.mY);
    m1.set(2, 2, v.mZ);
    m2 = *this * m1;
    m2.changed();
    return(m2);
}

GMatrix4 GMatrix4::invScale(const GVector3& v) const {
    GMatrix4 m1, m2;
    m1.set(0, 0, 1 / v.mX);
    m1.set(1, 1, 1 / v.mY);
    m1.set(2, 2, 1 / v.mZ);
    m2 = m1 * *this;
    m2.changed();
    return(m2);
}

GMatrix4 GMatrix4::rotate(const GVector3& v) const {
    GMatrix4 m1, m2;
    GFloat sinx = sin(-v.mX);
    GFloat siny = sin(-v.mY);
    GFloat sinz = sin(-v.mZ);

    GFloat cosx = cos(v.mX);
    GFloat cosy = cos(v.mY);
    GFloat cosz = cos(v.mZ);

    m1.val[0][0] = cosy * cosz;
    m1.val[0][1] = cosy * sinz;
    m1.val[0][2] = -siny;

    m1.val[1][0] = cosz * sinx * siny - cosx * sinz;
    m1.val[1][1] = cosx * cosz + sinx * siny * sinz;
    m1.val[1][2] = cosy * sinx;

    m1.val[2][0] = cosx * cosz * siny + sinx * sinz;
    m1.val[2][1] = -(cosz * sinx) + cosx * siny * sinz;
    m1.val[2][2] = cosx * cosy;

    m2 = *this * m1;
    m2.changed();
    return(m2);
}

GMatrix4 GMatrix4::invRotate(const GVector3&v) const {
    GMatrix4 m1, m2;
    GFloat sinx = sin(v.mX);
    GFloat siny = sin(v.mY);
    GFloat sinz = sin(v.mZ);

    GFloat cosx = cos(v.mX);
    GFloat cosy = cos(v.mY);
    GFloat cosz = cos(v.mZ);

    m1.val[0][0] = cosy * cosz;
    m1.val[0][1] = cosz * sinx * siny + cosx * sinz;
    m1.val[0][2] = -(cosx * cosz * siny) + sinx * sinz;

    m1.val[1][0] = -(cosy * sinz);
    m1.val[1][1] = cosx * cosz - sinx * siny * sinz;
    m1.val[1][2] = cosz * sinx + cosx * siny * sinz;

    m1.val[2][0] = siny;
    m1.val[2][1] = -(cosy * sinx);
    m1.val[2][2] = cosx * cosy;

    m2 = m1 * *this;
    m2.changed();
    return(m2);
}

void GMatrix4::unity(void) {
    mChanged = false;
    val[0][1] = val[0][2] = val[0][3] =
	val[1][0] = val[1][2] = val[1][3] = 
	    val[2][0] = val[2][1] = val[2][3] =
		val[3][0] = val[3][1] = val[3][2] = 0;
    val[0][0] = val[1][1] = val[2][2] = val[3][3] = 1;
}

GMatrix4 GMatrix4::inverse(void) const {
    GMatrix4 im;
    GFloat a0 = val[0][0];
    GFloat a1 = val[0][1];
    GFloat a2 = val[0][2];
    GFloat a3 = val[0][3];
    GFloat a4 = val[1][0];
    GFloat a5 = val[1][1];
    GFloat a6 = val[1][2];
    GFloat a7 = val[1][3];
    GFloat a8 = val[2][0];
    GFloat a9 = val[2][1];
    GFloat a10 = val[2][2];
    GFloat a11 = val[2][3];
    GFloat a12 = val[3][0];
    GFloat a13 = val[3][1];
    GFloat a14 = val[3][2];
    GFloat a15 = val[3][3];
    GFloat a1a11a14a4 = a1*a11*a14*a4;
    GFloat a1a10a15a4 = a1*a10*a15*a4;
    GFloat a11a13a2a4 = a11*a13*a2*a4;
    GFloat a10a13a3a4 = a10*a13*a3*a4;
    GFloat a0a11a14a5 = a0*a11*a14*a5;
    GFloat a0a10a15a5 = a0*a10*a15*a5;
    GFloat a11a12a2a5 = a11*a12*a2*a5;
    GFloat a10a12a3a5 = a10*a12*a3*a5;
    GFloat a1a11a12a6 = a1*a11*a12*a6;
    GFloat a0a11a13a6 = a0*a11*a13*a6;
    GFloat a1a10a12a7 = a1*a10*a12*a7;
    GFloat a0a10a13a7 = a0*a10*a13*a7;
    GFloat a15a2a5a8 = a15*a2*a5*a8;
    GFloat a14a3a5a8 = a14*a3*a5*a8;
    GFloat a1a15a6a8 = a1*a15*a6*a8;
    GFloat a13a3a6a8 = a13*a3*a6*a8;
    GFloat a1a14a7a8 = a1*a14*a7*a8;
    GFloat a13a2a7a8 = a13*a2*a7*a8;
    GFloat a15a2a4a9 = a15*a2*a4*a9;
    GFloat a14a3a4a9 = a14*a3*a4*a9;
    GFloat a0a15a6a9 = a0*a15*a6*a9;
    GFloat a12a3a6a9 = a12*a3*a6*a9;
    GFloat a0a14a7a9 = a0*a14*a7*a9;
    GFloat a12a2a7a9 = a12*a2*a7*a9;
    GFloat a11a14a5 = a11*a14*a5;
    GFloat a10a15a5 = a10*a15*a5;
    GFloat a11a13a6 = a11*a13*a6;
    GFloat a10a13a7 = a10*a13*a7;
    GFloat a15a6a9 = a15*a6*a9;
    GFloat a14a7a9 = a14*a7*a9;
    GFloat a1a11a14 = a1*a11*a14;
    GFloat a1a10a15 = a1*a10*a15;
    GFloat a11a13a2 = a11*a13*a2;
    GFloat a10a13a3 = a10*a13*a3;
    GFloat a15a2a9 = a15*a2*a9;
    GFloat a14a3a9 = a14*a3*a9;
    GFloat a15a2a5 = a15*a2*a5;
    GFloat a14a3a5 = a14*a3*a5;
    GFloat a1a15a6 = a1*a15*a6;
    GFloat a13a3a6 = a13*a3*a6;
    GFloat a1a14a7 = a1*a14*a7;
    GFloat a13a2a7 = a13*a2*a7;
    GFloat a11a2a5 = a11*a2*a5;
    GFloat a10a3a5 = a10*a3*a5;
    GFloat a1a11a6 = a1*a11*a6;
    GFloat a1a10a7 = a1*a10*a7;
    GFloat a3a6a9 = a3*a6*a9;
    GFloat a2a7a9 = a2*a7*a9;
    GFloat a11a14a4 = a11*a14*a4;
    GFloat a10a15a4 = a10*a15*a4;
    GFloat a11a12a6 = a11*a12*a6;
    GFloat a10a12a7 = a10*a12*a7;
    GFloat a15a6a8 = a15*a6*a8;
    GFloat a14a7a8 = a14*a7*a8;
    GFloat a0a11a14 = a0*a11*a14;
    GFloat a0a10a15 = a0*a10*a15;
    GFloat a11a12a2 = a11*a12*a2;
    GFloat a10a12a3 = a10*a12*a3;
    GFloat a15a2a8 = a15*a2*a8;
    GFloat a14a3a8 = a14*a3*a8;
    GFloat a15a2a4 = a15*a2*a4;
    GFloat a14a3a4 = a14*a3*a4;
    GFloat a0a15a6 = a0*a15*a6;
    GFloat a12a3a6 = a12*a3*a6;
    GFloat a0a14a7 = a0*a14*a7;
    GFloat a12a2a7 = a12*a2*a7;
    GFloat a11a2a4 = a11*a2*a4;
    GFloat a10a3a4 = a10*a3*a4;
    GFloat a0a11a6 = a0*a11*a6;
    GFloat a0a10a7 = a0*a10*a7;
    GFloat a3a6a8 = a3*a6*a8;
    GFloat a2a7a8 = a2*a7*a8;
    GFloat a11a13a4 = a11*a13*a4;
    GFloat a11a12a5 = a11*a12*a5;
    GFloat a15a5a8 = a15*a5*a8;
    GFloat a13a7a8 = a13*a7*a8;
    GFloat a15a4a9 = a15*a4*a9;
    GFloat a12a7a9 = a12*a7*a9;
    GFloat a1a11a12 = a1*a11*a12;
    GFloat a0a11a13 = a0*a11*a13;
    GFloat a1a15a8 = a1*a15*a8;
    GFloat a13a3a8 = a13*a3*a8;
    GFloat a0a15a9 = a0*a15*a9;
    GFloat a12a3a9 = a12*a3*a9;
    GFloat a1a15a4 = a1*a15*a4;
    GFloat a13a3a4 = a13*a3*a4;
    GFloat a0a15a5 = a0*a15*a5;
    GFloat a12a3a5 = a12*a3*a5;
    GFloat a1a12a7 = a1*a12*a7;
    GFloat a0a13a7 = a0*a13*a7;
    GFloat a1a11a4 = a1*a11*a4;
    GFloat a0a11a5 = a0*a11*a5;
    GFloat a3a5a8 = a3*a5*a8;
    GFloat a1a7a8 = a1*a7*a8;
    GFloat a3a4a9 = a3*a4*a9;
    GFloat a0a7a9 = a0*a7*a9;
    GFloat a10a13a4 = a10*a13*a4;
    GFloat a10a12a5 = a10*a12*a5;
    GFloat a14a5a8 = a14*a5*a8;
    GFloat a13a6a8 = a13*a6*a8;
    GFloat a14a4a9 = a14*a4*a9;
    GFloat a12a6a9 = a12*a6*a9;
    GFloat a1a10a12 = a1*a10*a12;
    GFloat a0a10a13 = a0*a10*a13;
    GFloat a1a14a8 = a1*a14*a8;
    GFloat a13a2a8 = a13*a2*a8;
    GFloat a0a14a9 = a0*a14*a9;
    GFloat a12a2a9 = a12*a2*a9;
    GFloat a1a14a4 = a1*a14*a4;
    GFloat a13a2a4 = a13*a2*a4;
    GFloat a0a14a5 = a0*a14*a5;
    GFloat a12a2a5 = a12*a2*a5;
    GFloat a1a12a6 = a1*a12*a6;
    GFloat a0a13a6 = a0*a13*a6;
    GFloat a1a10a4 = a1*a10*a4;
    GFloat a0a10a5 = a0*a10*a5;
    GFloat a2a5a8 = a2*a5*a8;
    GFloat a1a6a8 = a1*a6*a8;
    GFloat a2a4a9 = a2*a4*a9;
    GFloat a0a6a9 = a0*a6*a9;

    GFloat det = a1a11a14a4 - a1a10a15a4 - a11a13a2a4 + a10a13a3a4 - 
	a0a11a14a5 + a0a10a15a5 + a11a12a2a5 - a10a12a3a5 - 
	    a1a11a12a6 + a0a11a13a6 + a1a10a12a7 - a0a10a13a7 - 
		a15a2a5a8 + a14a3a5a8 + a1a15a6a8 - a13a3a6a8 - 
		    a1a14a7a8 + a13a2a7a8 + a15a2a4a9 - a14a3a4a9 - 
	       a0a15a6a9 + a12a3a6a9 + a0a14a7a9 - a12a2a7a9;

    im.val[0][0] = (-a11a14a5 + a10a15a5 + a11a13a6 
		    -a10a13a7 - a15a6a9 + a14a7a9)/det;
    im.val[0][1] = ( a1a11a14 - a1a10a15 - a11a13a2 
		+ a10a13a3 + a15a2a9 - a14a3a9)/det;
    im.val[0][2] = (-a15a2a5 + a14a3a5 + a1a15a6 
		    - a13a3a6 - a1a14a7 + a13a2a7)/det;
    im.val[0][3] = (a11a2a5 - a10a3a5 - a1a11a6 
		    + a1a10a7 + a3a6a9 - a2a7a9)/det; 
    im.val[1][0] = (a11a14a4 - a10a15a4 - a11a12a6 
		    + a10a12a7 + a15a6a8 - a14a7a8)/det;
    im.val[1][1] = (-a0a11a14 + a0a10a15 + a11a12a2 
		    - a10a12a3 - a15a2a8 + a14a3a8)/det;
    im.val[1][2] = (a15a2a4 - a14a3a4 - a0a15a6 
		    + a12a3a6 + a0a14a7 - a12a2a7)/det;
    im.val[1][3] = (-a11a2a4 + a10a3a4 + a0a11a6 
		    - a0a10a7 - a3a6a8 + a2a7a8)/det; 
    im.val[2][0] = (-a11a13a4 + a11a12a5 - a15a5a8 
		    + a13a7a8 + a15a4a9 - a12a7a9)/det;
    im.val[2][1] = (- a1a11a12 + a0a11a13 + a1a15a8 
		    - a13a3a8 - a0a15a9 + a12a3a9)/det;
    im.val[2][2] = (-a1a15a4 + a13a3a4 + a0a15a5 
		    - a12a3a5 + a1a12a7 - a0a13a7)/det;
    im.val[2][3] = (a1a11a4 - a0a11a5 + a3a5a8 
		    - a1a7a8 - a3a4a9 + a0a7a9)/det;
    im.val[3][0] = (a10a13a4 - a10a12a5 + a14a5a8 
		    -a13a6a8 - a14a4a9 + a12a6a9)/det;
    im.val[3][1] = (a1a10a12 - a0a10a13 - a1a14a8 
		    + a13a2a8 + a0a14a9 - a12a2a9)/det;
    im.val[3][2] = (a1a14a4 - a13a2a4 - a0a14a5 
		    + a12a2a5 - a1a12a6 + a0a13a6)/det;
    im.val[3][3] = (-a1a10a4 + a0a10a5 - a2a5a8 
		    + a1a6a8 + a2a4a9 - a0a6a9)/det;

    im.changed();
    return(im);
}

GMatrix4 GMatrix4::transpose(void) const {
    GMatrix4 tm;
    for (int i = 0; i < 4; i++) 
	for (int j = 0; j < 4; j++)
	    tm.val[i][j] = val[j][i];
    tm.changed();
    return(tm);
}

GMatrix4 GMatrix4::operator+(const GMatrix4& b) const {
    GMatrix4 m;
    for (int i = 0; i < 4; i++) 
	for (int j = 0; j < 4; j++)
	    m.val[i][j] = val[i][j] + b.val[i][j];
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::operator-(const GMatrix4& b) const {
    GMatrix4 m;
    for (int i = 0; i < 4; i++) 
	for (int j = 0; j < 4; j++)
	    m.val[i][j] = val[i][j] - b.val[i][j];
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::operator*(const GMatrix4& b) const {
    GMatrix4 m;
    for (int i = 0; i < 4; i++) 
	for (int j = 0; j < 4; j++) {
	    m.val[i][j] = 0.0;
	    for (int k = 0; k < 4; k++)
		m.val[i][j] += val[i][k] * b.val[k][j];
	}
    m.changed();
    return(m);
}

GVector3 GMatrix4::operator*(const GVector3& v) const {
    GVector3 r;
    r.mX = v.mX * val[0][0] + v.mY * val[0][1] + v.mZ * val[0][2] + val[0][3];
    r.mY = v.mX * val[1][0] + v.mY * val[1][1] + v.mZ * val[1][2] + val[1][3];
    r.mZ = v.mX * val[2][0] + v.mY * val[2][1] + v.mZ * val[2][2] + val[2][3];
    
    GFloat w = v.mX * val[3][0]
	+ v.mY * val[3][1]
	    + v.mZ * val[3][2]
		+ val[3][3];

    r.mX /= w; r.mY /= w; r.mZ /= w;
    return(r);
}








GMatrix4 GMatrix4::translate(const GVector3& v) const {
    GMatrix4 tmp, m;
    tmp.val[0][3] = v.mX; 
    tmp.val[1][3] = v.mY;
    tmp.val[2][3] = v.mZ;
    m = *this * tmp;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::invTranslate(const GVector3& v) const {
    GMatrix4 tmp, m;
    tmp.val[0][3] = -v.mX;
    tmp.val[1][3] = -v.mY;
    tmp.val[2][3] = -v.mZ;
    m = tmp * (*this);
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::rotateX(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(-angle);
    tmp.val[1][1] = cos_val; 
    tmp.val[2][2] = cos_val;
    tmp.val[1][2] = sin_val;
    tmp.val[2][1] = -sin_val;
 
    m = *this * tmp;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::invRotateX(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(angle);
    tmp.val[1][1] = cos_val; 
    tmp.val[2][2] = cos_val;
    tmp.val[1][2] = sin_val;
    tmp.val[2][1] = -sin_val;
    
    m = tmp * *this;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::rotateY(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(-angle);
    tmp.val[0][0] = cos_val; 
    tmp.val[2][2] = cos_val;
    tmp.val[0][2] = -sin_val;
    tmp.val[2][0] = sin_val;
    
    m = *this * tmp;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::invRotateY(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(angle);
    tmp.val[0][0] = cos_val; 
    tmp.val[2][2] = cos_val;
    tmp.val[0][2] = -sin_val;
    tmp.val[2][0] = sin_val;
    
    m = tmp * *this;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::rotateZ(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(-angle);
    tmp.val[0][0] = cos_val; 
    tmp.val[1][1] = cos_val;
    tmp.val[0][1] = sin_val;
    tmp.val[1][0] = -sin_val;
 
    m = *this * tmp;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::invRotateZ(const GFloat angle) const {
    GMatrix4 tmp, m;
    GFloat cos_val = cos(angle);
    GFloat sin_val = sin(angle);
    tmp.val[0][0] = cos_val; 
    tmp.val[1][1] = cos_val;
    tmp.val[0][1] = sin_val;
    tmp.val[1][0] = -sin_val;
 
    m = tmp * *this;
    m.changed();
    return(m);
}

GMatrix4 GMatrix4::operator*(const GFloat& d) const {
    GMatrix4 m;
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    m.val[i][j] = val[i][j] * d; 
    m.changed();
    return(m);
}

ostream& operator<<(ostream& pStream, const GMatrix4& pMatrix) {
    pStream << '{';
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    pStream << pMatrix.val[i][j] << ' ';
    return(pStream << '}');
}



GMatrix4::GMatrix4(const GMatrix4& b) { 
    for (register int i = 0; i < 4; i++) 
	for (register int j = 0; j < 4; j++)
	    val[i][j] = b.val[i][j];
    
    mChanged = b.mChanged;
}

GMatrix4& GMatrix4::operator=(const GMatrix4& b) { 
    for (register int i = 0; i < 4; i++) 
	for (register int j = 0; j < 4; j++)
	    val[i][j] = b.val[i][j];
    
    mChanged = b.mChanged;
    return(self);
}

GMatrix4& GMatrix4::operator+=(const GMatrix4& b) {
    for (register int i = 0; i < 4; i++) 
	for (register int j = 0; j < 4; j++)
	    val[i][j] += b.val[i][j];
    
    changed();
    return(self);
}

GMatrix4& GMatrix4::operator-=(const GMatrix4& b) {
    for (register int i = 0; i < 4; i++) 
	for (register int j = 0; j < 4; j++)
	    val[i][j] -= b.val[i][j];
    
    changed();
    return(self);
}

GMatrix4& GMatrix4::operator*=(const GMatrix4& b) {
    GMatrix4 m;
    for (register int i = 0; i < 4; i++) 
	for (register int j = 0; j < 4; j++) {
	    m.val[i][j] = 0.0;
	    for (register int k = 0; k < 4; k++)
		m.val[i][j] += (val[i][k] * b.val[k][j]);
	}
    
    self = m;
    changed();
    return(self);
}


GMatrix4& GMatrix4::operator*=(const GFloat& d) {
    for (register int i = 0; i < 4; i++)
	for (register int j = 0; j < 4; j++)
	    val[i][j] *= d; 
    
    changed();
    return(self);
}

// Create a general rotation matrix:
GMatrix4 GMatrix4::rotate(const GVector3& pAxis, GFloat pAngle) const {
    // unity matrix
    GMatrix4 m;
    
    GFloat sin_a, cos_a, n1, n2, n3;
    GVector3 n = pAxis.unitize();

    gSinCos(pAngle, sin_a, cos_a);

    n1 = n[0]; n2 = n[1]; n3 = n[2];

    m.val[0][0] = (n1*n1 + (1. - n1*n1)*cos_a);
    m.val[1][0] = (n1*n2*(1. - cos_a) + n3*sin_a);
    m.val[2][0] = (n1*n3*(1. - cos_a) - n2*sin_a);
    
    m.val[0][1] = (n1*n2*(1. - cos_a) - n3*sin_a);
    m.val[1][1] = (n2*n2 + (1. - n2*n2)*cos_a);
    m.val[2][1] = (n2*n3*(1. - cos_a) + n1*sin_a);
    
    m.val[0][2] = (n1*n3*(1. - cos_a) + n2*sin_a);
    m.val[1][2] = (n2*n3*(1. - cos_a) - n1*sin_a);
    m.val[2][2] = (n3*n3 + (1. - n3*n3)*cos_a);
    
    GMatrix4 ret = self * m;
    ret.changed();
    
    return (ret);
}



GVector3 GMatrix4::transformNormal(const GVector3& pN) const {
    GVector3 tN;
    tN.mX = pN.mX * val[0][0] + pN.mY * val[1][0] + pN.mZ * val[2][0];
    tN.mY = pN.mX * val[0][1] + pN.mY * val[1][1] + pN.mZ * val[2][1];
    tN.mZ = pN.mX * val[0][2] + pN.mY * val[1][2] + pN.mZ * val[2][2];
   
    return(tN.unitize());
} 
