//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gparam1D.h"

// methods of approximating b-spline curve:
GApproxSplineCurve::GApproxSplineCurve(int pNumber, const GResolution& pResol, const GSplineMode& pMode):
GCurveTopology(pNumber), mSpline(GSplineValues(pNumber, 0, pNumber >= 3 ? 3 : 1, pMode, G_CLAMPED)) {
    setResolution(pResol);
}

GApproxSplineCurve::GApproxSplineCurve(const GCurveVertexAggregate& pPoints, const GResolution& pResol, const GSplineMode& pMode): GCurveTopology(pPoints), mSpline(GSplineValues(pPoints.getNumber(), 0, pPoints.getNumber() >= 3 ? 3 : 1, pMode, G_CLAMPED)) {
    setResolution(pResol);
}

// set/get spline values
void GApproxSplineCurve::setOrder(int pOrder) {
    mSpline.setSpline().order = pOrder;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineCurve::setMode(const GSplineMode& pMode) {
    mSpline.setSpline().mode = pMode;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineCurve::setEnd(const GSplineEnd& pEnd) {
    mSpline.setSpline().end = pEnd;
    geometryChanged();
}

void GApproxSplineCurve::geometryChanged() {

    if(getNumber() != mSpline.getSpline().numVal)
	mSpline.setSpline().numVal = getNumber();

    if(getCurveDim() != mSpline.getSpline().dim)
	mSpline.setSpline().dim = getCurveDim();

    GMatrix tBase = mSpline.getBaseMatrix();
    // spline base
    GMatrix tBase1 = mSpline.getBaseMatrix1();
    // first derivatives of the spline base
    GMatrix tBase2 = mSpline.getBaseMatrix2();
    // second derivatives of the spline base

    for(int t = 0; t < getCurveDim(); t++) {
	GVector3 tPoint, tTang, tGrad;

	for(int i = 0; i < getNumber(); i++) {
	    tPoint += getVertex(i).getPoint() * tBase(t,i);
	    tTang += getVertex(i).getPoint() * tBase1(t,i);
	    tGrad += getVertex(i).getPoint() * tBase2(t,i);
	}

	setCurveVertex
	    (t, GCurveVertex(tPoint, tTang, tGrad, mSpline.getParVector()[t]));
    }

    computeNormals();
}

const GVector& GApproxSplineCurve::getParameterVector() {
    return mSpline.getParVector();
}

const GCurveVertex GApproxSplineCurve::getCurveVertexOn(GFloat pT) {
    GVector3 tPoint, tTang, tGrad, tNorm;
    GVector tTrans, tTrans1, tTrans2;

    tTrans = mSpline.getBaseTranslation(pT);
    tTrans1 = mSpline.getBaseTranslation1(pT);
    tTrans2 = mSpline.getBaseTranslation2(pT);

    for(int i = 0; i < getNumber(); i++) {
	tPoint += getVertex(i).getPoint() * tTrans[i];
	tTang += getVertex(i).getPoint() * tTrans1[i];
	tGrad += getVertex(i).getPoint() * tTrans2[i];
    }

    tNorm = tTang.cross(tTang.cross(tGrad));
    // compute normal

    return GCurveVertex(tPoint, tNorm, tTang, tGrad, pT);
}
     
// Fitting b-spline curve with natural contraints:
GFittingSplineCurve::GFittingSplineCurve(int pNumber, const GResolution& pResol, const GSplineMode& pMode): GCurveTopology(pNumber), mSpline(GSplineValues(pNumber, 0, 4, pMode, G_FREE)) {
    setResolution(pResol);
}

GFittingSplineCurve::GFittingSplineCurve(const GCurveVertexAggregate& pPoints, const GResolution& pResol, const GSplineMode& pMode): GCurveTopology(pPoints), mSpline(GSplineValues(pPoints.getNumber(), 0, 4, pMode, G_FREE)) {
    setResolution(pResol);
}

void GFittingSplineCurve::setMode(const GSplineMode& pMode) {
    mSpline.setMode(pMode);
    
    resolutionChanged();
    geometryChanged();
}

void GFittingSplineCurve::geometryChanged() {
    int i, t;

    if(getCurveDim() != mSpline.getDimension())
	mSpline.setDimension(getCurveDim());
    
    GMatrix mInput(getNumber(), 3);

    for(i = 0; i < getNumber(); i++)
	mInput.setRow(i, getVertex(i).getPoint());
    // copy the values to be interpolated
    
    mSpline.setValuesMatrix(mInput);

    for(t = 0; t < getCurveDim(); t++)
	setCurveVertex
	    (t, getCurveVertexOn(getParameterVector()[t]));

    computeNormals();
}

const GVector& GFittingSplineCurve::getParameterVector() {
    return mSpline.getParVector();
}

const GCurveVertex GFittingSplineCurve::getCurveVertexOn(GFloat pT) {
    GVector3 tPoint, tTang, tGrad, tNorm;
    GVector tTrans, tTrans1, tTrans2;

    GMatrix tVertices = mSpline.getVertexMatrix();
    
    tTrans = mSpline.getBaseTranslation(pT);
    tTrans1 = mSpline.getBaseTranslation1(pT);
    tTrans2 = mSpline.getBaseTranslation2(pT);

    for(int i = 0; i < tVertices.Rows(); i++) {
	tPoint += tVertices.getRow(i) * tTrans[i];
	tTang += tVertices.getRow(i) * tTrans1[i];
	tGrad += tVertices.getRow(i) * tTrans2[i];
    }

    tNorm = tTang.cross(tTang.cross(tGrad));
    // compute normal

    return GCurveVertex(tPoint, tNorm, tTang, tGrad, pT);
}

// Fitting b-spline curve with tangents on both ends set
GGeneralFittingSplineCurve::GGeneralFittingSplineCurve(int pNumber, const GResolution& pResol, const GSplineMode& pMode, const GVector3& pBeg, const GVector3& pEnd): GCurveTopology(pNumber), mSpline(GSplineValues(pNumber, 0, 4, pMode, G_FREE), pBeg, pEnd) {
    setResolution(pResol);
}

GGeneralFittingSplineCurve::GGeneralFittingSplineCurve(const GCurveVertexAggregate& pPoints, const GResolution& pResol, const GSplineMode& pMode, const GVector3& pBeg, const GVector3& pEnd): GCurveTopology(pPoints), mSpline(GSplineValues(pPoints.getNumber(), 0, 4, pMode, G_FREE), pBeg, pEnd) {
    setResolution(pResol);
}

void GGeneralFittingSplineCurve::setFirstTangent(const GVector3& pTng) {
    mSpline.setTangentBeg(pTng);
    geometryChanged();
}

void GGeneralFittingSplineCurve::setLastTangent(const GVector3& pTng) {
    mSpline.setTangentEnd(pTng);
    geometryChanged();
}

void GGeneralFittingSplineCurve::setMode(const GSplineMode& pMode) {
    mSpline.setMode(pMode);

    resolutionChanged();
    geometryChanged();
}

void GGeneralFittingSplineCurve::geometryChanged() {
    int i, t;

    if(getCurveDim() != mSpline.getDimension())
	mSpline.setDimension(getCurveDim());
    
    GMatrix mInput(getNumber(), 3);

    for(i = 0; i < getNumber(); i++)
	mInput.setRow(i, getVertex(i).getPoint());
    // copy the values to be interpolated
    
    mSpline.setValuesMatrix(mInput);

    for(t = 0; t < getCurveDim(); t++)
	setCurveVertex
	    (t, getCurveVertexOn(getParameterVector()[t]));
    
    computeNormals();
}

const GVector& GGeneralFittingSplineCurve::getParameterVector() {
    return mSpline.getParVector();
}

const GCurveVertex GGeneralFittingSplineCurve::getCurveVertexOn(GFloat pT) {
    GVector3 tPoint, tTang, tGrad, tNorm;

    GMatrix tVertices = mSpline.getVertexMatrix();
    GVector tTrans = mSpline.getBaseTranslation(pT);

    for(int i = 0; i < tVertices.Rows(); i++) {
	tPoint += tVertices.getRow(i) * tTrans[i];
	tTang += tVertices.getRow(i) * tTrans[i];
	tGrad += tVertices.getRow(i) * tTrans[i];
    }

    tNorm = tTang.cross(tTang.cross(tGrad));
    // compute normal

    return GCurveVertex(tPoint, tNorm, tTang, tGrad, pT);
}

